#include "registration.h"
RCSID(cPointing_cc, "$Id: cPointing.cc,v 1.25 2002/09/26 07:38:02 amn Exp $");

// cpointing.cc  --  pointing formulae

// $Log: cPointing.cc,v $
// Revision 1.25  2002/09/26 07:38:02  amn
// Defaulted to 22GHz VLBI pointing coeffiecient for Japanese Gigabit sampler expt.
//
// Revision 1.24  2001/05/29 09:36:48  amn
// Edits due to use as planet coordinate counter.
//
// Revision 1.23  2001/02/12 13:17:11  amn
// Added explicit cPointing::-references to calls of searchParam() in cPointin.g
//
// Revision 1.22  2001/02/12 12:25:32  amn
// Added explicit class names to calls of 'searchParam()'.
//
// Revision 1.21  2001/02/12 11:58:14  amn
// Added user interface coefficients for PP_ pointing parameters.
//
// Revision 1.20  2001/01/16 09:42:10  amn
// Added string-based update/get of pointing parameters.
//
// Revision 1.19  2000/12/30 09:46:39  amn
// Clipped refraction in range 0.0--0.622 degrees elevation.
//
// Revision 1.18  2000/12/18 07:33:38  amn
// Switched from 22GHz VLBI to 86GHz VLBI.
//
// Revision 1.17  2000/11/08 12:10:19  amn
// Updated az/el offset origin for 22GHz VLBI RX.
//
// Revision 1.16  2000/11/03 07:38:59  amn
// Updated IERS constants, activated 22GHz VLBI pointing coefficients.
//
// Revision 1.15  2000/07/28 07:35:25  amn
// Added HP GPS reported height.
//
// Revision 1.14  2000/07/28 07:09:43  amn
// Added a comment about location discovered / set into HP GPS.
//
// Revision 1.13  2000/04/13 19:22:48  amn
// First Apr-2000 90GHz RX pointing measurement fitting results used.
//
// Revision 1.12  2000/04/12 09:43:14  amn
// Back to using Apr-1999 90GHz RX pointing coefficients.
//
// Revision 1.11  2000/04/11 12:53:02  amn
// Changed to Nov-1998 90GHz RX pointing coefficients.
//
// Revision 1.10  2000/04/06 07:44:02  amn
// Put in initial 90GHz receiver pointing coefficients.
//
// Revision 1.9  2000/02/22 04:53:17  amn
// Changed 22GHz VLBI receiver pointing origin according to ptn measurements.
//
// Revision 1.8  2000/01/30 13:50:38  amn
// Formulae work as in old VAX vlbi_tracker.
//
// Revision 1.7  2000/01/29 20:30:59  amn
// Debugging offset formulae.
//
// Revision 1.6  2000/01/28 13:59:24  amn
// First use in 'antcon' for Linux-calculated coordinates.
//
// Revision 1.5  1999/12/03 15:08:39  amn
// Added references to VAX formulae origin.
//
// Revision 1.4  1999/12/03 13:20:36  amn
// Cleanups.
//
// Revision 1.3  1999/12/02 12:54:35  amn
// Added testing all planets, seems to work, quickly compared Sun and Moon
// to today's values in Astronomical Almanac 1999.
//
// Revision 1.2  1999/12/02 11:52:06  amn
// Created Makefile to support USNO Novas 2.0 plus JPL Ephemeris DE405.
//
// Revision 1.1  1999/12/01 15:53:55  amn
// First adaptation with both Novas and VAS Ada-based formulae.
//

/* OBJECT POINTING                                                  */
/*                                                                  */
/* Calculate the pointing (or tracking) error of a radio telescope. */
/*                                                                  */
/* Definitions: POINTING error is the angle between the direction   */
/*              in earth coordinates of the RF beam axis and the    */
/*              direction of the beam axis as indicated by the      */
/*              azimuth and elevation angle encoders.               */
/*              TRACKING error is the difference between the        */
/*              direction in earth coordinates of the RF beam axis  */
/*              and the direction angle (in earth coordinates)      */
/*              commanded by the computer. The tracking error       */
/*              includes all of the contributions of the pointing   */
/*              error plus the servo lag error, servo limit cycle   */
/*              and the computer error.                             */
/* The offset equations used in this program are the same as in the */
/* old Contobs program written in Ada.                              */
/*                                                                  */
/* Esa Vilenius, Metshovi Radio Observatory 16.11.1999             */


// Debug printf()'s, time-related and/or coordinate calculation-related.
#define DEBUG_TIME 0
#define DEBUG_COORDS 0


#include <iostream.h>
#include <math.h>
#include <assert.h>

#include <sys/time.h>  // for gettimeofday() and friends
#include <unistd.h>
#include <stdio.h>  // for printf() and friends

// ------ Math and trigonometrics...
// Pi and radians...
#ifndef M_PI
  static const double cPi = 3.141592653589793;
#else
  static const double cPi = M_PI;
#endif
static const double c2Pi = (2 * cPi);
static const double cRadToDeg = (180.0 / cPi);
static const double cDegToRad = (cPi / 180.0);
#define radToDeg(a) ((a) * cRadToDeg)
#define degToRad(a) ((a) * cDegToRad)
static const double cDegToHrs = (1 / 15.0);
static const double cHrsToDeg = (15.0);
#define degToHrs(a) ((a) * cDegToHrs)
#define hrsToDeg(a) ((a) * cHrsToDeg)
static const double cRadToHrs = (cRadToDeg * cDegToHrs);
static const double cHrsToRad = (cHrsToDeg * cDegToRad);
#define radToHrs(a) ((a) * cRadToHrs)
#define hrsToRad(a) ((a) * cHrsToRad)


// ----- Pointing offsets
// The generic pointing class, pointing functions and their
// private parameters.
class cPointing {
public:
  cPointing();
  virtual ~cPointing();
  virtual void printParameters();
  // Update named parameter with value 'p', return ==0 if no such parameter.
  virtual int updateParam(const char *pn, double p);
  // Get the value of named parameter, return ==0 if no such parameter.
  virtual int getParam(const char *pn, double *p);

  // The final pointing offsets derived from the pointing model.
  virtual double offsetAz(double az, double el);
  virtual double offsetEl(double az, double el);

  // Refraction modelling in this pointing formulae collection.
  virtual void refract(double inAz, double inEl,
                       double *outAz, double *outEl);
  // Simpler "elevation-only" refraction modelling.
  virtual double refractElOffset(double inEl);

protected:
  // Searches for a pointer to a named parameter plus its UI-coefficient.
  virtual double *searchParam(const char *pn);
  double paramUICoeff;
  // Location parameters used in ephemeric conversions.
  double latitude;  // positive is degrees north
  double longitudeDeg;  // positive is degrees east of Greenwich
  //double longitudeHrs;  // positive is hours east of Greenwich

  // "Surrounding" parameters used in refraction.
  double height;  // height (meters) (xxx: from sea level or what???)
  double temp;  // out temperature, in C
  double press;  // out atmospheric pressure, in mbar
  double relHum;  // out relative humidity, 100.0%
};  // cPointing

cPointing::cPointing()
{
  // Geodetic latitude 60d 12' 58" north, in degrees.
  this->latitude  = ( 60.0 + (12.0 + 58.0/60.0)/60.0);

  // Geodetic longitude 1h 37m 34.3s east, in hours.
  //this->longitudeHrs = (  1.0 + (37.0 + 34.3/60.0)/60.0);
  this->longitudeDeg = hrsToDeg(1.0 + (37.0 + 34.3/60.0)/60.0);;

  // xxx: On 28-Jul-2000 our HP GPS output in 
  // 'maser:/var/log/ntpstats/clockstats' had the following
  // (may have been set to the GPS, though...)
  // LAT      N  60:13:03.651
  // LON      E  24:23:37.065
  // HGT      +69.36 m  (MSL)

  // "Surrounding" parameters used in various refraction models.
  this->height = 85.0;  // height (meters), (xxx: from sea level or what???)
  this->temp = 2.0;  // out temperature, in C
  this->press = 1011.0;  // out atmospheric pressure, in mbar
  this->relHum = 30.0;  // out relative humidity, max. 100.0%
}

cPointing::~cPointing()
{
}

void
cPointing::printParameters()
{
  cout << "class cPointing:" << endl;
  cout << " latitude    " << this->latitude << "deg north" << endl;
  cout << " longitude   " << this->longitudeDeg << "deg east" << endl;
  //cout << " longitude   " << this->longitudeHrs << "hrs east" << endl;
  cout << " height      " << this->height << "m" << endl;
  cout << " temperature " << this->temp << "C" << endl;
  cout << " pressure    " << this->press << "mbar" << endl;
  cout << " humidity    " << this->relHum << "%" << endl;
}

// Searches for a pointer to a named parameter.
double *
cPointing::searchParam(const char *pn)
{
  this->paramUICoeff = 1.0;
  if (strcmp(pn, "latitude") == 0) {
    return &(this->latitude);
  } else if (strcmp(pn, "longitude") == 0) {
    return &(this->longitudeDeg);
  } else if (strcmp(pn, "height") == 0) {
    return &(this->height);
  } else if (strcmp(pn, "temperature") == 0) {
    return &(this->temp);
  } else if (strcmp(pn, "pressure") == 0) {
    return &(this->press);
  } else if (strcmp(pn, "humidity") == 0) {
    return &(this->relHum);
  } else {
    return NULL;
  } 
}  // searchParam

// Update named parameter with value 'p', return ==0 if no such parameter.
int
cPointing::updateParam(const char *pn, double p)
{
  double *d;

  if ( (d = cPointing::searchParam(pn)) ) {
    *d = p;
    return 1;
  } else {
    return 0;
  }
}  // updateParam

// Get the value of named parameter, return ==0 if no such parameter.
int
cPointing::getParam(const char *pn, double *p)
{
  double *d;

  if ( (d = cPointing::searchParam(pn)) ) {
    *p = *d;
    return 1;
  } else {
    return 0;
  }
}  // getParam

double
cPointing::offsetAz(double az, double el)
{
  assert( ("abstract cPointing::offsetAz() called" == NULL) );
  return (0.0);
}

double
cPointing::offsetEl(double az, double el)
{
  assert( ("abstract cPointing::offsetEl() called" == NULL) );
  return (0.0);
}

// The default implementation of az/el refraction:
// calculate refraction offset of el only to be summed to true el.
void
cPointing::refract(double inAz, double inEl,
                   double *outAz, double *outEl)
{
  *outAz = inAz;
  *outEl = inEl + this->refractElOffset(inEl);
}  // refract

double
cPointing::refractElOffset(double inEl)
{
  assert( ("abstract cPointing::refractElOffset() called" == NULL) );
  return (0.0);
}


// ----- The traditional VAX/Contobs/Fivpt pointing formulae.
// These formulae are calculated the pointing in the following order:
//     -- 1)Ephemeric_m.Convert
//     -- 2)Refraction correction (Refraction.Apparent)
//     -- 3)Pointing correction   (Pointing.Correct)
//     -- 4)Yawed  (Receiver_beams.Yawed)
//           ** just a cosine-corrected az offset from rx feed geometry

class cVAXPointing : public cPointing {
protected:  // cannot be private since these will be inherited to subclasses
  // Latest apparent coordinates.
  double appRa, appDec;
  // Temporary ra-dec offsets to the apparent ra-dec coordinates.
  double appRaOff, appDecOff;
  // Latest apparent coordinates.
  double appAz, appEl;  // without refraction nor pointing model
  double refrAppAz, refrAppEl;  // with refraction
  double finalAz, finalEl;  // with refr, pointing, beam yaw

  // 'public:' interface is the same as in superclass 'cPointing'
public:
  cVAXPointing();
  virtual ~cVAXPointing();
  virtual void printParameters();
  // Update named parameter with value 'p', return ==0 if no such parameter.
  virtual int updateParam(const char *pn, double p);
  // Get the value of named parameter, return ==0 if no such parameter.
  virtual int getParam(const char *pn, double *p);
  // 
  virtual double offsetAz(double az, double el);
  virtual double offsetEl(double az, double el);
  //
  virtual double refractElOffset(double inEl);

  // Sets this->appRa/Dec and ra-dec offsets.
  virtual void setApp(double ra, double dec);
  virtual void getApp(double *ra, double *dec);
  virtual void setRadecOff(double raoff, double decoff);

  // Updates from this->appRa/Dec --> appAz/El:
  virtual void convertUTC(struct timeval& tv);

  // Applies refraction, pointing corrections, and beam yaw 
  // to this->appRa/Dec --> appAz/El -coordinates ---> this->refrAppAz/El.
  virtual void applyCorrections(void);

  // Gets the final az/el for outputting to antenna computer.
  virtual double getFinalAz(void);
  virtual double getFinalEl(void);

protected:
  // Searches for a pointer to a named parameter.
  virtual double *searchParam(const char *pn);
private:
  // Pointing offset parameters in radians.
  double az_org;
  double az_null;
  double az_ell;
  double az_noe;
  double azhar;
  double azpha;
  double az_nod;
  double elhar;
  double elpha;
  double el_ell;
  double beam_ell;
  double beam_nod;

  // Receiver (azimuth) multi-beam offsets, in degrees.
  double beamYaw;

  // Refraction coefficients.
  double rc1, rc2, rc3, rc4, rc5, rc6;

  // Ephemeric pre-calculated values.
  double sinLatitude;  // sin(latitude)
  double cosLatitude;  // cos(latitude)
  double longitudeSecRad;  // longitude in seconds converted to radians
  double eq_eqSecRad;  // equation of the equinoxes, in seconds converted to radians
};  // cVAXPointing

  static const double second_of_time_in_radians_c = 0.727220521664304e-4;
  // directly from VAX ADA program to maintain maximum compatibility...


cVAXPointing::cVAXPointing()
{
#if 1
  // From VAX 'OFFSET22.VLBI;5' 16-MAR-1998 12:33:42.38.
  this->az_org   = degToRad( 39.18/1000.0);
  this->az_null  = degToRad( 27.93/1000.0);
  this->az_ell   = degToRad( -3.12/1000.0);
  this->az_noe   = degToRad(-46.01);
  this->azhar    = degToRad( -0.79/1000.0);
  this->azpha    = degToRad(104.80);
  this->az_nod   = degToRad(-11.48);
  this->elhar    = degToRad( -2.33/1000.0);
  this->elpha    = degToRad(359.54);
  this->el_ell   = degToRad( 26.44/1000.0);
  this->beam_ell = degToRad(267.42/1000.0);
  this->beam_nod = degToRad(  2.73/1000.0);

  // Change origin according to pointing measurements
  // on 20--22-Feb-2000:
  this->az_org   = degToRad( 29.18/1000.0);
  this->beam_ell = degToRad(282.42/1000.0);

  // Change origin according to pointing measurements
  // on 03--08-Nov-2000:
  this->az_org   = degToRad( 35.18/1000.0);
  this->beam_ell = degToRad(267.42/1000.0);

  // 22GHz VLBI RX beams:
  // 170.0      Beamyaw_eastCONST(12)  real beam???
  // -35.03     Beamyaw_westCONST(13)  planned virtual beam with chopper???
  // (These were just added to az, and we will continue that practice.)
  // From 'vlbi_tracker_p.ada;11':
  // -- 1997-04-11  amn  Now uses the last (second) 'beam_yaw' in pointing file
  // (before it was always 0.0).  It is beam B of 90GHz
  // receiver, ideally it should be A, but this needs
  // beam control box remote control.
  this->beamYaw = -35.03 / 1000.0;
    // (The second beam is the same the last VAX vlbi_tracker uses...)
#endif
#if 0
  // From VAX 'OFFSET_87.DAT;32 20-NOV-1998 12:23:27.63
  this->az_org   = degToRad( 21.96/1000.0);
  this->az_null  = degToRad( 49.05/1000.0);
  this->az_ell   = degToRad(  1.45/1000.0);
  this->az_noe   = degToRad( 11.01);
  this->azhar    = degToRad(  1.00/1000.0);
  this->azpha    = degToRad( 12.09);
  this->az_nod   = degToRad( -28.18);
  this->elhar    = degToRad( -3.27/1000.0);
  this->elpha    = degToRad(-14.28);
  this->el_ell   = degToRad( 26.64/1000.0);
  this->beam_ell = degToRad(269.46/1000.0);
  this->beam_nod = degToRad(  3.08/1000.0);

  // 90GHz VLBI RX beams:
  // 170.0      Beamyaw_eastCONST(12)  real beam??? A???
  // 116.5      Beamyaw_westCONST(13)  planned virtual beam with chopper??? B?
  // (These were just added to az, and we will continue that practice.)
  // From 'vlbi_tracker_p.ada;11':
  // -- 1997-04-11  amn  Now uses the last (second) 'beam_yaw' in pointing file
  // (before it was always 0.0).  It is beam B of 90GHz
  // receiver, ideally it should be A, but this needs
  // beam control box remote control.
  this->beamYaw = 170.00 / 1000.0;  // FS uses A beam
    // (The second beam is the same the last VAX vlbi_tracker uses...)
#endif
#if 0
  // From VAX 'OFFSET86.VLBI;29' 16-APR-1999 23:51:21.61.
  // Very strong suspicion that these produce quite large/wrong offsets.
  this->az_org   = degToRad( 28.23/1000.0);
  this->az_null  = degToRad( 44.25/1000.0);
  this->az_ell   = degToRad( -5.06/1000.0);
  this->az_noe   = degToRad(-53.72);
  this->azhar    = degToRad(  1.85/1000.0);
  this->azpha    = degToRad(105.06);
  this->az_nod   = degToRad(-131.84);
  this->elhar    = degToRad( -4.61/1000.0);
  this->elpha    = degToRad( 18.80);
  this->el_ell   = degToRad( 19.53/1000.0);
  this->beam_ell = degToRad(271.31/1000.0);
  this->beam_nod = degToRad( -4.69/1000.0);

  // 90GHz VLBI RX beams:
  // 170.0      Beamyaw_eastCONST(12)  real beam???
  // 164.90     Beamyaw_westCONST(13)  planned virtual beam with chopper???
  // (These were just added to az, and we will continue that practice.)
  // From 'vlbi_tracker_p.ada;11':
  // -- 1997-04-11  amn  Now uses the last (second) 'beam_yaw' in pointing file
  // (before it was always 0.0).  It is beam B of 90GHz
  // receiver, ideally it should be A, but this needs
  // beam control box remote control.
  this->beamYaw = 164.90 / 1000.0;
    // (The second beam is the same the last VAX vlbi_tracker uses...)
#endif
#if 0
  // From VAX 'OFFSET86.VLBI;28' 01-APR-1998 20:44.
  // First batch of pointing measurements in Apr-2000 done with this.
  this->az_org   = degToRad( 28.50/1000.0);
  this->az_null  = degToRad( 38.10/1000.0);
  this->az_ell   = degToRad( -0.57/1000.0);
  this->az_noe   = degToRad(-41.87);
  this->azhar    = degToRad(  2.56/1000.0);
  this->azpha    = degToRad( 13.06);
  this->az_nod   = degToRad(-126.70);
  this->elhar    = degToRad( -2.47/1000.0);
  this->elpha    = degToRad(-18.55);
  this->el_ell   = degToRad( 22.20/1000.0);
  this->beam_ell = degToRad(273.67/1000.0);
  this->beam_nod = degToRad( -1.93/1000.0);

  // 90GHz VLBI RX beams:
  // 170.0      Beamyaw_eastCONST(12)  real beam???
  // 162.38     Beamyaw_westCONST(13)  planned virtual beam with chopper???
  // (These were just added to az, and we will continue that practice.)
  // From 'vlbi_tracker_p.ada;11':
  // -- 1997-04-11  amn  Now uses the last (second) 'beam_yaw' in pointing file
  // (before it was always 0.0).  It is beam B of 90GHz
  // receiver, ideally it should be A, but this needs
  // beam control box remote control.
  this->beamYaw = 162.38 / 1000.0;
    // (The second beam is the same the last VAX vlbi_tracker uses...)
#endif
#if 0
  // From fitting of first batch of pointing measurements
  // of 90GHz receiver (mm-VLBI) in Apr-2000:
  this->az_org   = degToRad(  8.86/1000.0);
  this->az_null  = degToRad( 32.49/1000.0);
  this->az_ell   = degToRad( -4.99/1000.0);
  this->az_noe   = degToRad(-47.28);
  this->azhar    = degToRad( -4.19/1000.0);
  this->azpha    = degToRad( 74.56);
  this->az_nod   = degToRad(-35.51);
  this->elhar    = degToRad( -1.97/1000.0);
  this->elpha    = degToRad(  4.37);
  this->el_ell   = degToRad( 18.53/1000.0);
  this->beam_ell = degToRad(285.39/1000.0);
  this->beam_nod = degToRad( -2.20/1000.0);

  // 90GHz VLBI RX beams:
  // 170.0      Beamyaw_eastCONST(12)
  // 157.73     Beamyaw_westCONST(13)
  // (These were just added to az, and we will continue that practice.)
  // From 'vlbi_tracker_p.ada;11':
  // -- 1997-04-11  amn  Now uses the last (second) 'beam_yaw' in pointing file
  // (before it was always 0.0).
  this->beamYaw = 157.73 / 1000.0;
    // (The second beam is the same the last VAX vlbi_tracker uses...)
#endif

  // Old Contobs refraction coefficients:
  this->rc1 = 40.0;      // CONST(1)
  this->rc2 = 2.7;       // CONST(2)
  this->rc3 = 320.0e-6;  // CONST(3)
  this->rc4 = 42.5;      // CONST(4)
  this->rc5 = 0.4;       // CONST(5)
  this->rc6 = 2.64;      // CONST(6)

  //  Station location.
  this->sinLatitude = sin(degToRad(this->latitude));
  this->cosLatitude = cos(degToRad(this->latitude));
  // Our 'this->longitudeDeg' is in degrees (as required by Novas 2.0)
  // and VAX ADA code wants it in hours-->seconds, converted to radians.
  this->longitudeSecRad = degToHrs(this->longitudeDeg) * 60.0 * 60.0 * second_of_time_in_radians_c;
  // Equation of the equinoxes, in seconds converted to radians.
  // xxx: must update using 'earthtilt()' from Novas 2.0 library
  this->eq_eqSecRad = -0.9786;
  this->eq_eqSecRad *= second_of_time_in_radians_c;

  // Eliminate ra-dec offsets.
  this->appRaOff = this->appDecOff = 0.0;
}  // construct cVAXPointing

cVAXPointing::~cVAXPointing()
{
}

void
cVAXPointing::printParameters()
{
  cPointing::printParameters();
  cout << "class cVAXPointing:" << endl;
  cout << " az_org   " << radToDeg(az_org)  *1000.0 << "mdeg" << endl;
  cout << " az_null  " << radToDeg(az_null) *1000.0 << "mdeg" << endl;
  cout << " az_ell   " << radToDeg(az_ell)  *1000.0 << "mdeg" << endl;
  cout << " az_noe   " << radToDeg(az_noe)          << "deg" << endl;
  cout << " azhar    " << radToDeg(azhar)   *1000.0 << "mdeg" << endl;
  cout << " azpha    " << radToDeg(azpha)           << "deg" << endl;
  cout << " az_nod   " << radToDeg(az_nod)          << "deg" << endl;
  cout << " elhar    " << radToDeg(elhar)   *1000.0 << "mdeg" << endl;
  cout << " elpha    " << radToDeg(elpha)           << "deg" << endl;
  cout << " el_ell   " << radToDeg(el_ell)  *1000.0 << "mdeg" << endl;
  cout << " beam_ell " << radToDeg(beam_ell)*1000.0 << "mdeg" << endl;
  cout << " beam_nod " << radToDeg(beam_nod)*1000.0 << "mdeg" << endl;
  cout << " rc1..6   " 
       << this->rc1 << "  "
       << this->rc2 << "  "
       << this->rc3 << "  "
       << this->rc4 << "  "
       << this->rc5 << "  "
       << this->rc6 << endl;
}  // printParameters

// Searches for a pointer to a named parameter.
double *
cVAXPointing::searchParam(const char *pn)
{
  this->paramUICoeff = cRadToDeg * 1000.0;  //mdeg-->rad
  if (strcmp(pn, "az_org") == 0) {
    return &(this->az_org);
  } else if (strcmp(pn, "az_null") == 0) {
    return &(this->az_null);
  } else if (strcmp(pn, "az_ell") == 0) {
    return &(this->az_ell);
  } else if (strcmp(pn, "azhar") == 0) {
    return &(this->azhar);
  } else if (strcmp(pn, "elhar") == 0) {
    return &(this->elhar);
  } else if (strcmp(pn, "el_ell") == 0) {
    return &(this->el_ell);
  } else if (strcmp(pn, "beam_ell") == 0) {
    return &(this->beam_ell);
  } else if (strcmp(pn, "beam_nod") == 0) {
    return &(this->beam_nod);
  }
  this->paramUICoeff = cRadToDeg;  //deg-->rad
  if (strcmp(pn, "az_noe") == 0) {
    return &(this->az_noe);
  } else if (strcmp(pn, "azpha") == 0) {
    return &(this->azpha);
  } else if (strcmp(pn, "az_nod") == 0) {
    return &(this->az_nod);
  } else if (strcmp(pn, "elpha") == 0) {
    return &(this->elpha);
  }
  this->paramUICoeff = 1000.0;  //mdeg-->deg
  if (strcmp(pn, "beamYaw") == 0) {
    return &(this->beamYaw);
  }
  this->paramUICoeff = 1.0;
  if (strcmp(pn, "rc1") == 0) {
    return &(this->rc1);
  } else if (strcmp(pn, "rc2") == 0) {
    return &(this->rc2);
  } else if (strcmp(pn, "rc3") == 0) {
    return &(this->rc3);
  } else if (strcmp(pn, "rc4") == 0) {
    return &(this->rc4);
  } else if (strcmp(pn, "rc5") == 0) {
    return &(this->rc5);
  } else if (strcmp(pn, "rc6") == 0) {
    return &(this->rc6);
  } else {
    return NULL;
  } 
}  // searchParam

// Get the value of named parameter, return ==0 if no such parameter.
int
cVAXPointing::getParam(const char *pn, double *p)
{
  double *d;

  // Search for own parameters first (they "win" inherited ones).
  if ( (d = cVAXPointing::searchParam(pn)) ) {
    *p = *d * this->paramUICoeff;
    return 1;
  }
  // Search for inherited parameters if "own" parameters don't match.
  return (cPointing::getParam(pn, p));
}  // getParam

// Update named parameter with value 'p', return ==0 if no such parameter.
int
cVAXPointing::updateParam(const char *pn, double p)
{
  double *d;
  int found = 0;

  // Update all parameters with the same name, descendants first.
  if (cPointing::updateParam(pn, p)) {
    found = 1;
  }
  // Then update my "own" parameters.
  if ( (d = cVAXPointing::searchParam(pn)) ) {
    *d = p / this->paramUICoeff;
    found = 1;
  }
  // Fix "own" parameters that depend on previously-set values.
  if (strcmp(pn, "latitude") == 0) {
    this->sinLatitude = sin(degToRad(this->latitude));
    this->cosLatitude = cos(degToRad(this->latitude));
  } else if (strcmp(pn, "longitude") == 0) {
    // Our 'this->longitudeDeg' is in degrees (as required by Novas 2.0)
    // and VAX ADA code wants it in hours-->seconds, converted to radians.
    this->longitudeSecRad = degToHrs(this->longitudeDeg) * 60.0 * 60.0 * second_of_time_in_radians_c;
  }

  // xxx: double eq_eqSecRad;  handling missing

  return (found);
}  // updateParam


// These formulae are most probably based on the article
// "The Pointing Calibration of the Haystack Antenna"
// By M. L. Meeks, John A. Ball, and Anthony B. Hull
// In IEEE Transctions on Antennas and Propagation,
// Vol. AP-16, No. 6, November 1968.
// These formulae have assumed az in -180..180 degrees;
// this is fixed to the more general convention of 0..360.
double
cVAXPointing::offsetAz(double az, double el)
{
  double offset;

  az = degToRad(az - 180.0);
  el = degToRad(el);
#if 0
  offset  = -az_org+az_null*tan(el)+az_ell*cos(az_noe)*cos(az)*tan(el)
            -az_ell*sin(az_noe)*sin(az)*tan(el);
  offset += azhar*cos(azpha)*sin(2*az)+azhar*sin(azpha)*cos(2*az);
#elif 1
  offset  = -az_org + tan(el)*( az_null
                                + az_ell*cos(az_noe + az) )
            + azhar * sin(azpha + 2.0*az);
#else
  offset = 0.0;
#endif
  offset = -offset;  // VAX used to subtract, we always add offsets
  offset = radToDeg(offset);
  return(offset);
}  // offsetAz

double
cVAXPointing::offsetEl(double az, double el)
{
  double offset, dummy;

  az = degToRad(az - 180.0);
  el = degToRad(el);
#if 0
  dummy   = beam_nod*cos(az_nod)*sin(az)+beam_nod*sin(az_nod)*cos(az);
  offset  = -beam_ell+elhar*cos(elpha)*sin(2*az)+elhar*sin(elpha)*cos(2*az);
  offset += dummy-0.5*fabs(dummy)-el_ell*cos(el);
#elif 1
  dummy   = beam_nod * sin(az_nod + az);
  offset  = -beam_ell + dummy - 0.5*fabs(dummy) - el_ell*cos(el)
            + elhar * sin(elpha + 2.0*az);
#else
  offset = 0.0;
#endif
  offset = -offset;  // VAX used to subtract, we always add offsets
  offset = radToDeg(offset);
  return(offset);
}  // offsetEl

// Traditional refraction correction.
// This is based on the following information from MicroVAX on 19-Nov-1999:
#if 0
$ ll $DISK1:<kurpitsa>refr.dat

Directory $DISK1:<KURPITSA>

REFR.DAT;9                1  21-APR-1989 16:20:39.37  (RWED,RWED,RE,RWED)
REFR.DAT;8                1  12-APR-1989 11:00:17.14  (RWED,RWED,RE,RWED)
REFR.DAT;7                1  22-AUG-1988 15:34:54.38  (RWED,RWED,RE,RWED)

Total of 3 files, 3 blocks.
$ type $DISK1:<kurpitsa>refr.dat;9
40.0            CONST(1)
2.7             CONST(2)
320.0E-6        CONST(3)
42.5            CONST(4)
0.4             CONST(5)
2.64            CONST(6)
16.0            CONST(7)
0.01            CONST(8)
0.032           CONST(9)
0.014           CONST(10)
0.099           CONST(11)
0.107           CONST(12)
0.5             CONST(13)
0.350           CONST(14)
0.0185          CONST(15)
0.0465          CONST(16)
0.128           CONST(17)
$
#endif

// Calculate only elevation refraction offset based on el only.
// This refraction formula most probably originates from
// W. R. Iliff and J. M. Holt, "Use of surface refractivity
// in the empirical prediction of total atmospheric refraction,"
// J. Research NBS (Radio Propagation), vol. 87D, pp. 31--35,
// January--February 1963.
double
cVAXPointing::refractElOffset(double inEl)
{
  if (inEl <= 0.0) {
    // The formula doesn't work at all with negative elevation.
    return (0.0);
  } else if (inEl < 0.622) {
    // Near zero the formula returns very large refraction offsets.
    // Fix this by forcing the offset to go to zero.
    // 0.622 is the elevation when the formula returns 0.622 ==> 1.244
    // degrees for the total "refracted" elevation.
    return (inEl);
  }

  // We have elevation >0.0, calculate corrected elevation.
  return (radToDeg( -degToRad( rc1 / pow(inEl + rc2, 4.0) )
    + rc3 * ( 1.0/tan(degToRad(inEl)) - rc4 / pow(inEl + rc5, rc6) )
  ));
  // xxx: we need a new formula with takes temp/pressure/humid into account.
}  // refractElOffset

// Original VAX comments:
//  --  Number of Julian days since J2000.0,
//  --  the IAU(1976) standard epoch. The epoch
//  --  is AD 2000 Jan 1.5, JD 2451545.0.
//  --  Valid 1900..2099.
//  --  N. Holsti 1984.7.23 A.
//  --              87.3.21    Ada version
double
jd2m(int year,  // regular "2000" year
     int month,  // 1..12
     int day,  // 1..31
     double seconds  // 0.0..86400.0
)
{
  static const int firstDay[12] = {
    0,31,59,90,120,151,181,212,243,273,304,334
  };
  int y, leap, leaps;

  assert( (year >= 1900) && (year <= 2099) );
  assert( (month >= 1) && (month <= 12) );
  assert( (day >= 1) && (day <= 31) );
  // seconds can be outside of the "typical" range 0..86400.0, xxx: or can it?

  y = year - 2000;

  // "compute number of leap years in the interval"
  if (y < 1) {
    leaps = y/4;  // not after 2000
  } else {
    leaps = (y-1)/4 + 1;  // after 2000 (2000 is a leap year in itself).
  }

  // "see if this year is a leap year and after the leap day"
  if ( ((year % 4) == 0) && (month > 2) ) {
    leap = 1;  // leap year, later than Feb
  } else {
    leap = 0;  // not leap year, or Jan or Feb.
  }

  return ( (double)(y*365  // "number of years"
            + leaps  // "number of leap years"
            + firstDay[month-1]  // "number of days to day 0 of month"
            + leap  // "including leap day, if any"
            + day)
           + seconds / 86400.0  // "fraction of day"
           - 1.5 );  // "since the epoch is Jan 1.5"
}  // jd2m

// Returns the current Greenwich Mean Sidereal Time in radians
// derived from computer UTC clock (in VAX-compatible way...)
// (from VAX qso_sources:ephemeric.ada).
double
gmstRadUTC(struct timeval& tv)
{
  struct tm *p;
  double seconds;
  double jc, gz;
  static const double gf_c = 1.0027379093;
  double temp;

  assert( (p = gmtime(&(tv.tv_sec))) != NULL );
  // 'p->tm_sec' is in range 0..59 (/60/61, takes leap seconds into account)

  // Original VAX comments:
  // "Julian centuries from J2000.0 to midnight"
  seconds = (double)(p->tm_sec + 60*(p->tm_min + 60*p->tm_hour))
    + (double)tv.tv_usec / 1000000.0;  // 0..86400.0...
  jc = jd2m(p->tm_year+1900,
            p->tm_mon+1,
            p->tm_mday,
            /*seconds=>*/ 0.0
            ) / 36525.0;
  // "GMST at 0h UT on day T"
  gz = ((( 0.0 - 6.2e-6           * jc
               + 0.093104       ) * jc
               + 8640184.812866 ) * jc
               + 24110.54841    ) * second_of_time_in_radians_c;
  gz = gz - (floor(gz/c2Pi) * c2Pi);  // xxx: floor() or rint()?
  if (gz < 0.0) {
    gz = gz + c2Pi;
  }
  temp = gz + (seconds * gf_c * second_of_time_in_radians_c);
  // xxx: temp in 0.0..4*pi??? apparently ok
#if DEBUG_TIME
  printf("gmstRadUTC: jd2m = %f, GMST = %f (in rad %f, gz = %f)\n",
         jc, temp/second_of_time_in_radians_c/60.0/60.0, temp, gz);
#endif
  return (temp);
}  // gmstRadUTC

void
cVAXPointing::setApp(double ra, double dec)
{
  this->appRa = ra;
  this->appDec = dec;
}  // setApp

void
cVAXPointing::getApp(double *ra, double *dec)
{
  *ra = this->appRa;  // without offsets; apparent (not topocentric)
  *dec = this->appDec;
}  // setApp

void
cVAXPointing::setRadecOff(double raoff, double decoff)
{
  this->appRaOff = raoff;
  this->appDecOff = decoff;
}  // setRadecOff

// Ephemeric conversion from current ra (in hours) & dec (in degrees)
// to non-refracted/non-pointing modelled az/el coordinates.
void
cVAXPointing::convertUTC(struct timeval& tv)
{
  double localHourAngle;  // in radians
  double decRad;
  double cosDec, sinDec, cosLhaCosDec;
  double naz, nel;  // new position

  localHourAngle =
    gmstRadUTC(tv) + this->eq_eqSecRad
    + this->longitudeSecRad
    - hrsToRad(this->appRa + this->appRaOff);
  decRad = degToRad(this->appDec + this->appDecOff);
  cosDec = cos(decRad);
  sinDec = sin(decRad);
  cosLhaCosDec = cos(localHourAngle) * cosDec;

  naz = atan2( /*y=>*/ sin(localHourAngle)*cosDec,
               /*x=>*/ cosLhaCosDec*this->sinLatitude - sinDec*this->cosLatitude );
  nel = asin( cosLhaCosDec*this->cosLatitude + sinDec*this->sinLatitude );
  // xxx: fix -180..180 to a wider range
  this->appAz = radToDeg(naz) + 180.0;  // fix VAX convention of -180..180
  this->appEl = radToDeg(nel);
}  // convertUTC

// Applies refraction, pointing corrections, and beam yaw 
// to this->appRa/Dec --> appAz/El -coordinates ---> this->finalAz/El.
void
cVAXPointing::applyCorrections(void)
{
  double oaz, oel;

  // In the following order (from original VAX/Ada/Contobs code):
  //     -- 1)Ephemeric_m.Convert
  //     -- 2)Refraction correction (Refraction.Apparent)
  //     -- 3)Pointing correction   (Pointing.Correct)
  //     -- 4)Yawed  (Receiver_beams.Yawed)
  //           ** just a cosine-corrected az offset from rx feed geometry.

  // 1: was done in 'updateApparent()' (or 'convertUTC()' old VAX code).
  // 2: refraction.
  // (This will overwrite 'refr*' values by Novas library...)
  this->refract(this->appAz, this->appEl,
                &(this->refrAppAz), &(this->refrAppEl));
    // ("VAX" refraction doesn't care if az is -180..180 or 0..360.)
  // 3: pointing model.
  oaz = this->offsetAz(this->refrAppAz, this->refrAppEl);
  oel = this->offsetEl(this->refrAppAz, this->refrAppEl);
  // 4: beam selection.
  // Yaw in azimuth to change beams of a multi-horn receiver.
  oaz += this->beamYaw / cos(degToRad(this->refrAppEl + oel));

  this->finalAz = this->refrAppAz + oaz;
  this->finalEl = this->refrAppEl + oel;
}  // applyCorrections(void)

double cVAXPointing::getFinalAz(void) {
  return (this->finalAz - 180.0  /* xxx: fixfix...*/);
}  // getFinalAz
double cVAXPointing::getFinalEl(void) {
  return (this->finalEl);
}  // getFinalEl


// ----- Another variant of refraction.
// From:
// "Ephem - an interactive astronomical ephemeris program"
// "Version 4.28 February 25, 1992"
// "Copyright (c) 1990,1991,1992 by Elwood Charles Downey"
// "Many formulas and tables are based, with permission, on material found in",
// "\"Astronomy with your Personal Computer\" by Dr. Peter Duffett-Smith,",
// "Cambridge University Press, (c) 1985.


/* correct the true altitude, ta, for refraction to the apparent altitude, aa,
 * returning the required offset to be added to true altitude
 * each in radians, given the local atmospheric pressure, pr, in mbars, and
 * the temperature, tr, in degrees C.
 */
static double rfract (double pr, double tr, double ta)
{
  double r;	/* refraction correction*/

  if (ta >= degToRad(15.0)) {
    /* Model for altitudes at least 15 degrees above horizon. */
    r = 7.888888e-5*pr/((273+tr)*tan(ta));

  } else if (ta > degToRad(-5.0)) {
    /* Hairier model for altitudes at least -5 and below 15 degrees. */
    double a, b, tadeg;

    tadeg = radToDeg(ta);
    a = ((2e-5*tadeg+1.96e-2)*tadeg+1.594e-1)*pr;
    b = (273+tr)*((8.45e-2*tadeg+5.05e-1)*tadeg+1);
    r = degToRad(a/b);
  } else {
    /* Do nothing if more than 5 degrees below horizon. */
    r = 0;
  }

  return (r);
}  // rfract

/* correct the apparent altitude, aa, for refraction to the true altitude, ta,
 * each in radians, given the local atmospheric pressure, pr, in mbars, and
 * the temperature, tr, in degrees C.
 */
#if 0
// Still unused...
static void
unrfract (double pr, double tr, double aa, double *ta)
{
  double err;
  double appar;
  double truAlt;

  /* iterative solution: search for the truAlt that refracts to the
   *   given apparent.
   * since refract() is discontinuous at -5 degrees, there is a range
   *   of apparent altitudes between about -4.5 and -5 degrees that are
   *   not invertable (the graph of ap vs. truAlt has a vertical step at
   *   truAlt = -5). thus, the iteration just oscillates if it gets into
   *   this region. if this happens the iteration is forced to abort.
   *   of course, this makes unrefract() discontinuous too.
   */
  truAlt = aa;
  do {
    appar = truAlt + rfract (pr, tr, truAlt);
    err = appar - aa;
    truAlt -= err;
  } while (fabs(err) >= 1e-6 && truAlt > degToRad(-5));

  *ta = truAlt;
}  // unrfract
#endif

// ----- This is for testing the alternative refraction formula
// from 'Ephem' above; apparently no good at radio wavelengths.
class cAltVAXPointing : public cVAXPointing {
  // 'public:' interface is the same as in superclass 'cVAXPointing'
public:
  virtual double refractElOffset(double inEl);
};  // cAltVAXPointing

double
cAltVAXPointing::refractElOffset(double inEl)
{
  return( radToDeg( ::rfract (this->press, this->temp, degToRad(inEl)) ) );
}  // refractElOffset


// ----- Then USNO Novas 2.0 -based cNovasPointing
extern "C" {
#include "../novas/novascon.h"
#include "../novas/solarsystem.h"
#include "../novas/novas.h"
};

class cNovasPointing : public cVAXPointing {
private:
  // Site location & environment formatted to Novas 2.0 'site_info' struct.
  site_info nloc;

  // ---- IERS time/earth rotation parameters ----
  // Conventionally-defined x coordinate of celestial ephemeris 
  // pole with respect to IERS reference pole, in arcseconds, directly from
  // http://maia.usno.navy.mil IERS Bulletin A 'finals.daily' files.
  // ("cep"=celestial ephemeris pole, "pm"=polar motion)
  double cep_pm_x;
  double cep_pm_y;

  // Celestial pole offsets in arcseconds from
  // http://maia.usno.navy.mil IERS Bulletin A 'finals.daily' files
  // (files have milliarcseconds).
  double delta_dpsi;
  double delta_deps;

  // Instantaneous (UT1 - UTC) time difference.
  // xxx: needs updating daily, perhaps interpolating more often?
  double deltaUT;

  // (TAI - UTC), integral # of seconds, "leap second adjustment".
  // xxx: must be changed at the same time when UTC/xntpd/GPS changes
  double deltaTAI_UTC;

  // (TDT - UTC), add this to get TT/TDT from UTC.
  double deltaTDT_UTC;

  // deltaT + UT1 --> TDT, the basis for Novas ephemeric formulae.
  // (Time difference of (TT/TDT - UT1).)
  // This is needed in 'equ2hor()' and other ephemeric calculations.
  double deltaT;

  // The latest star catalog entry (ra/dec) or a planet identification.
  int isPlanet;  // 0==is star, !=0 is a planet, below:
  body aPlanet;
  cat_entry aStar;

  // Latest apparent/topocentric coordinates.
  // double appRa, appDec;  // these inherited from cVAXPointing
  double appDist;  // from Earth in AU; planets only
  double topoRa, topoDec;
  double topoDist;  // from observer in AU; planets only

  // Ra-Dec offsets.
  // (Mainly for 'radecsweep's; Novas can only apply these to topo-ra-dec
  //  although the name implies "apparent".)
  // double appRaOff;  // inherited
  // double appDecOff;

  // Horizontal coordinates.
  // double appAz, appEl;  // inherited; without refraction nor pointing model
  // With (optical Novas) refraction (not really used):
  double refrTopoRa, refrTopoDec;
  // double refrAppAz, refrAppEl;  // inherited
protected:
  // Searches for a pointer to a named parameter.
  virtual double *searchParam(const char *pn);

  // 'public:' interface is the same as in superclass 'cVAXPointing'
public:
  cNovasPointing();
  virtual ~cNovasPointing();
  virtual void printParameters(void);
  // Update named parameter with value 'p', return ==0 if no such parameter.
  virtual int updateParam(const char *pn, double p);
  // Get the value of named parameter, return ==0 if no such parameter.
  virtual int getParam(const char *pn, double *p);

  // Julian dates from Unix clock.
  virtual double jd_timeval(struct timeval& tv, double secoff);
  virtual double utc_jd_utc(struct timeval& tv);
  virtual double tdt_jd_utc(struct timeval& tv);
  virtual double tdt_jd_now(void);

  // Set catalog-based ra/dec coordinates.
  // 'epoch' can be a fractional year like 1950.0 or 2000.0 (if < 10000.0)
  // or it can be a Julian date.
  virtual void setSource(const char *name,
                         double ra, double dec,
                         double epoch = 2000.0,
                         double radialvel = 0.0,
                         double promora = 0.0,
                         double promodec = 0.0,
                         double parallax = 0.0);
  // Routines for setting ra/dec to solar system objects.
  virtual void setSun(void);
  //virtual void setMoon(void);
  virtual void setPlanet(int planetNum);

  // Recalculate apparent/topocentric/horizontal coordinates.
  virtual void updateApparent(double tdt_jd);

  // Test VAX version of ephemeric conversions.
  virtual void testVAXApparent(struct timeval& tv);

  // Test calculating equation of the equinoxes and sideral times.
  virtual void testEEST(double tdt_jd);
};  // cNovasPointing

cNovasPointing::cNovasPointing()
{
  // Copy our location info to Novas 'site_info' struct.
  this->nloc.latitude = this->latitude;
  this->nloc.longitude = this->longitudeDeg;
  this->nloc.height = this->height;
  this->nloc.temperature = this->temp;  // xxx: needs updating after init
  this->nloc.pressure = this->press;  // xxx: needs updating after init

  // Conventionally-defined x coordinate of celestial ephemeris 
  // pole with respect to IERS reference pole, in arcseconds, directly from
  // http://maia.usno.navy.mil IERS Bulletin A 'finals.daily' files.
  // ("cep"=celestial ephemeris pole, "pm"=polar motion)
  this->cep_pm_x = -0.070193;
  this->cep_pm_y = +0.407899;

  // Celestial pole offsets in arcseconds, IERS in milliarcsecs.
  this->delta_dpsi = -0.053231;
  this->delta_deps = -0.001590;
  ::cel_pole(this->delta_dpsi, this->delta_deps);
  // xxx: need to call again, once a day/week???

  // (UT1 - UTC) time difference xxx: today.
  // This is available from IERS A bulletins 'finals.daily' / 'final.data'.
  // 'ftp://maia.usno.navy.mil/ser7/finals.daily'
  // xxx: needs updating daily, perhaps interpolating more often?
  this->deltaUT = +0.0882853;

  // (TAI - UTC), integral # of seconds, "leap second adjustment",
  // can be checked at 'ftp://maia.usno.navy.mil/ser7/tai-utc.dat'.
  // xxx: must be changed at the same time when UTC/xntpd/GPS changes
  // i.e. introduces leap seconds into UTC/xntpd time
  // => we'll need to know the UTC timepoint when it will change
  this->deltaTAI_UTC = 32.0;

  // (TDT - UTC), how to get TT/TDT from UTC.
  // This is needed when TDT Julian dates a generated
  // from computer/xntpd UTC time.
  // TDT = UT1 + deltaT = UTC + deltaUT + deltaT
  //     = UTC + (TAI - UTC) + 32.184
  this->deltaTDT_UTC = this->deltaTAI_UTC + 32.184;

  // Delta T, time difference of (TT/TDT - UT1).
  // This is also needed in 'equ2hor()' and other ephemeric calculations.
  // (You can check this at 'ftp://maia.usno.navy.mil/ser7/deltat.data'.)
  // deltaT = (TAI - UTC) + 32.184 - (UT1 - UTC)
  //        = deltaTDT_UTC - deltaUT;
  // xxx: needs updating daily
  this->deltaT = this->deltaTDT_UTC - this->deltaUT;

  // Set the default object to be the Sun.
  this->setSun();
}  // construct cNovasPointing

cNovasPointing::~cNovasPointing()
{
}

void
cNovasPointing::printParameters(void)
{
  cVAXPointing::printParameters();
  cout << "class cNovasPointing:" << endl;
  cout << " cep_pm_x     " << this->cep_pm_x << "arcsec" << endl;
  cout << " cep_pm_y     " << this->cep_pm_y << "arcsec" << endl;
  cout << " delta_dpsi   " << this->delta_dpsi << "arcsec" << endl;
  cout << " delta_deps   " << this->delta_deps << "arcsec" << endl;
  cout << " deltaUT      " << this->deltaUT << "sec" << endl;
  cout << " deltaTAI_UTC " << this->deltaTAI_UTC << "sec" << endl;
  cout << "  - calculated:" << endl;
  cout << " deltaTDT_UTC " << this->deltaTDT_UTC << "sec" << endl;
  cout << " deltaT       " << this->deltaT << "sec" << endl;
}  // printParameters

// Searches for a pointer to a named parameter.
double *
cNovasPointing::searchParam(const char *pn)
{
  this->paramUICoeff = 1.0;
  if (strcmp(pn, "latitude") == 0) {
    return &(this->nloc.latitude);
  } else if (strcmp(pn, "longitude") == 0) {
    return &(this->nloc.longitude);
  } else if (strcmp(pn, "height") == 0) {
    return &(this->nloc.height);
  } else if (strcmp(pn, "temperature") == 0) {
    return &(this->nloc.temperature);
  } else if (strcmp(pn, "pressure") == 0) {
    return &(this->nloc.pressure);
    // no humidity, btw...
  } else if (strcmp(pn, "cep_pm_x") == 0) {
    return &(this->cep_pm_x);
  } else if (strcmp(pn, "cep_pm_y") == 0) {
    return &(this->cep_pm_y);
  } else if (strcmp(pn, "delta_dpsi") == 0) {
    return &(this->delta_dpsi);
  } else if (strcmp(pn, "delta_deps") == 0) {
    return &(this->delta_deps);
  } else if (strcmp(pn, "deltaUT") == 0) {
    return &(this->deltaUT);
  } else if (strcmp(pn, "deltaTAI_UTC") == 0) {
    return &(this->deltaTAI_UTC);
  } else if (strcmp(pn, "deltaTDT_UTC") == 0) {  // get/display only
    return &(this->deltaTDT_UTC);
  } else if (strcmp(pn, "deltaT") == 0) {  // get/display only
    return &(this->deltaT);
  } else {
    return NULL;
  } 
}  // searchParam

// Get the value of named parameter, return ==0 if no such parameter.
int
cNovasPointing::getParam(const char *pn, double *p)
{
  double *d;

  // Search for own parameters first (they "win" inherited ones).
  if ( (d = cNovasPointing::searchParam(pn)) ) {
    *p = *d;
    return 1;
  }
  // Search for inherited parameters if "own" parameters don't match.
  return (cVAXPointing::getParam(pn, p));
}  // getParam

// Update named parameter with value 'p', return ==0 if no such parameter.
int
cNovasPointing::updateParam(const char *pn, double p)
{
  double *d;
  int found = 0;

  // Update all parameters with the same name, descendants first.
  if (cVAXPointing::updateParam(pn, p)) {
    found = 1;
  }
  // Then update my "own" parameters.
  if ( (d = cNovasPointing::searchParam(pn)) ) {
    if ( (d == &(this->deltaTDT_UTC)) || (d == &(this->deltaT)) ) {
      // Cannot change calculated parameters.
      return 0;  // not found
    }
    *d = p;
    found = 1;
  }
  // Fix "own" parameters that depend on previously-set values.
  // (Update every time, strcmp() is slower than recalculating...)
  this->deltaTDT_UTC = this->deltaTAI_UTC + 32.184;
  this->deltaT = this->deltaTDT_UTC - this->deltaUT;

  return (found);
}  // updateParam


// Converts a Unix 2xlong timeval (UTC/whatever) to Julian date.
double
cNovasPointing::jd_timeval(struct timeval& tv, double secoff)
{
  struct tm *p;
  double hourofday;
  double jd;

  assert( (p = gmtime(&(tv.tv_sec))) != NULL );
  // 'p->tm_sec' can be in range 0..59 (/60/61, takes leap seconds into account)
  // xxx: should check if another value for 'this->deltaTDT_UTC'
  //      is needed (not the value for current day)
  hourofday = p->tm_hour*60.0*60.0 + p->tm_min*60.0 + p->tm_sec
    + ((double)tv.tv_usec / 1000000.0) + secoff;
  hourofday /= 3600.0;  // from seconds to hours
#if DEBUG_TIME
  printf("tdt_jd_utc: %04d-%02d-%02d %02d:%02d:%02d.%06ld + secoff %.3f = hourofday %f\n",
         p->tm_year+1900,
         p->tm_mon+1,
         p->tm_mday,
         p->tm_hour,
         p->tm_min,
         p->tm_sec,
         tv.tv_usec,
         secoff,
         hourofday);
#endif
  jd = ::julian_date(p->tm_year+1900, p->tm_mon+1, p->tm_mday, hourofday);
#if DEBUG_TIME
  {
    short int yyyy, mm, dd;

    ::cal_date(jd, &yyyy, &mm, &dd, &hourofday);
    printf("tdt_jd_utc: tdt_jd = %f -> %04d-%02d-%02d hourofday %f\n",
           jd,
           yyyy, mm, dd,
           hourofday);
  }
#endif
  return (jd);
}  // jd_timeval

// Converts an xntpd/UTC time without any
// corrections/differences to get UTC as a Julian date.
double
cNovasPointing::utc_jd_utc(struct timeval& tv)
{
  return ( this->jd_timeval(tv, /*secoff=>*/ 0.0) );
}  // utc_jd_utc

// Converts an xntpd/UTC time applying
// all corrections/differences to get TDT,
// terrestial dynamic time for Novas 2.0 ephemeric calculations.
double
cNovasPointing::tdt_jd_utc(struct timeval& tv)
{
  return ( this->jd_timeval(tv, this->deltaTDT_UTC) );
}  // tdt_jd_utc

// Gets the current computer (xntpd/UTC) time
// applies all corrections/differences to get TDT,
// terrestial dynamic time for Novas 2.0 ephemeric calculations.
double
cNovasPointing::tdt_jd_now(void)
{
  struct timeval tv;

  assert( gettimeofday(&tv, /*tz=>*/ NULL) == 0 );
  return ( this->tdt_jd_utc(tv) );
  // assert( (p = gmtime(&(tv.tv_sec))) != NULL );
  // 'p->tm_sec' is in range 0..59 (/60/61, takes leap seconds into account)
  // xxx: 'p->tm_yday' could be put available somewhere 0..365, xxx:+1 for DOY?
  // (hmm, the typical convention in Astronomical Almanac seems to start
  //   day numbers from 1 for 1-Jan)
  // xxx: should check if 'this->deltaTDT_UTC' needs updating (leap sec)
}  // tdt_jd_now

// ----- Novas 2.0 basic structs for Earth, Sun, and Moon.
static /*const*/ body novasBodyEarth = {
  0, /* type: 0==major planet */
  3, /* number: 3==Earth */
  "Earth"
};
static /*const*/ body novasBodySun = {
  0, /* type: 0==major planet */
  10, /* number: 10==Sun */
  "Sun"
};
#if 0
static /*const*/ body novasBodyMoon = {  /* works only with solsys2.c /JPL data??? */
  0, /* type: 0==major planet */
  11, /* number: 11==Moon */
  "Moon"
};
#endif

// Set catalog-based ra/dec coordinates.
void
cNovasPointing::setSource(const char *name,
                          double ra, double dec,
                          double epoch = 2000.0,
                          double radialvel = 0.0,
                          double promora = 0.0,
                          double promodec = 0.0,
                          double parallax = 0.0)
{
  if ((epoch == 2000.0) || (epoch == T0)) {  // conventional or Julian 2000.0
    ::make_cat_entry("FK5", (char *)name, /*star_num=>*/ 0,
                     ra, dec, promora, promodec,
                     parallax, radialvel,
                     &(this->aStar));
  } else {
    cat_entry incat;

    ::make_cat_entry("FK5", (char *)name, /*star_num=>*/ 0,
                     ra, dec, promora, promodec,
                     parallax, radialvel,
                     &incat);
    // 'epoch' can be a fractional year like 1950.0 or 2000.0 (if < 10000.0)
    // or it can be a Julian date.
    ::transform_cat(/*option=>*/ 3,  // both equator and equinox and epoch
                    epoch, &incat,
                    /*newjd=>*/ T0, /*newcat_id=>*/ "FK5",
                    /*newcat=>*/ &(this->aStar));
  }  // else have to change epoch
  
  // We have now a star/quasar/whatever, not a planet.
  this->isPlanet = 0;
}  // setSource

void
cNovasPointing::setSun(void)
{
  this->isPlanet = 1;
  this->aPlanet = novasBodySun;
}  // setSun

void
cNovasPointing::setPlanet(int planetNum)
{
  this->isPlanet = 1;
  this->aPlanet.type = 0;  // major planet
  this->aPlanet.number = planetNum;
  strncpy(this->aPlanet.name, "Planet", sizeof(this->aPlanet.name));  // xxx:
}  // setSun

void
cNovasPointing::updateApparent(double tdt_jd)
{
  double dummy;

  if (this->isPlanet) {
    // Something in our solar system.
    assert(::app_planet(tdt_jd, &(this->aPlanet), &(novasBodyEarth),
                        &(this->appRa), &(this->appDec), &(this->appDist))
           == 0);
    assert(::topo_planet(tdt_jd, &(this->aPlanet), &(novasBodyEarth),
                         this->deltaT, &(this->nloc),
                         &(this->topoRa), &(this->topoDec), &(this->topoDist))
           == 0);
  } else {
    // Something not in our solar system...
    assert(::app_star(tdt_jd, &(novasBodyEarth),
                      &(this->aStar),
                      &(this->appRa), &(this->appDec))
           == 0);
    this->appDist = 0.0;  // don't know about distance...
    assert(::topo_star(tdt_jd, &(novasBodyEarth),
                       this->deltaT,
                       &(this->aStar),
                       &(this->nloc),
                       &(this->topoRa), &(this->topoDec))
           == 0);
    this->topoDist = 0.0;
  }

  // Now we have apparent and topocentric coordinates, calc horizontal.
  // Novas2.0 'equ2hor()' can only use topocentric coordinates to calc horiz.
  // xxx: hmm, ra-dec offsets to 'topo*' vs. 'app*'...?
  
  ::equ2hor(tdt_jd, this->deltaT,
            this->cep_pm_x, this->cep_pm_y,
            &(this->nloc),
            (this->topoRa + this->appRaOff), 
            (this->topoDec + this->appDecOff),
            /*ref_option=>*/ 0,  // no refraction at first
            &(this->appEl),
            &(this->appAz),
            &dummy, &dummy);
  this->appEl= 90.0 - this->appEl;  // from zenith distance to elevation
#if 0
  // And the same with refraction.  (Not any more, optical refr.)
  ::equ2hor(tdt_jd, this->deltaT,
            this->cep_pm_x, this->cep_pm_y,
            &(this->nloc),
            (this->topoRa + this->appRaOff), 
            (this->topoDec + this->appDecOff),
            /*ref_option=>*/ 2,  // refraction with temp/pressure
            &(this->refrAppEl),
            &(this->refrAppAz),
            &(this->refrTopoRa),
            &(this->refrTopoDec));
  this->refrAppEl = 90.0 - this->refrAppEl;
#else
  this->refrAppEl = this->appEl;
  this->refrAppAz = this->appAz;
  this->refrTopoRa = this->topoRa;
  this->refrTopoDec = this->topoDec;
#endif

#if DEBUG_COORDS
  printf("updateApparent: apparent    ra %f, dec %f, dist %f\n",
         appRa, appDec, appDist);
#if 0
  printf("updateApparent: topocentric ra %f, dec %f, dist %f\n",
         topoRa, topoDec, topoDist),
#endif
  printf("updateApparent: horizontal  az %11.6f, el %11.6f, refrEl %11.6f, diff %9.6f\n",
         appAz, appEl, refrAppEl, refrAppEl-appEl);
  // With refraction:
  //double refrTopoRa, refrTopoDec;
  //double refrAppAz, refrAppEl;
#endif
}  // updateApparent

void
cNovasPointing::testVAXApparent(struct timeval& tv)
{
  // Uses this->appRa, this->appDec; overwrites 'this->appAz/El'.
  this->convertUTC(tv);
  this->appDist = 0.0;
  this->applyCorrections();
#if DEBUG_COORDS
  printf("testVAXApparent: apparent   az %11.6f, el %11.6f\n",
         this->appAz, this->appEl);
  printf("testVAXApparent: final      az %11.6f, el %11.6f\n",
         this->finalAz, this->finalEl);
#endif
}  // testVAXApparent

void
cNovasPointing::testEEST(double tdt_jd)
{
  double dummy;
  double secdiff;
  double tdb_jd;
  double mobl, tobl, ee, dpsi, deps;
  double gmst, gast;

  // Find out equation of the equinoxes.
  // Compute first TDB date from TDT date.
  ::tdb2tdt(tdt_jd, &dummy, &secdiff);
  tdb_jd = tdt_jd + secdiff / 86400.0;
  // Call 'earthtilt()' to find out ee.
  ::earthtilt(tdb_jd,
              &mobl,&tobl,&ee,&dpsi,&deps);
  // Compute sidereal times, mean and apparent.
  ::sidereal_time(floor(tdt_jd), tdt_jd-floor(tdt_jd),
                  /*ee=>*/0.0, &gmst);
  ::sidereal_time(floor(tdt_jd), tdt_jd-floor(tdt_jd),
                  ee, &gast);
#if DEBUG_TIME
  printf("testEEST: eq_eq = %f, GMST = %f, GAST = %f\n",
         ee, gmst, gast);
#endif
}  // testEEST


// This is not a good idea: it doesn't iterate
// and this is apparently for optical wavelengths
// and it masks out old VAX refraction formula.
#if 0
double
cNovasPointing::refractElOffset(double inEl)
{
  return( ::refract(&(this->nloc), /*ref_option=>*/ 2, (90.0 - inEl)) );
  // xxx: doesn't iterate like 'equ2hor()' does!
}  // refractElOffset
#endif


// ----- Then new cMetPointing...
// (Pointing offset correction formulae replaced VAX-->something else.)


#if INCLUDE_TEST_MAIN
// ----- Test main program.
int
main(void)
{
  class cVAXPointing vVAX;
  class cAltVAXPointing vAlt;
  class cNovasPointing vNov;
  class cPointing *myPointing;
  double tjd;
  struct timeval tv;
  struct tm *p;
  struct tm tm;
  long int max;

  myPointing = &vNov;
#if 0
  myPointing->printParameters();
#endif
  myPointing = &vVAX;
#if 0
  cout << "OffsetAz " << myPointing->offsetAz(0.0, 45.0) << endl;
  cout << "OffsetEl " << myPointing->offsetEl(0.0, 45.0) << endl;
  cout << "Refr     " << myPointing->refractElOffset(7.0) << endl;
  myPointing = &vAlt;
  cout << "RefrAlt  " << myPointing->refractElOffset(7.0) << endl;
  myPointing = &vNov;
  cout << "RefrNov  " << myPointing->refractElOffset(7.0) << endl;
#endif
  // A test source for Novas library:
  //vNov.setSun();
  {
    double ra =  20.0 + 37.0/60.0 + 14.3/60.0/60.0;
    double dec = 42.0 +  9.0/60.0 +  0.0/60.0/60.0;
    vNov.setSource("dr21", ra, dec, /*epoch=>*/ 1950.0);
  }
  assert( gettimeofday(&tv, /*tz=>*/ NULL) == 0 );
  assert( (p = gmtime(&(tv.tv_sec))) != NULL );
  tm = *p;  // copy the struct
  tm.tm_sec = tm.tm_min = tm.tm_hour = 0;  // start at midnight UTC
  tv.tv_sec = mktime(&tm); tv.tv_usec = 0;
#if 0
  // xxx: mktime() interprets local time...
  tv.tv_sec += 2*(3600);
#endif
  max = tv.tv_sec + 86400;  // seconds, one day
#if 0
  for (;tv.tv_sec < max; tv.tv_sec += 3600 /*seconds, one hour*/) {
    tjd = vNov.tdt_jd_utc(tv);
    vNov.updateApparent(tjd);
    vNov.testEEST(tjd);
    vNov.testVAXApparent(tv);
  }
#endif

#if 1
  {
    int pl;

#if 0
    assert( gettimeofday(&tv, /*tz=>*/ NULL) == 0 );
    tv.tv_sec += 12*(3600);  // mid-UTC-day
#endif
    assert( (p = gmtime(&(tv.tv_sec))) != NULL );
    printf("At UTC %04d-%02d-%02d %02d:%02d:%02d:\n",
           p->tm_year + 1900, p->tm_mon + 1, p->tm_mday,
           p->tm_hour, p->tm_min, p->tm_sec);
    tjd = vNov.tdt_jd_utc(tv);
    for (pl=1; pl<=5; pl++) {
      vNov.setPlanet(pl);
      vNov.updateApparent(tjd);
    }  // for
    tv.tv_sec += 86400;
    assert( (p = gmtime(&(tv.tv_sec))) != NULL );
    printf("\nAt UTC %04d-%02d-%02d %02d:%02d:%02d:\n",
           p->tm_year + 1900, p->tm_mon + 1, p->tm_mday,
           p->tm_hour, p->tm_min, p->tm_sec);
    tjd = vNov.tdt_jd_utc(tv);
    for (pl=1; pl<=5; pl++) {
      vNov.setPlanet(pl);
      vNov.updateApparent(tjd);
    }  // for
  }
#endif
}  // main
#endif
