/* antcli.cc  --  Command line interface (via TCP/IP) for antcon. */

// $Log: antcli.cc,v $
// Revision 1.31  2005/10/12 06:30:53  lindi
// Document -U and -P in usage message
//
// Revision 1.30  2005/10/12 06:29:42  lindi
// Add support for -P<pidfile> and -U<user> (priviledge dropping is not implemented though)
//
// Revision 1.29  2005/05/16 11:16:23  lindi
// Fixed SOURCE handling with 4 arguments.
//
// Revision 1.28  2002/11/03 09:59:12  amn
// Taking the new 'common/tstamp.c' into use.
//
// Revision 1.27  2002/11/03 09:23:20  amn
// Moved UTC time ASCII --> double conversions to tstamp.cc.
//
// Revision 1.26  2002/04/05 09:55:13  amn
// Added clip[min|max][az|el]limits.
//
// Revision 1.25  2001/02/12 11:56:56  amn
// Added waiting for antcon PP_ updates before antcli replying.
//
// Revision 1.24  2001/01/16 09:41:36  amn
// Added antcli --> antcon --> sh updates of af/met/cPointing.cc parameters.
//
// Revision 1.23  2000/12/30 11:18:33  amn
// Automatically recompile both regular and SunAnt, refraction fix.
//
// Revision 1.22  2000/10/06 11:00:57  amn
// Fine-tuned LOG commands.
//
// Revision 1.21  2000/09/20 10:28:43  amn
// Added offset-free calculated azel to 'AZEL' and 'LOG=' reports.
//
// Revision 1.20  2000/09/20 10:15:04  amn
// Added min/ave/max azel error to LOG=.
//
// Revision 1.19  2000/07/28 12:02:02  amn
// Corrected 'radecsweep' RA acceleration (hours == /15.0!).
//
// Revision 1.18  2000/07/28 10:37:08  amn
// Wrong acceleration formula; removed centi- and millisecond time input.
//
// Revision 1.17  2000/07/28 09:52:10  amn
// Log file activation was forgotten; new screen layout in antcon.
//
// Revision 1.16  2000/07/28 09:28:17  amn
// Changed to using VC 'open'(1) and added logging 'log=' to antcli.
//
// Revision 1.15  2000/07/26 09:48:17  amn
// Added 'radecsweep' plus its display on antcon screen.
//
// Revision 1.14  2000/07/25 12:27:30  amn
// azelsweep works about ok.
//
// Revision 1.13  2000/07/24 09:57:11  amn
// Added 'updateSweep()' and reading in azelsweep values in 'antcli'.
//
// Revision 1.12  2000/06/20 13:42:32  amn
// Added 'tSweep' structs to sh.h for both ra-dec and az-el.
// Still missing 'updateSweep()' and antcli:'readSweep()'.
//
// Revision 1.11  2000/04/12 09:42:48  amn
// Added zeroing not-yet-available proper motion & parallax in 'source='.
//
// Revision 1.10  2000/04/06 07:46:59  amn
// Corrected source radial velocity input unit from degrees to m/s.
//
// Revision 1.9  2000/03/10 11:27:24  amn
// BBC osc frequency calculation in LINEFREQ/SOURCEVEL.
//
// Revision 1.8  2000/02/21 09:05:50  amn
// Changed all network write()s to 'sureWrite()' (no assert()s).
//
// Revision 1.7  2000/02/20 14:22:32  amn
// Added "safe position" limits to all tracking modes.
// Antcli now reports the true encoder position with 'azel'.
//
// Revision 1.6  2000/02/17 13:46:23  amn
// Added line frequency Doppler shift calculation.
//
// Revision 1.5  2000/02/16 12:58:48  amn
// Flushed stdout for disk file logging after each command line.
//
// Revision 1.4  2000/02/16 11:59:12  amn
// Added many commands to 'antcli'.
//
// Revision 1.3  2000/02/09 07:36:08  amn
// Azeloff, slewlimit etc. keywords.
//
// Revision 1.2  2000/02/06 17:56:35  amn
// Added the basic commands what the FS needs for pointing measurements.
//
// Revision 1.1  2000/02/06 15:07:03  amn
// Initial version.
//


#include "registration.h"
RCSID(antcli_cc, "$Id: antcli.cc,v 1.31 2005/10/12 06:30:53 lindi Exp $");

// Configuration/debugging defines.
// none yet

#define MY_NAME "antcli: "


#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <assert.h>

// For shared memory.
#include <sys/ipc.h>
#include <sys/shm.h>


// Socket server common code.
// (Consists of classes 'cMessageProcessor' and 'cSocketServer'.)
#include "ss.cc"


// For 'downlink.h'...
#define LINUX_TIME_AVAIL 1
#include "ctime.cpp"
#include "interloc.cpp"

// For 'tCoord'.
#include "downlink.h"

/*------------------------------------------------------------*/
// The complete and common shared struct
// across ant/dos/up/ant.cpp -- antcond (==antcon+antcli) -- xantcon.
//
// ant.cpp <--> antcond is transmitted field-by-field using COM2 serial,
// antcond <--> xantcon uses shared memory and the same struct layout.
#include "sh.h"

// A pointer to it, for shared memory references in Linux with 'sh->'.
ptSh sh = NULL;


// --------------------------------------------------
// Pi and radians...
#ifndef M_PI
  static const double cPi = 3.141592653589793;
#else
  static const double cPi = M_PI;
#endif
static const double c2Pi = (2 * cPi);
static const double cRadToDeg = (180.0 / cPi);
static const double cDegToRad = (cPi / 180.0);
#define radToDeg(a) ((a) * cRadToDeg)
#define degToRad(a) ((a) * cDegToRad)
static const double cDegToHrs = (1 / 15.0);
static const double cHrsToDeg = (15.0);
#define degToHrs(a) ((a) * cDegToHrs)
#define hrsToDeg(a) ((a) * cHrsToDeg)
static const double cRadToHrs = (cRadToDeg * cDegToHrs);
static const double cHrsToRad = (cHrsToDeg * cDegToRad);
#define radToHrs(a) ((a) * cRadToHrs)
#define hrsToRad(a) ((a) * cHrsToRad)

// --------------------------------------------------
// All-purpose routines for reading hours, degrees, years, Julian dates...
typedef enum {
  eHours,
  eDegrees
} tDHMSType;

int
getDHMS(
  const char *word,
  tDHMSType deflt,
  double *result
) {
  double temp;

  // xxx: Need to add recognizing field-separated formats.

  if (sscanf(word, "%lg", &temp) == 1) {
    // Conversion successful.
    *result = temp;
    return 0;
  } else {
    // Conversion failed, leave the result alone/unchanged.
    return 1;
  }
}  // getDHMS

int
getEpoch(
  const char *word,
  double *result
) {
  double temp;

  if (sscanf(word, "%lg", &temp) == 1) {
    // Conversion successful.
    // Novas 2.0 takes 0.0..9999.0 to mean year, and
    // 10000.0.. to mean Julian date, so we'll just allow everything positive.
    if (temp < 0.0) {
      return 1;  // failed
    }
    *result = temp;
    return 0;
  } else {
    // Conversion failed, leave the result alone/unchanged.
    return 1;
  }
}  // getEpoch

int
getVel(
  const char *word,
  double *result
) {
  double temp;

  if (sscanf(word, "%lg", &temp) == 1) {
    // Conversion successful.
    // xxx: For now, we'll just allow everything.
    *result = temp;
    return 0;
  } else {
    // Conversion failed, leave the result alone/unchanged.
    return 1;
  }
}  // getVel

typedef enum {
  eMHz
} tFreqType;

int
getFreq(
  const char *word,
  tFreqType deflt,  // MHz or something else
  double *result
) {
  double temp;

  if (sscanf(word, "%lg", &temp) == 1) {
    // Conversion successful.
    // xxx: For now, we'll just allow everything.
    *result = temp;
    return 0;
  } else {
    // Conversion failed, leave the result alone/unchanged.
    return 1;
  }
}  // getFreq

// Timestamp formatting.
#include "tstamp.c"


// ----------------------------------------------------------------------
// Code from Kari Leppanen's original 1992 VLBI pointing program 'peak.c'.
// Calculates radial observer velocity (in km/s?)
// at a given Unix (seconds since 1970) UTC time 'when'.

  // xxx: This should really be replaced with something
  // from Novas 2.0, as it calculates velocity vectors
  // internally all the time.
  // (It doesn't seem to export velocity functions, though.)

double
Observer_Velocity (double ra, double decl, time_t when)
{
  double T,eps,e,lop,M,C,v,V0,sdx,sdy,sdz,vx,vy,vz;

  T = (when/86400.0 - 10957.5) / 36525;  /* Julian centuries from J2000.0 */
  eps = degToRad(23.439291-.01300417*T);

  /* Elements of Earth's orbit */

  e = .016708617-.000042037*T-.0000001236*T*T;
  lop = degToRad(102.937348+1.7195269*T+.00045962*T*T+.000000499*T*T*T);
  M = degToRad(357.5291+35999.0503*T-.0001559*T*T-.00000048*T*T*T);
  C = degToRad((1.9146-.004817*T-.000014*T*T)*sin(M)
      + (.019993-.000101*T)*sin(2*M) + .00029*sin(3*M));
  v = M+C;  /* True anomaly */

  /* Earth velocity vector wrt LSR */

  V0 = sqrt(887.12696/(1-e*e));
  vx = 0.58-V0*(sin(v+lop)+e*sin(lop));
  vy = -16.71+V0*cos(eps)*(cos(v+lop)+e*cos(lop));
  vz = 9.85+V0*sin(eps)*(cos(v+lop)+e*cos(lop));

  ra = hrsToRad(ra); decl = degToRad(decl);

  sdx = cos(ra)*cos(decl);  /* Source direction unit vector */
  sdy = sin(ra)*cos(decl);
  sdz = sin(decl);

  return(vx*sdx+vy*sdy+vz*sdz);
}  // Observer_Velocity

double
Dopplered_Line_Freq (double ra, double decl, double source_vel,
                     double line_rest_freq)
{
  struct timeval tv;
  struct tm *tim;
  //double epoch_now;
  double observer_vel;

  assert( gettimeofday(&tv, NULL) == 0 );
  tim = gmtime(&tv.tv_sec);
  //epoch_now = 1900.0+tim->tm_year+tim->tm_yday/365.0;
  //ApparentCoordinates(ra,decl,epoch,&ra_now,&decl_now,epoch_now);

  observer_vel = Observer_Velocity(ra, decl, tv.tv_sec);

  return ( line_rest_freq
    * (1 + (observer_vel-source_vel)/299792.5)
  );
}  // Dopplered_Line_Freq


// ----------------------------------------------------------------------
// Planet name <--> (Novas) number table.
typedef struct sPlanetTableEntry {
  char *cliName;
  char *planetName;
  int planetType;
  int planetNumber;
} tPlanetTableEntry, *ptPlanetTableEntry;
tPlanetTableEntry planetTable[] = {
  { "sun", "Sun", 0, 10 },
  { "aurinko", "Sun", 0, 10 },
  { "moon", "Moon", 0, 11 },
  { "kuu", "Moon", 0, 11 },
  { "mercury", "Mercury", 0, 1 },
  { "merkurius", "Merkurius", 0, 1 },
  { "venus", "Venus", 0, 2 },
  { "earth", "Earth", 0, 3 },
  { "maa", "Maa", 0, 3 },
  { "mars", "Mars", 0, 4 },
  { "jupiter", "Jupiter", 0, 5 },
  { "saturn", "Saturn", 0, 6 },
  { "saturnus", "Saturnus", 0, 6 },
  { "uranus", "Uranus", 0, 7 },
  { "neptune", "Neptune", 0, 8 },
  { "neptunus", "Neptunus", 0, 8 },
  { "pluto", "Pluto", 0, 9 }
  // xxx: could add minor planets, ' 1, nnn'...
};  // planetTable
int numPlanets = (sizeof(planetTable)/sizeof(tPlanetTableEntry));


/*------------------------------------------------------------*/
// Antcon command line message processor is based on the generic
// TCP/IP socket server 'cMessageProcessor'.
class cMPAnt : public cMessageProcessor {
private:
  // Background/logging.
  int nextChannel;
public:
  virtual void init(void);
  virtual void cleanup(void);
  virtual ~cMPAnt();
  virtual int process_message_buffer(int filedes, char *buffer, int nbytes);
  virtual void doBackground(void);
};  // class cMPAnt


void
cMPAnt::init(void)
{
  this->nextChannel = 0;
}  // init

void
cMPAnt::cleanup(void)
{
}  // cleanup

cMPAnt::~cMPAnt(void)
{
  // xxx: ???
}  // destructor

void
sureWrite(int fd, const void *buf, size_t count)
{
  size_t wr;
  
  wr = write(fd, buf, count);
  if (wr != count) {
    fprintf(stderr, MY_NAME "sureWrite: wrote %d bytes, should have written %d ('%*s')\n", wr, count, count, (char *)buf);
  }
}  // sureWrite

void
precalculateSweep(tSweep *swp, double xacc, double yacc)
{
  double dur1, dur2;  // for determining the longer acc duration

  // Sweep (constant) velocity.
  swp->v.x = (swp->end.x - swp->start.x) / swp->dur;
  swp->v.y = (swp->end.y - swp->start.y) / swp->dur;

    // Determine the time needed to accelerate to the greater of (x/y) speeds.
  dur1 = fabs(swp->v.x) / xacc;
  dur2 = fabs(swp->v.y) / yacc;
  dur1 = (dur1 > dur2) ? dur1 : dur2;
  swp->accdur = dur1;
  swp->decdur = dur1;  // equal durations
  swp->keepdur = 5.0;  // xxx: seconds (for now...)

  // xxx:
  // if (now + accdur > startTime) then cannot do, no time...

  // Pre-calculate accelerations/decelerations.
  swp->acca.x = (dur1 > 0.0) ? (swp->v.x / dur1) : 0.0;
  swp->acca.y = (dur1 > 0.0) ? (swp->v.y / dur1) : 0.0;
  swp->deca.x = -(swp->acca.x);
  swp->deca.y = -(swp->acca.y);

    // Pre-calculate start and end positions after adding accelerations.
  swp->accStart.x = swp->start.x - (swp->v.x / 2.0 * dur1);
  swp->accStart.y = swp->start.y - (swp->v.y / 2.0 * dur1);
  swp->decEnd.x = swp->end.x + (swp->v.x / 2.0 * dur1);
  swp->decEnd.y = swp->end.y + (swp->v.y / 2.0 * dur1);
}  // precalculateSweep


int
cMPAnt::process_message_buffer(int filedes,
                               char *buffer,
                               int nbytes)
{
  char *words[NWORDS];
  const char delims[] = " \t,=\n\r";
  int number_of_words = 0;
  char *tok;
  char *running = NULL;
  int retCode;
  char replyBuffer[256];
  int rbytes = 0;

  if (nbytes < 1) {
    return 0;  // null request, processed okay
  }

  buffer[nbytes-1] = '\0';  // ensure that 'buffer' ends as a string
  // (this is written over a line termination character)

#if 1
  // Log the incoming command line.
  fprintf(stdout, "%s\t%s\n", getUTC(NULL), buffer);
  fflush(stdout);
#endif

  // Separate command line into words / tokens.
  // Make a copy since 'strsep()' modifies its buffer (puts in '\0' chars).
  assert( (running = strdup(buffer)) != NULL );
  while ( (tok = strsep(&running, delims)) ) {
    if (number_of_words < NWORDS) {
      if (*tok != '\0') {  // skip null arguments
        words[number_of_words] = tok;
        number_of_words++;
      }
    }
  }  // while tokens available

#if 0
  fprintf(stderr, "number_of_words = %d\n", number_of_words);
  {
    int i;

    for (i=0; i<number_of_words; i++) {
      fprintf(stderr, "words[%d] = '%s'\n", i, words[i]);
    }
  }
#endif

  // xxx: Should actually accumulate bytes until a '\n' is found,
  //      and then process each line as a message.
  //      (Note that WinNT 4.0 'telnet' sends each character immediately
  //       and thus using it requires this change.)

  retCode = 0;  // the default return code

  if (number_of_words < 1) {
    retCode = 0;  // null request, processed okay
    goto lExit;
  }

  // Check the first word, that is, the 'antcon' command word.

  // STANDBY
  if ((strcmp(words[0], "standby")==0) || (strcmp(words[0], "STANDBY")==0) ) {
    // Reply to network.
    sureWrite(filedes, "OK\n", 3);

    // Signal 'antcon' by incrementing the update flag counter.
    sh->antcliMode = ACM_standby;
    sh->antcliUpdates++;
    goto lExit;
  }  // if "standby"

  // VAXTRACK
  if ((strcmp(words[0], "vaxtrack")==0) || (strcmp(words[0], "VAXTRACK")==0) ) {
    // Reply to network.
    sureWrite(filedes, "OK\n", 3);

    // Signal 'antcon' by incrementing the update flag counter.
    sh->antcliMode = ACM_VAX_track;
    sh->antcliUpdates++;
    goto lExit;
  }  // if "vaxtrack"

  // AZEL
  if ((strcmp(words[0], "azel")==0) || (strcmp(words[0], "AZEL")==0) ) {

    // xxx: "service", az=38.5, el=0.5?

    if (number_of_words < 2) {
      // Report back the current fixed position settings.
      // xxx: or should we return the "real" encoder numbers?
      // Let's try actual incremental encoder numbers...
      rbytes += sprintf(replyBuffer, "OK,%.6f,%.6f,%.6f,%.6f\n",
#if 0
                        sh->myFixedPos.az,
                        sh->myFixedPos.el
#endif
                        sh->incr.az,
                        sh->incr.el,
                        sh->myTarget.az,  // calc azel /wo offsets
                        sh->myTarget.el
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    if (number_of_words < 3) {
      // 2 words, 1 parameter, only az given, needs el, too,
      // report "missing value"
      sureWrite(filedes, "MV,el\n", 6);
      goto lExit;
    }

    // At least 3 words, 2 parameters: azoff, eloff.
    if (getDHMS(words[1], /*deflt=>*/ eDegrees, &(sh->myFixedPos.az))) {
      sureWrite(filedes, "IV,az\n", 6);
      goto lExit;
    }
    if (getDHMS(words[2], /*deflt=>*/ eDegrees, &(sh->myFixedPos.el))) {
      sureWrite(filedes, "IV,el\n", 6);
      goto lExit;
    }

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    // Signal 'antcon' by incrementing the update flag counter.
    sh->antcliMode = ACM_azel;
    sh->antcliUpdates++;
    goto lExit;
  }  // if "azel"

  // PP_*
  // all pointing parameters (double values) are handled in the same way
  if ((strncmp(words[0], "pp_", 3)==0) || (strncmp(words[0], "PP_", 3)==0) ) {

    if (number_of_words < 2) {
      // Trig 'antcon' to get the value from pointing package.
      strncpy(sh->ppname, &((words[0])[3]), sizeof(sh->ppname));
      sh->getPP++;
      // Wait for getPP to go back to zero.
      while (sh->getPP > 0) {
        usleep(1000L);  // sleep a little, 1msec
        // xxx: should abort sleeps of >1sec or so
      }
      if (sh->foundPP) {
        // Report back the current pointing parameter value.
        rbytes += sprintf(replyBuffer, "OK,%f\n",
                          sh->ppvalue
                          );
      } else {
        // Unknown pointing parameter.
        rbytes += sprintf(replyBuffer, "UN,%s\n",
                          sh->ppname
                          );
      }
      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    // At least 2 words, 1 parameter: new pointing parameter value.
    // xxx: need a new routine to replace getVel()
    if (getVel(words[1], &(sh->ppvalue))) {
      sureWrite(filedes, "IV\n", 3);
      goto lExit;
    }

    // Real update deferred to 'antcon'.
    // xxx: wait for updatePP to go back to zero
    strncpy(sh->ppname, &((words[0])[3]), sizeof(sh->ppname));
    sh->updatePP++;
    // Wait for updatePP to go back to zero.
    while (sh->updatePP > 0) {
      usleep(1000L);  // sleep a little, 1msec
      // xxx: should abort sleeps of >1sec or so
    }
    if (sh->foundPP) {
      /* Build network reply buffer. */
      rbytes += sprintf(replyBuffer, "OK\n");
    } else {
      // Unknown pointing parameter.
      rbytes += sprintf(replyBuffer, "UN,%s\n",
                        sh->ppname
                        );
    }
    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);
    goto lExit;
  }  // if "pp_"

  // SOURCE
  if ((strcmp(words[0], "source")==0) || (strcmp(words[0], "SOURCE")==0) ) {

    if (number_of_words < 2) {
      // Report back the current 'source' settings.
      rbytes += sprintf(replyBuffer, "OK,%f,%f,%s,%f,%f,%f,%f\n",
                        sh->appRa,
                        sh->appDec,
                        sh->sourceName,
                        sh->sourceRa,
                        sh->sourceDec,
                        sh->sourceEpoch,
                        sh->sourceRadialvel
                        // xxx: proper motion ra, dec; parallax
                        );
      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    if (number_of_words < 3) {
      // 2 words, 1 parameter, i.e. a special source name, like planet.

      // xxx: "service", az=38.5, el=0.5?

      // Search for a name --> planet number.
      int i;

      for (i=0; i < numPlanets; i++) {
        if (strcasecmp(words[1], planetTable[i].cliName) == 0) {
          // Copy planet settings to 'sh'.
          sh->sourcePlanetType = planetTable[i].planetType;
          sh->sourcePlanetNumber = planetTable[i].planetNumber;
          strncpy(sh->sourceName, planetTable[i].planetName, sizeof(sh->sourceName));
          sh->sourceName[sizeof(sh->sourceName)-1] = '\0';
          sh->isPlanet = 1;
          sh->sourceRa = sh->sourceDec
            = sh->sourceEpoch = sh->sourceRadialvel = 0.0;

          // Reply to network.
          sureWrite(filedes, "OK\n", 3);

          // Signal 'antcon' by incrementing the update flag counter.
          sh->antcliMode = ACM_source;
          sh->antcliUpdates++;
          goto lExit;
        }
      }  // for
      sureWrite(filedes, "IV,planet\n", 10);
      goto lExit;
    }

    if (number_of_words < 4) {
      // 3 words, only right ascension given, needs declination, too.
      // report "missing value"
      sureWrite(filedes, "MV,dec\n", 7);
      goto lExit;
    }

    // At least 4 words, 3 parameters: source name, ra, dec.
    if (getDHMS(words[2], /*deflt=>*/ eHours, &(sh->sourceRa))) {
      sureWrite(filedes, "IV,ra\n", 6);
      goto lExit;
    }
    if (getDHMS(words[3], /*deflt=>*/ eDegrees, &(sh->sourceDec))) {
      sureWrite(filedes, "IV,dec\n", 7);
      goto lExit;
    }
    // If 4th parameter, it must be epoch, year or Julian day.
    if (number_of_words < 5) {
      // Default to 2000.0.
      sh->sourceEpoch = 2000.0;
    } else {
      if (getEpoch(words[4], &(sh->sourceEpoch))) {
        sureWrite(filedes, "IV,epoch\n", 9);
        goto lExit;
      }
    }
    // If 5th parameter, it must be source radial velocity.
    if (number_of_words < 6) {
      // Default to 0.
      sh->sourceRadialvel = 0.0;
    } else {
      if (getVel(words[5], &(sh->sourceRadialvel))) {
        sureWrite(filedes, "IV,vel\n", 7);
        goto lExit;
      }
    }
    // If both coordinates + epoch were valid, only then copy source name.
    strncpy(sh->sourceName, words[1], sizeof(sh->sourceName));
    sh->sourceName[sizeof(sh->sourceName)-1] = '\0';
    sh->isPlanet = 0;

    // xxx: rest of source parameters?
    sh->sourcePromora = 0.0;
    sh->sourcePromodec = 0.0;
    sh->sourceParallax = 0.0;

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    // Signal 'antcon' by incrementing the update flag counter.
    sh->antcliMode = ACM_source;
    sh->antcliUpdates++;
    goto lExit;
  }  // if "source"

  // SOURCEVEL
  // (for FS compatibility; "adding" source velocity after FS 'source=')
  if ((strcmp(words[0], "sourcevel")==0) || (strcmp(words[0], "SOURCEVEL")==0) ) {

    if (number_of_words < 2) {
      // Report back the current source velocity.
      rbytes += sprintf(replyBuffer, "OK,%.6f\n",
                        sh->sourceRadialvel
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    // At least 2 words, 1 parameter: new source radial velocity.
    if (getVel(words[1], &(sh->sourceRadialvel))) {
      sureWrite(filedes, "IV\n", 3);
      goto lExit;
    }

    /* Build network reply buffer. */
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);
    goto lExit;
  }  // if "sourcevel"

  // LINEFREQ
  if ((strcmp(words[0], "linefreq")==0) || (strcmp(words[0], "LINEFREQ")==0) ) {
    double line_rest_freq;  // all in MHz
    double dopplered;
    double lo_freq = 0.0;
    double bandwidth = 0.0;
    double osc_freq;

    if (number_of_words < 2) {
      /* Send network error message back to the client. */
      sureWrite(filedes, "MV\n", 3);
      goto lExit;
    }

    // At least 2 words, 1 parameter: line rest frequency (in MHz).
    if (getFreq(words[1], /*deflt=>*/ eMHz, &line_rest_freq)) {
      sureWrite(filedes, "IV\n", 3);
      goto lExit;
    }

    // Get LO frequency, bandwidth if given (for BBC freq calculation).
    if (number_of_words >= 3) {
      if (getFreq(words[2], /*deflt=>*/ eMHz, &lo_freq)) {
        sureWrite(filedes, "IV\n", 3);
        goto lExit;
      }
    }
    if (number_of_words >= 4) {
      if (getFreq(words[3], /*deflt=>*/ eMHz, &bandwidth)) {
        sureWrite(filedes, "IV\n", 3);
        goto lExit;
      }
    }

    // Calculate the current apparent (Doppler-corrected) line frequency,
    // and if required, also the "peak in the middle" osc setting frequency.
    // xxx: !!! Should definitely use cPointing::appRa/Dec instead
    //          of coordinates in source epoch.
    //          (Works for now since the FS gives today's ra/dec...)
    dopplered = Dopplered_Line_Freq(
      sh->sourceRa,
      sh->sourceDec,
      sh->sourceRadialvel,
      line_rest_freq
    );
    osc_freq = dopplered - lo_freq - bandwidth/2.0;

    /* Build network reply buffer. */
    rbytes += sprintf(replyBuffer, "OK,%.2f,%.2f\n",
      dopplered,
      osc_freq
    );

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);
    goto lExit;
  }  // if "linefreq"

  // AZELOFF
  if ((strcmp(words[0], "azeloff")==0) || (strcmp(words[0], "AZELOFF")==0) ) {

    if (number_of_words < 2) {
      // Report back the current offset settings.
      rbytes += sprintf(replyBuffer, "OK,%.6f,%.6f\n",
                        sh->myOffset.az,
                        sh->myOffset.el
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    if (number_of_words < 3) {
      // 2 words, 1 parameter, only az given, needs el, too,
      // report "missing value"
      sureWrite(filedes, "MV,el\n", 6);
      goto lExit;
    }

    // At least 3 words, 2 parameters: azoff, eloff.
    if (getDHMS(words[1], /*deflt=>*/ eDegrees, &(sh->myOffset.az))) {
      sureWrite(filedes, "IV,az\n", 6);
      goto lExit;
    }
    if (getDHMS(words[2], /*deflt=>*/ eDegrees, &(sh->myOffset.el))) {
      sureWrite(filedes, "IV,el\n", 6);
      goto lExit;
    }

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    //// Signal 'antcon' by incrementing the update flag counter.
    //sh->antcliUpdates++;  no need, offsets just take effect immediately

    goto lExit;
  }  // if "azeloff"

  // RADECOFF
  if ((strcmp(words[0], "radecoff")==0) || (strcmp(words[0], "RADECOFF")==0) ) {

    if (number_of_words < 2) {
      // Report back the current offset settings.
      rbytes += sprintf(replyBuffer, "OK,%.6f,%.6f\n",
                        sh->raOffset,
                        sh->decOffset
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    if (number_of_words < 3) {
      // 2 words, 1 parameter, only az given, needs el, too,
      // report "missing value"
      sureWrite(filedes, "MV,dec\n", 7);
      goto lExit;
    }

    // At least 3 words, 2 parameters: raoff, decoff.
    // 'raOffset' is in hours, but we default to getting it in degrees.
    if (getDHMS(words[1], /*deflt=>*/ eHours, &(sh->raOffset))) {
      sureWrite(filedes, "IV,ra\n", 6);
      goto lExit;
    }
    if (getDHMS(words[2], /*deflt=>*/ eDegrees, &(sh->decOffset))) {
      sureWrite(filedes, "IV,dec\n", 7);
      goto lExit;
    }

    // Convert from degrees to hours.
    sh->raOffset = degToHrs(sh->raOffset);

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    //// Signal 'antcon' by incrementing the update flag counter.
    //sh->antcliUpdates++;

    goto lExit;
  }  // if "radecoff"

  // xxx: 'xyoff='

  // AZELSWEEP
  if ((strcmp(words[0], "azelsweep")==0) || (strcmp(words[0], "AZELSWEEP")==0) ) {
    if (number_of_words < 2) {
      // Report back the current az-el sweep settings.
      if (sh->azelsweep.isActive) {
        rbytes += sprintf(replyBuffer,
                          "OK,%.2f,%.6f,%.6f,%.6f,%.6f,%.2f,%.6f,%.6f\n",
                          (sh->azelsweep.dur - sh->azelsweep.t),  // remaining
                          sh->azelsweep.start.x,
                          sh->azelsweep.start.y,
                          sh->azelsweep.end.x,
                          sh->azelsweep.end.y,
                          sh->azelsweep.accdur,
                          sh->azelsweep.v.x,
                          sh->azelsweep.v.y
                          );
      } else {
        rbytes += sprintf(replyBuffer, "OK,0\n");
      }

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    if (number_of_words < 7) {
      // Needs at least six parameters, report "missing value".
      sureWrite(filedes, "MV,sutc,sx,sy,ex,ey,dur\n", 24);
      goto lExit;
    }

    // 'antcli' user wishes to make changes, check for pending activity.
    if (sh->azelsweep.isActive) {
      // Terminate it immediately, recording the 'remaining' seconds
      // of the previous, interrupted sweep.
      sh->azelsweep.isActive = 0;
      sh->azelsweep.prevSweepInterruptTime =
        sh->azelsweep.dur - sh->azelsweep.t;
      // Then proceed to load new values.
    } else {
      sh->azelsweep.prevSweepInterruptTime = 0.0;
    }

    // At least 7 words, 6 parameters: sutc,sx,sy,ex,ey,dur.
    if (getDateTime(words[1], &(sh->azelsweep.startTime))) {
      sureWrite(filedes, "IV,startUTC\n", 12);
      goto lExit;
    }
    if (getDHMS(words[2], /*deflt=>*/ eDegrees, &(sh->azelsweep.start.x))) {
      sureWrite(filedes, "IV,saz\n", 7);
      goto lExit;
    }
    if (getDHMS(words[3], /*deflt=>*/ eDegrees,  &(sh->azelsweep.start.y))) {
      sureWrite(filedes, "IV,sel\n", 7);
      goto lExit;
    }
    if (getDHMS(words[4], /*deflt=>*/ eDegrees, &(sh->azelsweep.end.x))) {
      sureWrite(filedes, "IV,eaz\n", 7);
      goto lExit;
    }
    if (getDHMS(words[5], /*deflt=>*/ eDegrees,  &(sh->azelsweep.end.y))) {
      sureWrite(filedes, "IV,eel\n", 7);
      goto lExit;
    }
    if (getDuration(words[6], &(sh->azelsweep.dur))) {
      // xxx: should be positive, yes?
      sureWrite(filedes, "IV,dur\n", 7);
      goto lExit;
    }

    // Pre-calculate values for 'antcon'.
    precalculateSweep(&(sh->azelsweep),
                      /*xacc=>*/ 0.05,
                      /*yacc=>*/ 0.05
                      );

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    // Signal 'antcon' by incrementing the update flag counter.
    sh->azelsweep.isActive++;

    goto lExit;
  }  // if "azelsweep"

  // RADECSWEEP
  if ((strcmp(words[0], "radecsweep")==0) || (strcmp(words[0], "RADECSWEEP")==0) ) {
    if (number_of_words < 2) {
      // Report back the current ra-dec sweep settings.
      if (sh->radecsweep.isActive) {
        rbytes += sprintf(replyBuffer,
                          "OK,%.2f,%.6f,%.6f,%.6f,%.6f,%.2f,%.6f,%.6f\n",
                          (sh->radecsweep.dur - sh->radecsweep.t),  // remaining
                          sh->radecsweep.start.x,
                          sh->radecsweep.start.y,
                          sh->radecsweep.end.x,
                          sh->radecsweep.end.y,
                          sh->radecsweep.accdur,
                          sh->radecsweep.v.x,
                          sh->radecsweep.v.y
                          );
      } else {
        rbytes += sprintf(replyBuffer, "OK,0\n");
      }

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    if (number_of_words < 7) {
      // Needs at least six parameters, report "missing value".
      sureWrite(filedes, "MV,sutc,sx,sy,ex,ey,dur\n", 24);
      goto lExit;
    }

    // 'antcli' user wishes to make changes, check for pending activity.
    if (sh->radecsweep.isActive) {
      // Terminate it immediately, recording the 'remaining' seconds
      // of the previous, interrupted sweep.
      sh->radecsweep.isActive = 0;
      sh->radecsweep.prevSweepInterruptTime =
        sh->radecsweep.dur - sh->radecsweep.t;
      // Then proceed to load new values.
    } else {
      sh->radecsweep.prevSweepInterruptTime = 0.0;
    }

    // At least 7 words, 6 parameters: sutc,sx,sy,ex,ey,dur.
    if (getDateTime(words[1], &(sh->radecsweep.startTime))) {
      sureWrite(filedes, "IV,startUTC\n", 12);
      goto lExit;
    }
    if (getDHMS(words[2], /*deflt=>*/ eDegrees, &(sh->radecsweep.start.x))) {
      sureWrite(filedes, "IV,sra\n", 7);
      goto lExit;
    }
    if (getDHMS(words[3], /*deflt=>*/ eDegrees,  &(sh->radecsweep.start.y))) {
      sureWrite(filedes, "IV,sdec\n", 8);
      goto lExit;
    }
    if (getDHMS(words[4], /*deflt=>*/ eDegrees, &(sh->radecsweep.end.x))) {
      sureWrite(filedes, "IV,era\n", 7);
      goto lExit;
    }
    if (getDHMS(words[5], /*deflt=>*/ eDegrees,  &(sh->radecsweep.end.y))) {
      sureWrite(filedes, "IV,edec\n", 8);
      goto lExit;
    }
    if (getDuration(words[6], &(sh->radecsweep.dur))) {
      // xxx: should be positive, yes?
      sureWrite(filedes, "IV,dur\n", 7);
      goto lExit;
    }

    // Convert ra's from degrees (convenient) to hours (what Novas uses).
    sh->radecsweep.start.x = degToHrs(sh->radecsweep.start.x);
    sh->radecsweep.end.x = degToHrs(sh->radecsweep.end.x);

    // Pre-calculate values for 'antcon'.
    precalculateSweep(&(sh->radecsweep),
                      /*xacc=>*/ degToHrs(0.05),  // hours!
                      /*yacc=>*/ 0.05
                      );

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    // Signal 'antcon' by incrementing the update flag counter.
    sh->radecsweep.isActive++;

    goto lExit;
  }  // if "radecsweep"

  // LOG
  if ((strcmp(words[0], "log")==0) || (strcmp(words[0], "LOG")==0) ) {

    if (number_of_words < 2) {
      // Report back the current log file settings.
      if (sh->log.isActive) {
        rbytes += sprintf(replyBuffer, "OK,%.2f,%.2f,%s\n",
                          (sh->log.endTime - sh->utc),  // remaining
                          sh->log.incr,
                          sh->log.filename
                          );
      } else {
        rbytes += sprintf(replyBuffer, "OK,0.00\n");
      }

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    if (number_of_words < 3) {
      // 2 words, 1 parameter, check if the only parameter is "off".
      if ((strcmp(words[1], "off")==0) || (strcmp(words[1], "OFF")==0) ) {
        if (sh->log.isActive) {
          sh->log.isActive++;  // increment to > 1 to signal off
        }
        sureWrite(filedes, "OK\n", 3);
        goto lExit;
      } else {
        sureWrite(filedes, "IV\n", 3);
        goto lExit;
      }
    }
    // We have at least 2 words.

    // It can be new end time, check it.
    {
      double newend;

      if (getDateTime(words[1], &newend)) {
        sureWrite(filedes, "IV,end\n", 7);
        goto lExit;
      }

      // Just change it without special signalling, active or not.
      sh->log.endTime = newend;
      // (If active, changes end time, if not, the change doesn't matter.)
      // (Wants a file name to activate a new log file.)
    }
    if (number_of_words < 3) {
      sureWrite(filedes, "OK\n", 3);
      goto lExit;
    }

    // At least 3 words, 2 parameters.
    {
      // Check if the 2nd parameter is a new increment / log interval time.
      double newincr;

      if (getDuration(words[2], &newincr)) {
        sureWrite(filedes, "IV,inc\n", 7);
        goto lExit;
      }
      // Just change it without special signalling, active or not.
      sh->log.incr = newincr;
        // (If active, changes logging interval immediately,
        //  if not, the change doesn't matter.)
        // (Needs a file name to activate a new log file.)
    }
    if (number_of_words < 4) {
      sureWrite(filedes, "OK\n", 3);
      goto lExit;
    }

    // At least 4 words, 3 parameters: 3rd is new file name.
    if (sh->log.isActive) {
      // A log file is active and open, cannot change it.
      sureWrite(filedes, "AC,file\n", 8);
      goto lExit;
    }

    // Okay, no previous log file is active, test-open the new one.
    {
      FILE *f;

      if ((f = fopen(words[3], "at"))) {
        // Successfully opened.
        (void)fclose(f);
        strncpy(sh->log.filename, words[3], sizeof(sh->log.filename));
        rbytes += sprintf(replyBuffer, "OK\n");
      } else {
        // Failed (no need to close).
        rbytes += sprintf(replyBuffer, "OP,fopen(): %s (%d) '%s'\n",
                          strerror(errno),
                          errno,
                          words[3]
                          );
      }
    }

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    // Signal 'antcon' by incrementing the update flag counter.
    sh->log.isActive++;

    goto lExit;
  }  // if "log"

  // ONSOURCE
  if ((strcmp(words[0], "onsource")==0) || (strcmp(words[0], "ONSOURCE")==0) ) {
    // Report back the current position error situation.
    rbytes += sprintf(replyBuffer, "%s,%.6f,%.6f,%.6f\n",
                      ((sh->vectPosErr < sh->slewLimit) ? "OK" : "SL"),
                      sh->vectPosErr,
                      sh->posErr.az,
                      sh->posErr.el
                      );

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);
    goto lExit;
  }  // if "onsource"

  // SLEWLIMIT
  if ((strcmp(words[0], "slewlimit")==0) || (strcmp(words[0], "SLEWLIMIT")==0) ) {

    if (number_of_words < 2) {
      // Report back the current "slewing" limit.
      rbytes += sprintf(replyBuffer, "OK,%.6f\n",
                        sh->slewLimit
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    // At least 2 words, 1 parameter: new slew limit.
    if (getDHMS(words[1], /*deflt=>*/ eDegrees, &(sh->slewLimit))) {
      sureWrite(filedes, "IV\n", 3);
      goto lExit;
    }

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    //// Signal 'antcon' by incrementing the update flag counter.
    //sh->antcliUpdates++;

    goto lExit;
  }  // if "slewlimit"

  // CLIPMINAZLIMIT
  if ((strcmp(words[0], "clipminazlimit")==0) || (strcmp(words[0], "CLIPMINAZ")==0) ) {

    if (number_of_words < 2) {
      // Report back the current "tracking clip" limit.
      rbytes += sprintf(replyBuffer, "OK,%.6f\n",
                        sh->clipLimitMin.az
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    // At least 2 words, 1 parameter: new tracking clip limit.
    if (getDHMS(words[1], /*deflt=>*/ eDegrees, &(sh->clipLimitMin.az))) {
      sureWrite(filedes, "IV\n", 3);
      goto lExit;
    }

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    //// Signal 'antcon' by incrementing the update flag counter.
    //sh->antcliUpdates++;

    goto lExit;
  }  // if "clipxxxxxlimit"

  // CLIPMINELLIMIT
  if ((strcmp(words[0], "clipminellimit")==0) || (strcmp(words[0], "CLIPMINEL")==0) ) {

    if (number_of_words < 2) {
      // Report back the current "tracking clip" limit.
      rbytes += sprintf(replyBuffer, "OK,%.6f\n",
                        sh->clipLimitMin.el
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    // At least 2 words, 1 parameter: new tracking clip limit.
    if (getDHMS(words[1], /*deflt=>*/ eDegrees, &(sh->clipLimitMin.el))) {
      sureWrite(filedes, "IV\n", 3);
      goto lExit;
    }

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    //// Signal 'antcon' by incrementing the update flag counter.
    //sh->antcliUpdates++;

    goto lExit;
  }  // if "clipxxxxxlimit"

  // CLIPMAXAZLIMIT
  if ((strcmp(words[0], "clipmaxazlimit")==0) || (strcmp(words[0], "CLIPMAXAZ")==0) ) {

    if (number_of_words < 2) {
      // Report back the current "tracking clip" limit.
      rbytes += sprintf(replyBuffer, "OK,%.6f\n",
                        sh->clipLimitMax.az
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    // At least 2 words, 1 parameter: new tracking clip limit.
    if (getDHMS(words[1], /*deflt=>*/ eDegrees, &(sh->clipLimitMax.az))) {
      sureWrite(filedes, "IV\n", 3);
      goto lExit;
    }

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    //// Signal 'antcon' by incrementing the update flag counter.
    //sh->antcliUpdates++;

    goto lExit;
  }  // if "clipxxxxxlimit"

  // CLIPMAXELLIMIT
  if ((strcmp(words[0], "clipmaxellimit")==0) || (strcmp(words[0], "CLIPMAXEL")==0) ) {

    if (number_of_words < 2) {
      // Report back the current "tracking clip" limit.
      rbytes += sprintf(replyBuffer, "OK,%.6f\n",
                        sh->clipLimitMax.el
                        );

      /* Send network reply back to the client. */
      sureWrite(filedes, replyBuffer, rbytes);
      goto lExit;
    }

    // At least 2 words, 1 parameter: new tracking clip limit.
    if (getDHMS(words[1], /*deflt=>*/ eDegrees, &(sh->clipLimitMax.el))) {
      sureWrite(filedes, "IV\n", 3);
      goto lExit;
    }

    /* Build network reply buffer. */
    // rbytes += sprintf(&(replyBuffer[2]), ",%+g,%+g\n", sh->incr.az, sh->incr.el);
    rbytes += sprintf(replyBuffer, "OK\n");

    /* Send network reply back to the client. */
    sureWrite(filedes, replyBuffer, rbytes);

    //// Signal 'antcon' by incrementing the update flag counter.
    //sh->antcliUpdates++;

    goto lExit;
  }  // if "clipxxxxxlimit"

  // xxx: "terminate" (for compatibility with the FS).

  if (strcmp(words[0], "GET")==0) {
    // xxx: Reply with real HTTP message. */
  }  // if "GET"

  if (strcmp(words[0], "reload")==0) {
    // xxx: Reload channel name & settings file.
    // Run configuration checks;
    // if no responses, generate error message
    // with 'sl.whatPathname()' telling the serial port used.
  }  // if "reload"

  // Nothing matched, let the inherited method do the work.
  retCode = cMessageProcessor::process_message_buffer(filedes,
                                                      buffer,
                                                      nbytes);
  goto lExit;

lExit:
  if (running) {
    free(running);
  }
  return retCode;
}  // process_message_buffer

void
cMPAnt::doBackground(void) {
#if 0
  time_t now;
  char replyBuffer[256];
  int i = this->nextChannel;

  now = time(NULL);
  if ( (now - ct[i].lastLogged) > ct[i].interval ) {
    // Log it.
    printf("%s\t%s\t%+g\n", getUTC(NULL), ct[i].name1, value);
    
    // Make sure the line gets immediately to stdout
    // (where 'arranger.pl' takes it and writes it into a file).
    fflush(stdout);

    ct[i].lastLogged = now;  // successful or not, for now...
  }  // if expired

  if (i < (ctSize-1)) {
    this->nextChannel++;
  } else {
    this->nextChannel = 0;
  }
#endif

  // Do default background processing.
  cMessageProcessor::doBackground();
}  // doBackground



// ----------------------------------------------------------------------
int
main (int argc, char **argv)
{
  cSocketServer ss;
  cMPAnt mp;

  int option_char;

  // Startup version message.
  fprintf (stderr, MY_NAME "$Id: antcli.cc,v 1.31 2005/10/12 06:30:53 lindi Exp $\n");

  // Configuration file / command line option defaults.
  // Command line wins, default conf file is next,
  // and these hard-coded values are the last resort.
  int my_port_number = 5555;

  // unpriviledged user name
  char *my_user_name = NULL;

  // pid file name
  char *my_pid_file = NULL;

  // Process default conf file.
  // xxx process_conf(/*file=>*/ "/etc/dammer.conf" );

  // Invokes member function `int operator ()(void);'
  while ((option_char = getopt (argc, argv, "c:p:U:P:")) != EOF)
    switch (option_char) {
       case 'c': {
         assert( optarg != NULL );
	 // xxx process_conf(/*file=>*/getopt.optarg);
	 break;
       }
       case 'p': {
	 my_port_number = atoi (optarg);
	 break;
       }
       case 'U': {
	assert ((my_user_name = strdup(optarg)) != NULL);
	// TODO: figure out where priviledges can be dropped
	break;
       }
       case 'P': {
	assert ((my_pid_file = strdup(optarg)) != NULL);
	break;
       }
       case ':':  // missing parameter after a option letter
       case '?': {
	 fprintf (stderr,
		  "usage: %s [-p<port>] [-U<user>] [-P<pidfile>]\n", argv[0]);
	 exit(EXIT_FAILURE);
       }
    }  // switch

  // Write pid file
  if (my_pid_file) {
	  FILE *fp;
	  assert((fp = fopen(my_pid_file, "w")) != NULL);
	  fprintf(fp, "%d\n", getpid());
	  fclose(fp);
  }

  // Shared memory unique identifier (key) and the returned reference id.
  key_t shKey;
  int shId = -1;

  // Find and attach to 'antcon--antcli--xantcon' shared memory.
  shKey = fourCharLong('a','n','t','c');
  // Don't create, must exist (created by 'antcon')
  shId = shmget(shKey, sizeof(tSh), 0777);
  if (shId == -1) {
    perror(MY_NAME "shmget(): could not find 'antcon' shared memory");
    exit(EXIT_FAILURE);
  }
  assert( (sh = (ptSh)shmat(shId, NULL, 0)) != (ptSh)-1 );
  // Initialize 'antcli' portion of shared memory.
  // xxx: init like sh->target.az = sh->target.el = 0.0;

  // Prime the message processor.
  mp.init();

  // ------------------------------------------------------------
  // Start to listen and serve socket/port forever until killed.
  // Command lines from network are interpreted
  // and the values are written to shared memory for 'antcon'.
  ss.serve_socket (my_port_number, mp);
  // ------------------------------------------------------------

  // xxx: Is the following cleanup code ever executed?
  mp.cleanup();

#if 0
  printf("shKey=0x%08X, shId=%d, sh=%08X\n",
         shKey, shId, (unsigned int)sh);
#endif
  if ((shId != -1) && (sh != (ptSh)-1) && (sh != NULL)) {
    // It is not our business to remove the seqment,
    // 'antcon' takes care of managing removal.
    // assert( shmctl(shId, IPC_RMID, NULL) == 0 );
    assert( shmdt(sh) == 0 );
  }  // if shared memory was attached to

  return (EXIT_SUCCESS);
}  // main
