/* sh.h -- The complete and common shared struct */
/* across ant/dos/up/ant.cpp -- antcond (==antcon+antcli) -- xantcon. */

#include "registration.h"
RCSID(sh_cc, "$Id: sh.h,v 1.14 2002/04/05 09:55:14 amn Exp $");

/*
 * $Log: sh.h,v $
 * Revision 1.14  2002/04/05 09:55:14  amn
 * Added clip[min|max][az|el]limits.
 *
 * Revision 1.13  2001/02/12 11:56:58  amn
 * Added waiting for antcon PP_ updates before antcli replying.
 *
 * Revision 1.12  2001/01/16 09:41:37  amn
 * Added antcli --> antcon --> sh updates of af/met/cPointing.cc parameters.
 *
 * Revision 1.11  2000/09/20 10:15:06  amn
 * Added min/ave/max azel error to LOG=.
 *
 * Revision 1.10  2000/07/28 09:28:20  amn
 * Changed to using VC 'open'(1) and added logging 'log=' to antcli.
 *
 * Revision 1.9  2000/07/25 12:27:31  amn
 * azelsweep works about ok.
 *
 * Revision 1.8  2000/07/24 09:57:14  amn
 * Added 'updateSweep()' and reading in azelsweep values in 'antcli'.
 *
 * Revision 1.7  2000/06/20 13:42:35  amn
 * Added 'tSweep' structs to sh.h for both ra-dec and az-el.
 * Still missing 'updateSweep()' and antcli:'readSweep()'.
 *
 * Revision 1.6  2000/04/06 07:47:22  amn
 * Added time fields (JD, MJD etc).
 *
 * Revision 1.5  2000/02/16 12:01:21  amn
 * Closed an unterminated comment.
 *
 * Revision 1.4  2000/02/16 11:59:13  amn
 * Added many commands to 'antcli'.
 *
 * Revision 1.3  2000/02/06 17:56:36  amn
 * Added the basic commands what the FS needs for pointing measurements.
 *
 * Revision 1.2  2000/02/04 13:55:33  amn
 * Added 'source=' command sh-> variables.
 *
 * Revision 1.1  2000/02/02 15:25:18  amn
 * Initial revision.
 * */

/* ant.cpp <--> antcond is transmitted field-by-field using COM2 serial, */
/* antcond <--> xantcon uses shared memory and the same struct layout. */
/* Internally antcond has 'antcon <--> antcli'. */

/* Per motor / FET board settings and measurements. */
typedef struct sMotor {
  int pwm;  /* as in cPWMISA, -1023..1023 */
  int antiph;  /* as in cPWMISA, -1023..1023, only 16 real steps */
  unsigned int qdr;  /* latest QDR tacho value */
  double qdrVel;  /* calculated QDR velocity, counts/s */
  double qdrAcc;  /* acceleration, counts/s */
  double vsupply;  /* power supply of this motor */
  double icurr;  /* latest current measurement of this motor */
  double temp;  /* temperature of FET board */
} tMotor, *ptMotor;

/* Ra-dec or az-el as well. */
typedef struct sAnyCoord {
  double x;
  double y;
} tAnyCoord, *ptAnyCoord;

/* The sweep struct.  Defines start and end offsets plus durations. */
typedef struct sSweep {
  int isActive;  /* ==0, skipped altogether */
  char status;  /* descriptive status letter, which phase is going */

  /* From antcli: */
  /* (If sweep is active and new values are coming from antcli, */
  /*  record the old 'remaining' time here:) */
  double prevSweepInterruptTime;
  /*  time of interrupt of previous scan, >dur means completed.) */
  double startTime;  /* in Unix (fractional) seconds, representing UTC */
  tAnyCoord start;  /* start and end offsets, /wo accel/decel */
  tAnyCoord end;
  double dur;  /* sweep duration in (fractional) seconds */
  double accdur;  /* acceleration seconds (before startTime) */
  double decdur;  /* deceleration seconds (after startTime+dur) */
  double keepdur;  /* keep final position (seconds); after it back to zero */

  /* Precalculated convenience variables. */
  /* ('antcli' recalculates these every time they are changed from cli.) */
  tAnyCoord accStart;  /* start accelerating here */
  tAnyCoord acca;  /* acceleration before sweep */
  tAnyCoord v;  /* calculated constant speed during sweep */
  tAnyCoord deca;  /* deceleration after sweep */
  tAnyCoord decEnd;  /* stop decelerating here */

  /* Run-time variables. */
  double t;  /* sweep time, in (fractional) seconds, 0.0==start point */
  tAnyCoord o;  /* calculated offsets */
  /* Sweep phasing: */
  /* - goto start-acc point: t<(-accdur) */
  /* - accelerate: t<0.0 */
  /* - constant velocity: t<dur */
  tCoord startError;  /* store azel(?) error when starting sweep => detect too "tight" sweeps */
  /* - decelerate: t<(dur+decdur) */
  /* - keep end point: t<(dur+decdur+keepdur) */
  /* - deactivate, offsets to zero: t >= (dur+decdur+keepdur) */
} tSweep, *ptSweep;


/* ---------- */
/* The struct :-) */

typedef struct sSh {
  /* Current time, UTC in Unix 2xlong format, */
  /* UTC in Unix double (fractional seconds) format, */
  /* and m/usecs since program start. */
  tTime coord_calc_utc_tv;
  double utc;
  unsigned long msecs;
  double utcJD;  /* UTC in Julian days */
  double utcMJD;  /* UTC in Modified Julian days; for IERS */
  int dayOfYear;  /* 1..366, as in Astronomical Almanac */
  double tdtJD;  /* TDT time (terrestial dynamic time, "observation time") */
  double tdbJD;  /* TDB, terrestial barycentric dynamical time, for eq_eq */
  double eq_eq;  /* equation of the equinoxes */
  double gmst;  /* Greenwich Mean Sidereal Time */
  double gast;  /* Apparent (local) Sidereal Time */

  /* ----- Pedestal PC hardware-related: */
  /* Interlock chain bits. */
  tInterlockBits interlock;

  /* Angle encoders, all in degrees, az=-180..180 "old" convention. */
  tCoord abs;
  tCoord incr;
  tCoord tilt;

  /* Per motor / FET board settings and measurements. */
  union {
    tMotor motors[5];
    struct {
      tMotor az1;
      tMotor az2;
      tMotor el1;
      tMotor el2;
      tMotor subr;
    } motor;
  };

  /* Brakes (brake Ibrake currents). */
  struct {
    double az;
    double el;
  } ibrake;

  /* Supply voltages from power supplies. */
  struct {
    double v5;
    double v12;
    double v24;
    double vn15;
    double v15;
    double vgnd;
  } volt;

  /* Temperatures. */
  struct {
    double tilt;
    double racklow;
    double rackmiddle;
    double rackpc;
  } temp;

  /* ----- ant/dos/up/ant.cpp -related: */
  /* Azimuth system used, 0.0 for "traditional" Metsahovi -180..180 */
  /* and 180.0 for international 0..360 azimuth. */
  int azel;  /* 1== Az first, El then; 0==El first, Az then */
  double azsys;  /* 0.0 or 180.0 */
  /* Add 'azsys' to all internal azimuths before displaying to the user, */
  /* and subtract it before accepting azimuths from the user. */

  /* Current antenna position (==incr), angular velocity, and acceleration. */
  tCoord curr;
  tCoord currSp;
  tCoord currAcc;

  /* The latest uplink message time from downstairs 'antcond', */
  /* i.e. xntpd time. */
  tTime xntpdTime;

  /* Current "momentary" (==for this loop run) target position. */
  tCoord goalPos;
  tCoord goalVel;
  tCoord goalAcc;

  /* Manual/keyed in main antenna angular speed. */
  tCoord goalSpeed;

  /* ----- ant/antcon/antcon.cc -related: */
  /* PWM settings from "service" mode panel knobs. */
  double knobMot1pwm, knobMot2pwm;

  /* Target is what we ask upstairs to go to. */
  tCoord target;
  /* Temporary user-settable az/el limits where tracking is "clipped". */
  /* (Cannot override hard-coded safety limits.) */
  tCoord clipLimitMin;
  tCoord clipLimitMax;
  /* vPosErr is how much we are off from vTarget. */
  tCoord posErr;
  /* Is the "Pythagoran" sum of az/el errors, "error distance". */
  double vectPosErr;
  /* 'antcli' slewing limit, vector error less than this */
  /* reports back 'OK' to 'onsource'. */
  double slewLimit;
  /* Panel offsets are added to what VAX/tracking formulae tells us, */
  /* and subtracted before we tell VAX/other computers back where we are. */
  tCoord panelOffset;  /* 'antcon' panel offsets */
  /* We keep our own target (calculated from source ra/dec using Novas lib) */
  /* separate from VAX target, fixed az-el position, */
  /* and manual panel position (all of these /wo offsets). */
  tCoord myTarget;
  tCoord VAXTarget;
  tCoord myFixedPos;
  tCoord manualTarget;
  /* Manual panel speeds. */
  tCoord manualSpeed;
  /* The final position where antenna stopped while slowing down to standby. */
  tCoord standbyPos;

  /* xxx: any need for "pure" az/el coordinates (/wo pointing model)? */
  /* (remember FS 'source=azeluncr,...') */

  /* 'antcli'-maintained variables coming from command line. */
  int antcliUpdates;  /* incremented by 'antcli', decr by 'antcon' */
  /* --'standby', 'azel', or 'source='? */
  int antcliMode;
#define ACM_standby 1
#define ACM_azel 2
#define ACM_source 3
#define ACM_VAX_track 4
  int getPP;  /* incremented by 'antcli', decr by 'antcon' */
  int updatePP;  /* incremented by 'antcli', decr by 'antcon' */
  char ppname[50];  /* a single pointing parameter name & value */
  int foundPP;  /* set to ==0 if named parameter not found by 'antcon' */
  double ppvalue;

  /* --'source=' */
  int isPlanet;  /* 1==is a solar system planet */
  /* if 'isPlanet' ==1, then a planet object in below */
  int sourcePlanetType;
    /* 0 ... major planet, Sun, or Moon */
    /* 1 ... minor planet */
  int sourcePlanetNumber;
  /* For 'type' = 0: Mercury = 1, ..., Pluto = 9, */
  /*                 Sun = 10, Moon = 11  ('antcli' maps these) */
  /* For 'type' = 1: minor planet number */
  /* char sourcePlanetName[100];  merged to 'sourceName' */
  /* if 'isPlanet' ==0, then a "regular" object in below */
  char sourceName[51];  /* object/target name (as in Novas 2.0) */
  double sourceRa;  /* its right ascension, in (decimal) hours */
  double sourceDec;  /* declination, in (decimal) degrees */
  double sourceEpoch;  /* epoch of the previous, in either (decimal) years */
                       /* or in Julian days; typically 2000.0 (or 1950.0) */
  double sourceRadialvel;  /* radial velocity in m/s, used for Doppler calcs */
  double sourcePromora;
  double sourcePromodec;
  double sourceParallax;
  /* --'source=' end */

  /* 'source='-related additional variables changed by 'antcli'. */
  /* Current apparent (not topocentric) ra/dec without radec offsets. */
  /* (Updates for both extragalactic and for planets as UTC elapses.) */
  double appRa;
  double appDec;

  /* Ra/dec offsets. */
  double raOffset;
  double decOffset;
  /* Antenna coordinate offsets. */
  tCoord myOffset;  /* 'antcli' computer (FS/meacon) offsets */
  /* xxx: x-y-offsets?  (odd antenna mount system in FS) */

  /* Sweep variables, for both azelsweep and radecsweep. */
  tSweep radecsweep;
  tSweep azelsweep;

  /* antcli-started antcon performance logging. */
  struct {
    int isActive;  /* ==0 => no log time checking needs to be done */
    char filename[256];
    FILE *filep;  /* NULL => no file open */
    double nextTime;  /* in Unix UTC fractional seconds; */
    /* the next time log line should be written (initially the start time) */
    double incr;  /* in fractional seconds */
    double endTime;
    /* Minimum and maximum errors for both az and el during incr seconds. */
    tCoord minErr;  /* typically - */
    tCoord sumErr;  /* for average */
    int nErr;  /* for average */
    tCoord maxErr;  /* typically + */
  } log;

} tSh, *ptSh;
