#include "registration.h"
RCSID(c8255_cpp, "$Id: c8255.cpp,v 1.8 1999/10/12 07:49:16 amn Exp $");
/* ------------------------------------------------------------------ */
/* Represents a 8255 parallel TTL digital I/O, one 50-pin connector
   (3x8 bit ports A, B, C) per one 8255 chip. */

// $Log: c8255.cpp,v $
// Revision 1.8  1999/10/12 07:49:16  amn
// Added log.
//

class c8255 {
private:
  int baseAddr;
  int conn;  /* connector on ET-1722B board, 0..5 */
  int addressing;
  int relAddrIndex;  /* address of 8255 to be used, calc from 'baseAddr' and 'conn' */
    /* or in case of separate address register on card, its base value. */
protected:
  virtual void out(int offset, unsigned char value);
  virtual unsigned char in(int offset);
public:
  virtual void init(
	int baseAddr,
	int conn,
    int addressing  /* 0==four separate I/O; 1== only two regs, 0=addr, 1=data */
  );
  virtual void cleanup(void);
/* Directions for the following. */
#define IN 1
#define OUT 0
  virtual void setPortABCMode(
    int portA,
    int portB,
    int portCUpper,
    int portCLower
  );
};  /* c8255 */

void c8255::out(int offset, unsigned char value) {
  if (this->addressing) {  /* separate address register */
    outportb(this->baseAddr+0, (this->relAddrIndex + offset) );
    outportb(this->baseAddr+1, value);
  } else {  /* normal four I/O regs per chip */
    outportb(this->relAddrIndex + offset, value);
  }
}

unsigned char c8255::in(int offset) {
  if (this->addressing) {  /* separate address register */
    outportb(this->baseAddr+0, (this->relAddrIndex + offset) );
    return (inportb(this->baseAddr+1));
  } else {  /* normal four I/O regs per chip */
    return (inportb(this->relAddrIndex + offset));
  }
}

/* Initialize one 50-pin Opto-22 connector for one 8255 chip. */
void c8255::init(
  int baseAddr,  /* of ET1722B card (or any 8255-compatible parallel) */
  int conn,  /* 0..5 */
  int addressing
) {
  this->baseAddr = baseAddr;
  this->conn = conn;
  this->addressing = addressing;

  /* If addr.reg/data.reg mode... */
  if (addressing) {
    /* ...store connector offset: */
    this->relAddrIndex = (conn<<2);
#if LINUX_IOPERM
    if (ioperm(
      /*from address=>*/         baseAddr,
      /*number of addresses=>*/  2,
      /*turn on=>*/              1  /* 0 to turn off */
    ) < 0) {
      fprintf(stderr, MYNAME "access to I/O port 0x%x denied\n", baseAddr);
      exit(EXIT_FAILURE);
    }
#endif
  } else {
    /* ...store relative address of 8255 base (+0..3). */
    this->relAddrIndex = baseAddr + (conn<<2);
#if LINUX_IOPERM
    if (ioperm(
      /*from address=>*/         this->relAddrIndex,
      /*number of addresses=>*/  4,
      /*turn on=>*/              1  /* 0 to turn off */
    ) < 0) {
      fprintf(stderr, MYNAME "access to I/O port 0x%x denied\n", this->relAddrIndex);
      exit(EXIT_FAILURE);
    }
#endif
  }

  /* Set to input mode by default. */
  this->setPortABCMode(
	/*A=>*/ IN, /*B=>*/ IN, /*Cupper=>*/ IN, /*Clower=>*/ IN
  );
}

void c8255::cleanup() {
  this->setPortABCMode(
	/*A=>*/ IN, /*B=>*/ IN, /*Cupper=>*/ IN, /*Clower=>*/ IN
  );
}

/* Set ports of one 50-pin connector of any 8255 PIO to IN or OUT. */
void c8255::setPortABCMode(
  int portA,
  int portB,
  int portCUpper,
  int portCLower
) {
  /* Control word layout: D7==always 1, D4==port A direction (0==out),
	 D3==portCUpper, D2==always 0, D1==portB, D0==portCLower,
	 D6, D5==always 0. */
  this->out(
	/*offset=>*/ 3,
	0x80
	  | ((portA & 0x01)<<4)
	  | ((portCUpper & 0x01)<<3)
	  | ((portB & 0x01)<<1)
	  | (portCLower & 0x01)
  );
}


/* ------------------------------------------------------------------ */
/* All three subports are used together as a 24-bit output. */
class c24Out : public c8255 {
public:
  virtual void init(
	int baseAddr,
	int conn,
        int addressing
  );
  virtual void set(unsigned long int value);
  virtual void cleanup(void);
};

void c24Out::init(
  int baseAddr,  /* of ET1722B card (or any 8255-compatible parallel) */
  int conn,
  int addressing
) {
  c8255::init(baseAddr, conn, addressing);

  this->setPortABCMode(
	/*A=>*/ OUT, /*B=>*/ OUT, /*Cupper=>*/ OUT, /*Clower=>*/ OUT
  );
}

void c24Out::cleanup() {
  c8255::cleanup();
}

void c24Out::set(unsigned long int value) {
  this->out(0,  value      & 0x000000FF);
  this->out(1, (value>>8)  & 0x000000FF);
  this->out(2, (value>>16) & 0x000000FF);
}


/* ------------------------------------------------------------------ */
/* All three subports are used together as a 24-bit input. */
class c24In : public c8255 {
public:
  virtual void init(
	int baseAddr,
	int conn,
    int addressing
  );
  virtual unsigned long int get(void);
  virtual void cleanup(void);
};

void c24In::init(
  int baseAddr,  /* of ET1722B card (or any 8255-compatible parallel) */
  int conn,
  int addressing
) {
  c8255::init(baseAddr, conn, addressing);

  this->setPortABCMode(
	/*A=>*/ IN, /*B=>*/ IN, /*Cupper=>*/ IN, /*Clower=>*/ IN
  );
}

void c24In::cleanup() {
  c8255::cleanup();
}

unsigned long int
c24In::get(void) {
  return (((unsigned long int)this->in(2)) << 16)
       | (((unsigned long int)this->in(1)) <<  8)
       | (unsigned long int)this->in(0);
}  // get
