#include "registration.h"
RCSID(cet28xx_cpp, "$Id: cet28xx.cpp,v 1.4 1999/10/28 11:01:33 amn Exp $");
/* ------------------------------------------------------------------ */
/* Represents a ET-2816 (16-bit) or ET-xxx (12-bit) A/D converter board. */

// $Log: cet28xx.cpp,v $
// Revision 1.4  1999/10/28 11:01:33  amn
// Tried shortening channel/gain change delay from 1ms to 10usec.
//
// Revision 1.3  1999/10/12 07:51:18  amn
// Added registration and log.
//

#define DIFF 1
#define SE 0

class cET28xx : public cSensor {
private:
  int baseAddr;
  int boardModel;
  int numOfBitsMask;  /* all valid (12/16) bits set to 1, all others 0 */
  double binToVoltsCoeff;
  int loopLimit;
  int maxLoops;
  int currentChannel;
  double binToVoltsCoeffWithGain;

  void delay5us(void);

public:
  virtual void init(
	int baseAddr,
#define ET2816 (0)
#define ACL8113 (1)
    int boardModel = ET2816,
	int numOfBitsMask = 0xffff,
	int loopLimit =
#if DEBUG_REPEATS
	  10  /* ET-2816: 10us conversion time... */
#else
	  15
#endif
  );
  virtual void setChannel(
	int channel,
	int mode,  /* SE/DIFF */
	int gain  /* 1,2,4,8 == +/- 10, 5, 2.5, 1.25V */
  );
  virtual int getBin(void);
  virtual long int getRaw(void);
  virtual double getUncalibrated(void);
  virtual void cleanup(void);
};  /* cET28xx */

void cET28xx::init(
  int baseAddr,
  int boardModel,
  int numOfBitsMask,
  int loopLimit
) {
  cSensor::init();

  this->baseAddr = baseAddr;
  this->boardModel = boardModel;
  this->numOfBitsMask = numOfBitsMask;
  this->binToVoltsCoeff = 10.0 / (double)( (((unsigned int)numOfBitsMask)>>1) + 1 );
  this->loopLimit = loopLimit;
  this->maxLoops = 0;
  this->currentChannel = -1;
  this->binToVoltsCoeffWithGain = binToVoltsCoeff /* * 1.0 */;

  /* We will not touch baseAddr + 0..3, timer chip 8254. */

  if (boardModel == ET2816) {
    /* Four operation modes are available: */
    /* 0x00 == only external trigger pulse will start conversion */
    /* 0x01 == software will start and poll for completion */
    /* 0x02 == timer pacer will start and transfer via DMA */
    /* 0x0x == timer pacer will start and interrupt will occur after completion */
    /* We will use the software only mode. */
    outportb(baseAddr + 11, 0x01);

    /* Clear interrupt status flip-flop just in case. */
    outportb(baseAddr + 8, 0x00);

    /* Set digital I/O output to all zeroes. */
    outportb(baseAddr + 13, 0x00);
    outportb(baseAddr + 14, 0x00);

    /* Set both D/A channels to 0V. */
    outportb(baseAddr + 4, 0x00);  /* ch 1 LSB */
    outportb(baseAddr + 5, 0x00);  /* ch 1 MSB */
    outportb(baseAddr + 6, 0x00);  /* ch 2 LSB */
    outportb(baseAddr + 7, 0x00);  /* ch 2 MSB */
	  /* Writing MSB trigs LSB into D/A converter. */
  }  // if board is ET2816

  this->setChannel(0, SE, /*gain=>*/ 1);
}

void cET28xx::cleanup(void) {
#if DEBUG_REPEATS
  printf("cET28xx: max number of DRDY bit wait inportb() calls = %d\n",
	this->maxLoops
  );
#endif
  cSensor::cleanup();
}

void cET28xx::delay5us(void) {
  outportb(this->baseAddr, 0);
  outportb(this->baseAddr, 0);
  outportb(this->baseAddr, 0);
  outportb(this->baseAddr, 0);
  outportb(this->baseAddr, 0);
}  // delay5us

void cET28xx::setChannel(int channel, int mode /*SE/DIFF*/, int gain) {
  unsigned char temp;

  if (this->boardModel == ACL8113) {
    assert(mode == SE);
    channel &= 0x001F;  // only 5 bits, 0..31
    temp = channel;
  } else {
    // ET2816 board.
    channel &= 0x000F;
    if (mode == DIFF) {
	  temp = 0x30 | (channel & 0x07);
    } else /* if (mode == SE) */ {
	  temp = ((channel & 0x08) ? 0x20 : 0x10) | channel;
    }
  }  // else ET2816 board
  outportb(baseAddr + 10, temp);
#if DEBUG_CHANGE_CHANNEL
  {
    int x = wherex();
    int y = wherey();

    gotoxy(1, 10);
    printf("ET2816/ACL8113 channel byte 0x%03X = %02X", baseAddr+10, temp);
    gotoxy(x, y);
  }
#endif
  this->currentChannel = channel;

  /* Calculate a multiply coefficient for getting real volts from binary. */
  this->binToVoltsCoeffWithGain = 1.0 / (double)gain * this->binToVoltsCoeff;

  /* Map gain coefficient to bit value 1->0, 2->1, 4->2, 8->3. */
  {
	int temp;

	for (temp=-1; gain; gain>>=1, temp++) ;
	outportb(baseAddr + 9, (temp & 0x0f));
	  /* ET-2816 documents only two LSBits, but other similar 28xx boards
		 probably use four bits for even larger gains. */
#if DEBUG_CHANGE_CHANNEL
    {
      int x = wherex();
      int y = wherey();

      gotoxy(1, 11);
      printf("ET2816/ACL8113 gain select  0x%03X = %02X", baseAddr+9, temp);
      while (!kbhit()) ;
      (void)getch();
      gotoxy(x, y);
    }
#endif
  }
#if 0
  /* Take one sample to get enough delay... */
  (void)this->getBin();
    /* Is not enough, too bad... */
#endif
  if (this->boardModel == ACL8113) {
  this->delay5us();  // from Advantech PCL-813 manual/DEMO01.C:
  // "Important!!!  For channel setting."
  this->delay5us();  // ...another was there... ?! :-)
  } else {
    // ET2816 board.
    delay(1);  /* xxx: is 1ms enough?  Seems to be for ET2816.*/
  }
}  // setChannel

int cET28xx::getBin(void) {
  int n;
  unsigned char drdybit;
  int temp;
  unsigned char drdys[100];

#if NO_ET28xx_PRESENT
  return (0);
#endif

  /* Trigger next A/D conversion (by writing anything to address 12). */
  outportb(baseAddr + 12, 0x00);
  this->delay5us();  // From PCL-813: Delay 10us waiting for photocoupler
  this->delay5us();  // to transfer data...
  this->delay5us();  // (These two had no comments in PCL-813 code...)
  this->delay5us();

  this->delay5us();
  this->delay5us();
  this->delay5us();
  this->delay5us();

  this->delay5us();  // 8*5us seems to be enough, 7*5us is not,
                     // so having 9*5us seems to be on the safe side.

  /* Wait for completion or max number of loops. */
  /* Bit 5 of byte 'baseAddr + 8' goes to zero when ready. */
  n = 0;
  do {
	n++;
    if (this->boardModel == ET2816) {
      drdybit = inportb(baseAddr + 8) & 0x20/*DRDY*/;
    } else if (this->boardModel == ACL8113) {
      drdys[n] = inport(baseAddr + 5);
      drdybit = drdys[n] & 0x10/*DRDY*/;
      // No, no: drdybit = !(drdybit);  // fix-fix against doc/manual
      // ACL8113 DRDY-bit seems to be stuck == 0.
    } else {
      assert( ("cET28xx::getBin: unknown board model", 0) );
    }
#if 0
// Did not help on 16.02.1999 with ACL-8113 with 'n < 1'
// still random 0x0000, 0x0001 and like values...
// Forcing 60 loops, i.e. over the max A/D conversion time helped.
    if (n < 60) {  // 60*0.8==48us --> 21kHz << 25kHz
      // Then "fast CPU" fix: DRDY doesn't have enough time to go busy
      // after software trigger before the first query, so mask first
      // query to "busy":
      drdybit = 1;
    }
#endif
  } while ((drdybit) && (n < this->loopLimit));
	/* One loop should take about 800ns (one inportb()). */
  if (n > this->maxLoops) this->maxLoops = n;
#define DEBUG_DRDY 0
#if DEBUG_DRDY
  for (n=1; n<100; n++) {
    printf("%02X ", drdys[n]);
  }
  while (!kbhit()) ;
  (void)getch();
#endif
  if (n >= this->loopLimit) {
	/* xxx: What to return as error value? */
    /* xxx: This is about the only place to detect a faulty or missing board. */
#if DEBUG_REPEATS
	{
	  int x = wherex();
	  int y = wherey();

	  gotoxy(1, 8);
	  printf("A/D loops >= %d", n);
      sound(1000); delay(100); nosound();
      gotoxy(x, y);
    }
#endif
  }

  /* Get A/D converted value. */
  if (this->boardModel == ET2816) {
    temp = ((inportb(baseAddr + 5))<<8)  /* MSB */
            | inportb(baseAddr + 4);  /* LSB * /
            /* No need to mask 16-bit value, 'temp' is 16 bits. */
  } else if (this->boardModel == ACL8113) {
    temp = (((inportb(baseAddr + 5))<<8)  /* MSB */
            | inportb(baseAddr + 4)  /* LSB */
           ) & this->numOfBitsMask;
    // Now we have 0..4095 which needs to be mapped to -2047.5..+2047.5...
    temp -= 2048;
  } else {
    assert( ("cET28xx::getBin: unknown board model", 0) );
  }
  return (temp);
}

long int cET28xx::getRaw(void) {
  return (this->getBin());
}

double cET28xx::getUncalibrated(void) {
  return ( (double)this->getBin() * this->binToVoltsCoeffWithGain );
}
