#include "registration.h"
RCSID(cik121_cpp, "$Id: cik121.cpp,v 1.5 1999/10/14 12:48:29 amn Exp $");
/* -------------------------------------------------------------------- */
/* Heidenhain IK121 interface card. */

// $Log: cik121.cpp,v $
// Revision 1.5  1999/10/14 12:48:29  amn
// Added IK121 fault status detection; both down and uplinks.
//
// Revision 1.4  1999/10/13 12:15:40  amn
// Added routine for "stopped" and "searching for refmark" state queries.
//
// Revision 1.3  1999/10/12 07:52:04  amn
// Added registration and log.
//

class cIK121 /* xxx: will be ' : cSensor' */ {
private:
  unsigned int baseAddr;
  int addrReg;
    /* On-board register for internal chip address bus bits B4, B3, B2.
       B1 and B0 come directly from system bus.
       This is a cache-copy of the latest value written to the board. */
  int nextIOAddr;  /* precalculated I/O address for next outp/inp */

  void setAddr(int axis, int reg);
  void out(int axis, int reg, unsigned int value);
  unsigned int in(int axis, int reg);

  // Encoder signal amplitude bits, current and minimum.
  // Used for fault detection, both axen.
  int currAmpl[2];
  int minAmpl[2];

public:
  virtual void init(int baseAddr);
  virtual void cleanup(void);
  virtual void startCountingFromZeroAtRefMark(void);
  virtual int isSearchingForRefMark(int axis);
  virtual int isStopped(int axis);
  virtual char *amplStatus(int axis);
  virtual void resetAmplStatus(int axis);
  virtual void latchAtNextRefMark(void);
  virtual long int getLatchedRefMark(int axis);
  virtual void trigger(void);
private:
  int waitingForValues;
  long int latestValues[2];
public:
  virtual int isReady(void);
  virtual long int get(int axis);
};  /* cIK121 */

/* Extract bits B4, B3, and B2 of the chip register address */
/* and make sure that they are downloaded to the on-board address register. */
void cIK121::setAddr(int axis, int reg) {
  unsigned int temp = (reg & 0x1f) >> 2;

  if (temp != this->addrReg) {
    outport(this->baseAddr | 0x08, temp);
    this->addrReg = temp;
  }

  /* Calculate the I/O address for next outpw/inpw. */
  this->nextIOAddr = this->baseAddr | (axis ? 0x0004 : 0) | (reg & 0x03);
}

void cIK121::out(int axis, int reg, unsigned int value) {
  this->setAddr(axis, reg);
  outport(this->nextIOAddr, value);
}

unsigned int cIK121::in(int axis, int reg) {
  this->setAddr(axis, reg);
  return (inport(this->nextIOAddr));
}

void cIK121::init(int baseAddr) {
  this->baseAddr = baseAddr;
  this->addrReg = -1;  /* not set */
  this->waitingForValues = 0;
  /* xxx: detection, reg 0x1c has 0x08xx */
  out(0, 0x14, 0x0000);  /* axis 0, disable all interrupts */
  out(1, 0x14, 0x0000);  /* axis 1, disable all interrupts */
  out(0, 0x0c, 0xC001);  /* axis 0, interpolation mode, 32-bit counting */
  out(1, 0x0c, 0xC001);  /* axis 1, interpolation mode, 32-bit counting */
    /* Bits D15 == D14 == 1 mean that whenever reg 0x10 requests
       a counter value to be latched at ref mark, this latching
       will occur in data register 1 and not 0, the normal reg. */
  out(0, 0x0e, 0x0048); /* axis 0, reset error bit, start counter */
  out(1, 0x0e, 0x0048); /* axis 1, reset error bit, start counter */
  out (0, 0x1c, 0x0008);  /* fixed value to control register 2 */
  out (1, 0x1c, 0x0008);  /* fixed value to control register 2 */
  /* xxx: adjustments of 11uA offset-A/D-conversion registers 0x16, 0x18? */
  /* xxx: remember to calibrate digital potentiometers */

  // Amplitude watching.
  this->currAmpl[0] = this->currAmpl[1] = 0;  // normal
  this->minAmpl[0] = this->minAmpl[1] = 0;  // normal
}  // init

void cIK121::cleanup(void) {
}

void cIK121::startCountingFromZeroAtRefMark(void) {
  /* Stop and clear the counters. */
  out(0, 0x0e, 0x0030);  /* axis 0, stop & delete counter */
  out(1, 0x0e, 0x0030);  /* axis 1, stop & delete counter */
  out(0, 0x10, 0x0001);  /* axis 0, start counter at refmark */
  out(1, 0x10, 0x0001);  /* axis 1, start counter at refmark */
}

int cIK121::isSearchingForRefMark(int axis) {
  return ( (this->in(axis, 0x0e) & 0x0100) ? 1 : 0);
    /* xxx: note bit D15==logic level for reference mark */
    /* xxx: note bits D4==counter stopped and D6==frequency exceeded */
}

int cIK121::isStopped(int axis) {
  return ( (this->in(axis, 0x0e) & 0x0010) ? 1 : 0);
    /* xxx: note bits D4==counter stopped */
}

char *cIK121::amplStatus(int axis) {
  static char *expl[16] = {
  // D11 D10 D09 D08
  /*   0   0   0   0 */  "Ok Ok",
  /*   0   0   0   1 */  "Ok LO",
  /*   0   0   1   0 */  "Ok HI",
  /*   0   0   1   1 */  "Ok FA",
  /*   0   1   0   0 */  "LO Ok",
  /*   0   1   0   1 */  "LO LO",
  /*   0   1   1   0 */  "LO HI",
  /*   0   1   1   1 */  "LO FA",
  /*   1   0   0   0 */  "HI Ok",
  /*   1   0   0   1 */  "HI LO",
  /*   1   0   1   0 */  "HI HI",
  /*   1   0   1   1 */  "HI FA",
  /*   1   1   0   0 */  "FA Ok",
  /*   1   1   0   1 */  "FA LO",
  /*   1   1   1   0 */  "FA HI",
  /*   1   1   1   1 */  "FA FA",
  };
  int i;

  assert( (axis >= 0) && (axis <= 2) );
  i = (this->currAmpl[axis] << 2) | this->minAmpl[axis];
  return (expl[i]);
}  // amplStatus

void cIK121::resetAmplStatus(int axis) {
  out(axis, 0x0e, 0x0080);  // Only clear ampl value register with bit D7.
}  // resetAmplStatus

void cIK121::latchAtNextRefMark(void) {
  out(0, 0x10, 0x0008);  /* axis 0, latch counter at refmark */
  out(1, 0x10, 0x0008);  /* axis 1, latch counter at refmark */
}

long int cIK121::getLatchedRefMark(int axis) {
  /* xxx: while (this->isSearchingForRefMark(axis)) ; */
  return (((long)in(axis, 0x08)) << 16) | in(axis, 0x06);
}

void cIK121::trigger(void) {
  /* "Software latch" both axen simultaneously. */
  out(0, 0x0e, 0x0001);  /* axis 0, data reg 0 */
  out(1, 0x0e, 0x0001);  /* axis 1, data reg 0 */
    /* xxx: Should be changed to axis cascading and simultaneous latching. */
  this->waitingForValues = 1;
}

int cIK121::isReady(void) {
  unsigned int axis0;
  unsigned int axis1;

  if ( ! this->waitingForValues) {
    return (1);
  }

  /* Get 'ready' bits of both axen. */
  axis0 = this->in(0, 0x0e) & 0x0001;
  axis1 = this->in(1, 0x0e) & 0x0001;
    /* xxx: note bits D4==counter stopped and D6==frequency exceeded */
  return (axis0 && axis1);
}

long int cIK121::get(int axis) {
  if (this->waitingForValues) {
    /* Wait until both axen are ready. */
    while (!this->isReady()) ;
    this->waitingForValues = 0;

    /* Read them in. */
    this->latestValues[0] = (((long)in(0, 0x02)) << 16) | in(0, 0x00);
    this->latestValues[1] = (((long)in(1, 0x02)) << 16) | in(1, 0x00);

    { // Check reg. 0x10 for amplitude values: normal, low, high, faulty.
      // The IK121 manual says that the value should be frozen
      // before reading but that apparently stops the counter.
      // By testing with disconnecting the RON encoder we found
      // that the board reports immediately 11, "faulty low amplitude"
      // and that the minimum value is remembered on-board
      // until expressly cleared with control reg 1 bit D7==1.
      int i;

      i = in(0, 0x10) >> 8;
      this->currAmpl[0] = i & 0x3;
      this->minAmpl[0] = (i >> 2) & 0x3;
      i = in(1, 0x10) >> 8;
      this->currAmpl[1] = i & 0x3;
      this->minAmpl[1] = (i >> 2) & 0x3;
    }  // ampl check
  }  // if previously the board was latched for a new measurement

  return (this->latestValues[axis]);
}  // get
