#include "registration.h"
RCSID(cpwmisa_cpp, "$Id: cpwmisa.cpp,v 1.6 1999/10/18 12:58:49 amn Exp $");
/*------------------------------------------------------------*/
// New PWMISA board for five FET boards plus an UTC 1usec clock.

// $Log: cpwmisa.cpp,v $
// Revision 1.6  1999/10/18 12:58:49  amn
// Added comments about speed/direction settings.
//
// Revision 1.5  1999/10/12 08:32:46  amn
// Corrected registration identifier.
//
// Revision 1.4  1999/10/12 07:52:31  amn
// Added registration and log.
//

class cPWMISA {
private:
  unsigned int baseAddr;

// FET board / motor directions.
#define UP   0  /* Motor dir UP/CW */
#define DOWN 1  /* Motor dir DOWN/CCW */
#define TOGGLE 2  /* Motor dir / enable signal opposite than what it was. */

// PWM/antiphase max value.
#define MAX_PWM 1023

// Handling of DIRENA bits.
#define NUMOFMOTORS 5
  unsigned char direnaBits[NUMOFMOTORS];
#define DIRENA_DIR 0x01
#define DIRENA_ENABLE 0x02
#define DIRENA_RELAY 0x04
#define DIRENA_RESETLATCHED 0x08
  virtual void updateDirena(int motor);

public:
  virtual void init(int baseAddr = 0x280);
  virtual void cleanup(void);

  virtual void setBasePWMFreq(int freq);
#define PWMBASE20KHZ 0
#define PWMBASE40KHZ 1
#define PWMBASE10KHZ 2

  virtual void refreshWatchdog(void);

  // Returns 0..999999usecs, syncronized to UTC 1pps signal.
  virtual unsigned long int getRawUsecs(void);  // secs and usecs
  virtual unsigned long int getUsecs(void);  // only usecs
  virtual int getSecs(void);  // only secs, 0..3

  virtual void setEnable(int motor, unsigned char newState);
  virtual unsigned char getEnable(int motor);
  virtual void setRelay(int motor, unsigned char newState);
  virtual unsigned char getRelay(int motor);
  virtual void disableAll(void);

  virtual int getStatus(int motor);
#define ST_DIR_RB 0x01
#define ST_ENABLED 0x02
#define ST_RELAY_RB 0x04
#define ST_CHOP 0x08
#define ST_FAULT 0x10
#define ST_STOPPED 0x20
#define ST_ANTIPH_RB 0x03C0
  virtual void formatStatus(int status, char *formatted);
  virtual void resetStatus(int motor);

  // Sets both speed and direction, <0 down/ccw, >0 up/cw.
  virtual void setSpeed(int motor, int speed);
};  // class cPWMISA



void cPWMISA::disableAll(void) {
  int i;

  /* Disable all enable signals of all motors. */
  for (i=0; i < NUMOFMOTORS; i++) {
    setEnable(i, 0);
    setRelay(i, 0);
    resetStatus(i);
    setSpeed(i, /*speed=>*/ 0);
  }  // for each motor
}  // disableAll


void cPWMISA::init(int baseAddr) {
  int i;

  this->baseAddr = baseAddr;

  setBasePWMFreq(PWMBASE20KHZ);

  disableAll();
}  // init

void cPWMISA::cleanup(void) {
  disableAll();
}  // cleanup


void cPWMISA::updateDirena(int motor) {
  outport(
    this->baseAddr + (motor<<3) + 0x04,
    this->direnaBits[motor]
  );
}  // updateDirena


void cPWMISA::setBasePWMFreq(int freq) {
  assert( (freq >= 0) && (freq <= 2) );
  outport(this->baseAddr + 0x3a, freq);
}  // setBasePWMFreq

void cPWMISA::refreshWatchdog(void) {
  outport(this->baseAddr+0x38, 1);
}  // refreshWatchdog


unsigned long int cPWMISA::getRawUsecs(void) {
  unsigned long int c;

  // High 11 bits, latches also low bits.
  c = inport(this->baseAddr + 0x32) & 0x07ff;
  c = (c << 11) | (inport(this->baseAddr + 0x30) & 0x7ff);  // low 11 bits
  // Seems to return 22 1-bits == 2^23-1 if board is missing or powered off.
  return c;
}  // getRawUsecs

unsigned long int cPWMISA::getUsecs(void) {
  return (this->getRawUsecs() & 0x000fffff);
}  // getUsecs

int cPWMISA::getSecs(void) {
  return ( (this->getRawUsecs() & 0x00300000) >> 20);
}  // getSecs


void cPWMISA::setEnable(int motor, unsigned char newState) {
#define CHECKMOTOR assert( (motor >= 0) && (motor <= NUMOFMOTORS) )
  CHECKMOTOR;
  if (/* (newState < 0) || unsigned! */ (newState > 1)) {
    // Toggle.
    newState = ((this->direnaBits[motor] & DIRENA_ENABLE) ? 0 : 1);
  }
  if (newState) {
    this->direnaBits[motor] |= DIRENA_ENABLE;
  } else {
    this->direnaBits[motor] &= ~(DIRENA_ENABLE);
  }
  updateDirena(motor);
}  // setEnable

unsigned char cPWMISA::getEnable(int motor) {
  CHECKMOTOR;
  return ( (this->direnaBits[motor] & DIRENA_ENABLE) ? 1 : 0 );
}  // getEnable

void cPWMISA::setRelay(int motor, unsigned char newState) {
  CHECKMOTOR;
  if (/* (newState < 0) || unsigned! */ (newState > 1)) {
    // Toggle.
    newState = ((this->direnaBits[motor] & DIRENA_RELAY) ? 0 : 1);
  }
  if (newState) {
    this->direnaBits[motor] |= DIRENA_RELAY;
  } else {
    this->direnaBits[motor] &= ~(DIRENA_RELAY);
  }
  updateDirena(motor);
}  // setRelay

unsigned char cPWMISA::getRelay(int motor) {
  CHECKMOTOR;
  return ( (this->direnaBits[motor] & DIRENA_RELAY) ? 1 : 0 );
}  // getRelay


int cPWMISA::getStatus(int motor) {
  return (inport(this->baseAddr + (motor<<3) + 0x06));
}  // getStatus

void cPWMISA::formatStatus(int status, char *formatted) {
  formatted[7] = '\0';
  if (status & ST_DIR_RB) {
    formatted[6] = 'D';  // 1==ccw/down
  } else {
    formatted[6] = 'U';
  }
  if (status & ST_ENABLED) {
    formatted[5] = 'E';
  } else {
    formatted[5] = '_';
  }
  if (status & ST_RELAY_RB) {
    formatted[4] = 'R';
  } else {
    formatted[4] = '_';
  }
  if (status & ST_CHOP) {
    formatted[3] = 'C';
  } else {
    formatted[3] = '_';
  }
  if (status & ST_FAULT) {
    formatted[2] = 'F';
  } else {
    formatted[2] = '_';
  }
  if (status & ST_STOPPED) {
    formatted[1] = 'S';
  } else {
    formatted[1] = '_';
  }
  formatted[0] = ((status & ST_ANTIPH_RB) >> 6) + '0';
  if (formatted[0] > '9') {
    formatted[0] += ('A' - '0');
  }
}  // formatStatus

void cPWMISA::resetStatus(int motor) {
  // Turn bit to 1 but don't remember it, next time it will be 0 again.
  outport(
    this->baseAddr + (motor<<3) + 0x04,
    this->direnaBits[motor] | DIRENA_RESETLATCHED
  );
}  // resetStatus


// Set speed (and direction) of one FET board/motor.
void cPWMISA::setSpeed(
  int motor,
  int speed
) {
  CHECKMOTOR;
  if (speed < 0) {
    speed = -speed;
    this->direnaBits[motor] |= DIRENA_DIR;  // to down/ccw == 1
  } else {
    this->direnaBits[motor] &= ~(DIRENA_DIR);  // to up/cw == 0
  }
  outport(this->baseAddr + (motor<<3) + 0x00, speed);
  outport(this->baseAddr + (motor<<3) + 0x02, /*antiphase=>*/ 0);
  updateDirena(motor);
}  // setSpeed
