#include "registration.h"
RCSID(croc4xx_cpp, "$Id: croc4xx.cpp,v 1.7 2000/08/10 09:32:50 amn Exp $");
/* ------------------------------------------------------------------ */
// Handling of Heidenhain ROC4xx parallel absolute encoders
// connected to a ACL-712 parallel 50-pin I/O card.

// $Log: croc4xx.cpp,v $
// Revision 1.7  2000/08/10 09:32:50  amn
// Added support for SunAnt SSI encoders.
//
// Revision 1.6  1999/10/28 18:14:11  amn
// Debugged cROC4xx repeats (7 not enough...).
// Added ignoring values >10% off from last 4 in cSensorChannel.
// Added 'tTime' alias struct timeval (2x longs) into ctime.cpp.
//
// Revision 1.5  1999/10/28 08:26:49  amn
// Added CVS log.
//

/* ------------------------------------------------------------------ */
/* Gray code conversions. */
#include "gray.cpp"

/* ------------------------------------------------------------------ */
#define IN 1
#define OUT 0
/* Set ports of one 50-pin connector of ET-1722B to IN or OUT. */
void setPortABCMode(
  int baseAddr,
  int conn,
  int portA,
  int portB,
  int portCUpper,
  int portCLower
) {
  /* Control word layout: D7==always 1, D4==port A direction (0==out),
	 D3==portCUpper, D2==always 0, D1==portB, D0==portCLower,
	 D6, D5==always 0. */
  outportb(
	baseAddr + (conn<<2) + 3,
	0x80
	  | ((portA & 0x01)<<4)
	  | ((portCUpper & 0x01)<<3)
	  | ((portB & 0x01)<<1)
	  | (portCLower & 0x01)
  );
}


/* ------------------------------------------------------------------ */
/* Represents a HH ROC4xx parallel TTL absolute encoder connected to
   one 50-pin connector (3x8 bit ports A, B, C) of ET-1722B board. */
typedef unsigned int tROC4xxBits;
class cPortROC4xx : public cSensor {
private:
  int baseAddr;
  int conn;  /* connector on ET-1722B board, 0..5 */
  int relAddr;  /* address of 8255 to be used, calc from 'baseAddr' and 'conn' */
  tROC4xxBits numOfBitsMask;  /* MSB of ROC4xx set to 1, all others 0 */
	/* This determines number of bits in ROC4xx. */
  double binToDegreesCoeff;  /* calculated from 'numOfBitsMask' */
  int attemptLimit;  /* attempt this many times to get a stable reading */
  int maxAttempts;  /* max attempt count observed */

public:
  virtual void init(
	int baseAddr,
	int conn,
	tROC4xxBits numOfBitsMask,
	int attemptLimit =
#if DEBUG_REPEATS
	  7
#else
	  10
#endif
  );
protected:
  virtual tROC4xxBits convertPortToGray(unsigned char pa, unsigned char pb);
public:
  virtual tROC4xxBits getRawGray(void);
  virtual tROC4xxBits getStableGray(void);
  virtual tROC4xxBits getStableBin(void);
  virtual long int getRaw(void);
  virtual double getUncalibrated(void);
  virtual void cleanup(void);
};  /* cPortROC4xx */

/* Initialize one 50-pin connector for one ROC4xx. */
/* Port A (6 LSB) and port B (6 MSB) as inputs, port C as outputs */
/* (D0=release A, active 0, LSB out; D1=release B, active 0, MSB out). */
void cPortROC4xx::init(
  int baseAddr,  /* of ET1722B card (or any 8255-compatible parallel) */
  int conn,  /* 0..5 */
  tROC4xxBits numOfBitsMask,  /* only MSBit is == 1 */
  int attemptLimit
) {
  cSensor::init();

  this->baseAddr = baseAddr;
  this->conn = conn;
  this->relAddr = baseAddr + (conn<<2);
  this->numOfBitsMask = numOfBitsMask;
  this->binToDegreesCoeff = 360.0 / (double)(numOfBitsMask<<1);
  this->attemptLimit = attemptLimit;
  this->maxAttempts = 0;

  /* Init output port to a known value before changing direction;
	 port C to all ones to disable ROC4xx tri-state output buffers,
	 or to all zeroes to permanently enable all output bit lines. */
  /* xxx: ROC412 should have enabled output lines within 60ns of enable;
     in reality scope shows 1.4ms... Thus we keep output enabled. */
#if REAL_REL_AB
  outportb(relAddr + 2, 0xff);
#else
  outportb(relAddr + 2, 0x00);
#endif
  setPortABCMode(baseAddr, conn,
	/*A=>*/ IN, /*B=>*/ IN, /*Cupper=>*/ OUT, /*Clower=>*/ OUT
  );
  /* Set RelA, RelB of ROC4xx again; ET-1722B documentation conflicts
	 here: part of it says set output value before setting direction,
	 part of it says set direction before output value... */
#if REAL_REL_AB
  outportb(relAddr + 2, 0xff);
#else
  outportb(relAddr + 2, 0x00);
#endif
}

void cPortROC4xx::cleanup() {
  setPortABCMode(baseAddr, conn,
	/*A=>*/ IN, /*B=>*/ IN, /*Cupper=>*/ IN, /*Clower=>*/ IN
  );
#if DEBUG_REPEATS
  printf("cPortROC4xx: max number of read retries = %d\n",
	this->maxAttempts
  );
#endif
  cSensor::cleanup();
}

/* This function will be combine the bits from two 8-bit parallel ports */
/* to match the correct number of bits for a given ROC4xx. */
tROC4xxBits cPortROC4xx::convertPortToGray(
  unsigned char pa,
  unsigned char pb
) {
  (void)pa; (void)pb;  /* "use" parameters to avoid warning messages */
  assert(("abstract 'cPortROC4xx::convertPortToGray()' called", 0));
  return (-1);
}

/* Get a raw Gray code value from a ROC4xx. */
tROC4xxBits cPortROC4xx::getRawGray(void) {
  unsigned char pa, pb;

#if REAL_REL_AB
  outportb(relAddr + 2, 0xfe);  /* release A, LSB */
#endif
  pa = inportb(relAddr);  /* port A */
#if REAL_REL_AB
  outportb(relAddr + 2, 0xfd);  /* release B, MSB */
#endif
  pb = inportb(relAddr + 1);  /* port B */
#if REAL_REL_AB
  outportb(relAddr + 2, 0xff);  /* inhibit both LSB and MSB */
#endif
  return ( this->convertPortToGray(pa, pb) );
}

/* Read repeatedly a raw Gray code reading until two successive
   readouts return the same vale or tried too many times.
   Returns a negative value in case of too many retries. */
tROC4xxBits cPortROC4xx::getStableGray(void) {
  tROC4xxBits prev, now;
  int attempts;
#if DEBUG_REPEATS
  tROC4xxBits attValue[100];
#endif

  now = this->getRawGray();
  attempts = 1;
#if DEBUG_REPEATS
  attValue[attempts] = now;
#endif
  do {
	prev = now;
	now = this->getRawGray();
	attempts++;
#if DEBUG_REPEATS
	attValue[attempts] = now;
#endif
  } while ((prev != now) && (attempts < this->attemptLimit));
  if (attempts > this->maxAttempts) this->maxAttempts = attempts;
  if (prev == now) {
	return (now);
  } else {
#if DEBUG_REPEATS
    int x = wherex();
	int y = wherey();
	int i;

	gotoxy(50, 7);
	for (i=1; i <= this->attemptLimit; i++) {
	  printf("%04X ", attValue[i]);
	}
	// sound(1000); delay(100); nosound();
    gotoxy(x, y);
#endif
	return (0xffff);
  }
}

/* Get a stable reading out of ROC4xx and convert it to binary 0..n. */
/* Returns a negative value if a stable reading cannot be found. */
tROC4xxBits cPortROC4xx::getStableBin(void) {
  tROC4xxBits temp;

  temp = this->getStableGray();
  if (temp < (this->numOfBitsMask<<1) /* xxx: max 15 bit! */) {
	temp = gray2bin(temp, this->numOfBitsMask);
  }
  /* Else 'temp' will be left larger than the maximum, i.e. error. */
  return (temp);
}

/* A common version of 'getStableBin()' */
long int cPortROC4xx::getRaw(void) {
  return (this->getStableBin());
}

/* Convert binary value 0..n to degrees ('double'). */
/* This provides an implementation to abstract 'cSensor::getUncalibrated()'. */
double cPortROC4xx::getUncalibrated(void) {
  return ((double)this->getStableBin() * this->binToDegreesCoeff);
}

/* ------------------------------------------------------------------ */
/* A specific 'cPortROC4xx' for 12-bit ROC412. */
class cPortROC412 : public cPortROC4xx {
protected:
  virtual tROC4xxBits convertPortToGray(unsigned char pa, unsigned char pb);
};  /* cPortROC412 */

/* This function combines the bits from two 8-bit parallel ports */
/* 6 LSBits from port A and 6 MSBits from port B == 12 bits in total. */
tROC4xxBits cPortROC412::convertPortToGray(
  unsigned char pa,
  unsigned char pb
) {
  return ( ((pb & 0x3f)<<6) | (pa & 0x3f) );
}
