#include "registration.h"
RCSID(csensor_cpp, "$Id: csensor.cpp,v 1.4 1999/10/28 18:14:11 amn Exp $");
/* ------------------------------------------------------------------ */
/* An abstract base class for all sensor/encoder measurement objects. */
/* All these can return a 'double' type measurement value. */

// $Log: csensor.cpp,v $
// Revision 1.4  1999/10/28 18:14:11  amn
// Debugged cROC4xx repeats (7 not enough...).
// Added ignoring values >10% off from last 4 in cSensorChannel.
// Added 'tTime' alias struct timeval (2x longs) into ctime.cpp.
//
// Revision 1.3  1999/10/12 07:46:24  amn
// Added channel number to 'cSensorChannel'.
//

class cSensor {
private:
  double offset;
  double coeff;
public:
  virtual void init(void);
  virtual void cleanup(void);
  virtual long int getRaw(void);
  virtual double getUncalibrated(void);
  virtual double calibrated(double d);
  virtual void setCalibration(double offset, double coeff);
  virtual double get(void);
};  /* cSensor */

void cSensor::init(void) {
  this->setCalibration(0.0, 1.0);
}

void cSensor::cleanup(void) {
}

void cSensor::setCalibration(double offset, double coeff) {
  this->offset = offset;
  this->coeff = coeff;
}

long int cSensor::getRaw(void) {
  assert(("abstract 'cSensor::getRaw()' called", 0));
  return (0);
}

double cSensor::getUncalibrated(void) {
  assert(("abstract 'cSensor::getUncalibrated()' called", 0));
  return (0.0);
}

double cSensor::calibrated(double d) {
  return (d * this->coeff + this->offset);
}

double cSensor::get() {
  return (this->calibrated(this->getUncalibrated()));
}


/* ------------------------------------------------------------------ */
/* A class for individual A/D channels and other separately */
/* scaled/calibrated 'cSensor::get()' values. */
/* By the way, 'cSensorChannel' is a specialized 'cSensor' itself. */
class cSensorChannel : public cSensor {
private:
  class cSensor *mySensor;
  int channelNumber;
  double lastGets[4];  // latest uncalibrated values
  int lg;
public:
  virtual void init(class cSensor *sensor);
  /* inherited 'setCalibration(double offset, double coeff)' */
  virtual long int getRaw(void);
  virtual double getUncalibrated(void);
  /* inherited 'double calibrated(double d)' */
  /* inherited 'double get(void)' */
  virtual void setChannelNumber(int ch);
  virtual int getChannelNumber(void);
  virtual void cleanup(void);
};  /* cSensorChannel */

void cSensorChannel::init(class cSensor *sensor) {
  cSensor::init();
  this->mySensor = sensor;
  this->channelNumber = 0;
  this->setCalibration(0.0, 1.0);
  this->lastGets[0] = this->lastGets[1]
    = this->lastGets[2] = this->lastGets[3] = 0.0;
  this->lg = 0;
}

long int cSensorChannel::getRaw(void) {
  return (this->mySensor->getRaw());
}

double cSensorChannel::getUncalibrated(void) {
  double v = this->mySensor->get();
  double avg;

  avg = (this->lastGets[0] + this->lastGets[1]
    + this->lastGets[2] + this->lastGets[3]) / 4.0;

  this->lastGets[this->lg++] = v;
  if (this->lg >= 4) {
    this->lg = 0;
  }

  // If deviates more than 10% from old average, return old average.
  if (fabs(v - avg) > fabs(0.1*avg)) {
    return (avg);
  } else {
    // Return new fresh value.
    return (v);
  }
}  // getUncalibrated

void cSensorChannel::setChannelNumber(int ch) {
  this->channelNumber = ch;
}

int cSensorChannel::getChannelNumber(void) {
  return (this->channelNumber);
}

void cSensorChannel::cleanup(void) {
  this->mySensor = NULL;
  cSensor::cleanup();
}
