#include "registration.h"
RCSID(cssi_cpp, "$Id: cssi.cpp,v 1.2 2000/10/23 07:07:24 amn Exp $");
/* ------------------------------------------------------------------ */
// Handling of SSI (Synchronous Serial Interface) absolute encoders
// connected to a ACL-712 parallel 50-pin I/O card
// using a SSIDEC Metsahovi SSI decoder board.

// $Log: cssi.cpp,v $
// Revision 1.2  2000/10/23 07:07:24  amn
// Added a function for getting "degrees per bit" coefficient.
//
// Revision 1.1  2000/08/10 09:32:50  amn
// Added support for SunAnt SSI encoders.
//

// Gray code conversions.
#include "gray.cpp"

/* ------------------------------------------------------------------ */
#define IN 1
#define OUT 0
/* Set ports of one 50-pin connector of ET-1722B to IN or OUT. */
void setPortABCMode(
  int baseAddr,
  int conn,
  int portA,
  int portB,
  int portCUpper,
  int portCLower
) {
  /* Control word layout: D7==always 1, D4==port A direction (0==out),
	 D3==portCUpper, D2==always 0, D1==portB, D0==portCLower,
	 D6, D5==always 0. */
  outportb(
	baseAddr + (conn<<2) + 3,
	0x80
	  | ((portA & 0x01)<<4)
	  | ((portCUpper & 0x01)<<3)
	  | ((portB & 0x01)<<1)
	  | (portCLower & 0x01)
  );
}


/* ------------------------------------------------------------------ */
/* Represents a SSI absolute encoder connected via SSIDEC board to
   one 50-pin connector (3x8 bit ports A, B, C) of ET-1722B board. */
typedef unsigned int tSSIBits;
class cPortSSI : public cSensor {
private:
  int baseAddr;
  int conn;  /* connector on ET-1722B board, 0..5 */
  int relAddr;  /* address of 8255 to be used, calc from 'baseAddr' and 'conn' */
  tSSIBits numOfBitsMask;  /* MSB of SSI set to 1, all others 0 */
	/* This determines number of bits in SSI. */
  tSSIBits hiBitsAwayMask;  /* MSBs set to 1, all others 0 */
  double binToDegreesCoeff;  /* calculated from 'numOfBitsMask' */
  int attemptLimit;  /* attempt this many times to get a stable reading */
  int maxAttempts;  /* max attempt count observed */
  int maxWaits;  /* max # of wait loops for 'ready' signal */

public:
  virtual void init(
	int baseAddr,
	int conn,
	tSSIBits numOfBitsMask,
	int attemptLimit =
#if DEBUG_REPEATS
	  7
#else
	  10
#endif
  );
public:
  virtual double getBinToDegreesCoeff(void);
  virtual tSSIBits getRawGray(void);
  virtual tSSIBits getStableGray(void);
  virtual tSSIBits getStableBin(void);
  virtual long int getRaw(void);
  virtual double getUncalibrated(void);
  virtual void cleanup(void);
};  /* cPortSSI */

/* Initialize one 50-pin connector for one SSI. */
/* Port A (6 LSB) and port B (6 MSB) as inputs, port C as outputs */
/* (D0=release A, active 0, LSB out; D1=release B, active 0, MSB out). */
void cPortSSI::init(
  int baseAddr,  /* of ET1722B card (or any 8255-compatible parallel) */
  int conn,  /* 0..5 */
  tSSIBits numOfBitsMask,  /* only MSBit is == 1 */
  int attemptLimit
) {
  cSensor::init();

  this->baseAddr = baseAddr;
  this->conn = conn;
  this->relAddr = baseAddr + (conn<<2);
  this->numOfBitsMask = numOfBitsMask;
  this->hiBitsAwayMask = numOfBitsMask << 1;
  while (this->hiBitsAwayMask < 0x8000) {
    this->hiBitsAwayMask |= (this->hiBitsAwayMask << 1);
  }   // while not all MSBs are ==1
  this->hiBitsAwayMask = ~(this->hiBitsAwayMask);  // invert; MSBs ==0
  this->binToDegreesCoeff = 360.0 / (double)(numOfBitsMask<<1);
  this->attemptLimit = attemptLimit;
  this->maxAttempts = 0;
  this->maxWaits = 0;

  /* Init output port to a known value before changing direction;
	 to all zeroes. */
  outportb(relAddr + 2, 0x00);
  setPortABCMode(baseAddr, conn,
	/*A=>*/ IN, /*B=>*/ IN, /*Cupper=>*/ OUT, /*Clower=>*/ IN
  );
  /* Set RelA, RelB of SSI again; ET-1722B documentation conflicts
	 here: part of it says set output value before setting direction,
	 part of it says set direction before output value... */
  outportb(relAddr + 2, 0x00);
}  // init

void cPortSSI::cleanup() {
  setPortABCMode(baseAddr, conn,
	/*A=>*/ IN, /*B=>*/ IN, /*Cupper=>*/ IN, /*Clower=>*/ IN
  );
#if DEBUG_REPEATS
  printf("cPortSSI: max number of read retries = %d, waits = %d\n",
	this->maxAttempts,
	this->maxWaits
  );
#endif
  cSensor::cleanup();
}  // cleanup

double cPortSSI::getBinToDegreesCoeff(void) {
  return (this->binToDegreesCoeff);
}  // getBinToDegreesCoeff

/* Get a raw Gray code value from a SSI. */
tSSIBits cPortSSI::getRawGray(void) {
  int waits;
  unsigned char v;
  unsigned int t;
  int i;

  // Wait for as long as 'ready' stays ==0.  (Encoder still busy
  // with previous conversion or no new value available yet.)
  waits = 0;
  while ( (inportb(this->relAddr+2) & 0x04) == 0 ) {  // SSIDEC & ECN
    waits++;
    if (waits > 1000) {
      return (0xffff);  // xxx: ?
    }
  }  // while encoder not ready

  // Toggle 'start' bit 0-->1-->0.
  // xxx: check the length of the pulse, must be >cycle time of
  //      SSI clock (90kHz..500kHz => 2us..11.1us minimum).
  for (i=0; i<12; i++) {  // >12usec "delay"
    outportb(this->relAddr+2, 0x80);
  }
  outportb(this->relAddr+2, 0x00);

#if 1
  delay(1);
#else
  // Wait until chip has started a conversion (BUSY ==1).
  // xxx: actually now can be removed, because start pulse
  // is longer than one clock cycle
  while ( (inportb(this->relAddr+2) & 0x08) == 0 ) {  // while chip not busy
    waits++;
    if (waits > 1000) {
      return (0xffff);  // xxx: ?
    }
  }  // while chip is not yet busy
#endif

  // Wait until chip has received a new readout.
  // (Wait until READY-->1, includes both chip and encoder readiness.)
  while ( (inportb(this->relAddr+2) & 0x04) == 0 ) {
    waits++;
    if (waits > 1000) {
      return (0xffff);  // xxx: ?
    }
  }  // while encoder and SSIDEC chip not ready
  if (waits > this->maxWaits) {
    // Store new maxmimum wait loop counter.
    this->maxWaits = waits;
  }

  // Get high byte.
  t = inportb(this->relAddr+1);

  // Get low byte.
  t = (t << 8) | inportb(this->relAddr+0);

  t >>= 1;  // remove one LSB junk bit...
  t &= this->hiBitsAwayMask;  // ...and MSB junk bits
  return ( t );
}  // getRawGray

/* Read repeatedly a raw Gray code reading until two successive
   readouts return the same value or tried too many times.
   Returns a negative value in case of too many retries. */
tSSIBits cPortSSI::getStableGray(void) {
  tSSIBits prev, now;
  int attempts;
#if DEBUG_REPEATS
  tSSIBits attValue[100];
#endif

  now = this->getRawGray();
  attempts = 1;
#if DEBUG_REPEATS
  attValue[attempts] = now;
#endif
  do {
	prev = now;
	now = this->getRawGray();
	attempts++;
#if DEBUG_REPEATS
	attValue[attempts] = now;
#endif
  } while ((prev != now) && (attempts < this->attemptLimit));
  if (attempts > this->maxAttempts) this->maxAttempts = attempts;
  if (prev == now) {
	return (now);
  } else {
#if DEBUG_REPEATS
    int x = wherex();
	int y = wherey();
	int i;

	gotoxy(50, 7);
	for (i=1; i <= this->attemptLimit; i++) {
	  printf("%04X ", attValue[i]);
	}
	// sound(1000); delay(100); nosound();
    gotoxy(x, y);
#endif
	return (0xffff);
  }
}

/* Get a stable reading out of SSI and convert it to binary 0..n. */
/* Returns a negative value if a stable reading cannot be found. */
tSSIBits cPortSSI::getStableBin(void) {
  tSSIBits temp;

  temp = this->getStableGray();
  if (temp < (this->numOfBitsMask<<1) /* xxx: max 15 bit! */) {
	temp = gray2bin(temp, this->numOfBitsMask);
  }
  /* Else 'temp' will be left larger than the maximum, i.e. error. */
  return (temp);
}

/* A common version of 'getStableBin()' */
long int cPortSSI::getRaw(void) {
  return (this->getStableBin());
}

/* Convert binary value 0..n to degrees ('double'). */
/* This provides an implementation to abstract 'cSensor::getUncalibrated()'. */
double cPortSSI::getUncalibrated(void) {
  return ((double)this->getStableBin() * this->binToDegreesCoeff);
}
