/* down/ant.cpp  --  Showing sensor readouts in "down" PC.  */

/* When         Who  What */
/* 16-Apr-1997  amn  Adding logging. */
/* 02-Dec-1996  amn  Fixing 'getUntilStaysSame()'.
/* 29-Nov-1996  amn  Jouko's new VAXIF board. */
/* 01-Oct-1996  amn  c8255::setPortMode didn't use 'this->out()' correctly. */
/* 27-Sep-1996  amn  VAX, serial lost bytes check, VAX out. */
/* 09-Sep-1996  amn  Remote IK121. */
/* 08-Sep-1996  amn  Initializing the reference mark. */
/* 05-Sep-1996  amn  Original version. */

#include <conio.h>
#include <stdio.h>
#include <dos.h>

#include <assert.h>
#include <math.h>  /* for M_PI */
#include <mem.h>  /* for memcpy() */
#include <time.h>  /* for time functions */

/* Testing a local IK121 board. */
#define LOCAL_IK121 0
#define TEST_DA 0
#define TEST_POS 1

#include "serial.h"

/* Reads one character; if timeout, returns 0xff. */
unsigned char getOne(void) {
  unsigned int i = 50000;
  int c;

  while (i && ((c = ReadSerial(COM2)) == -1))
    i--;
  return ((unsigned char)c);
}


#if LOCAL_IK121
#include "cIK121.cpp"
#endif


/* -------------------------------------------------------------------- */
/* ET-1728 2ch D/A converter. */
class cET1728 {
private:
  unsigned int baseAddr;
  /* xxx: range: 5/10, uni/bipolar */
public:
  virtual void init(int baseAddr);
  virtual void cleanup(void);
  virtual void setRaw(int channel /* 0..1 */, int /* 12 bit */ value);
  virtual void set(int channel /* 0..1 */, double volts);
};  /* cET1728 */

void cET1728::init(int baseAddr) {
  this->baseAddr = baseAddr;
  /* Set outputs to 0V. */
  this->set(0, 0.0);
  this->set(1, 0.0);
}

void cET1728::cleanup(void) {
  /* Set outputs to 0V. */
  this->set(0, 0.0);
  this->set(1, 0.0);
}

void cET1728::setRaw(int channel /* 0..1 */, int /* 12 bit */ value) {
  unsigned int addr = this->baseAddr | ((channel & 0x01) << 1);

  outportb(addr, (unsigned char)((value >> 8) & 0x0F));  /* MSB (12-bit, upper nibble zero) */
  outportb(addr+1, (unsigned char)(value & 0xFF));  /* LSB; it latches new value to D/A chip */
}

void cET1728::set(int channel /* 0..1 */, double volts) {
  if (volts > 10.0) {
    volts = 10.0;
  } else if (volts < -10.0) {
    volts = -10.0;
  }
  this->setRaw(
    channel,
    (0x0fff & (
     ( ((int)(volts * 204.7)) + 2048 )
    ))
  );
}


/* -------------------------------------------------------------------- */
/* New VAXIF MACH211-based VAX interface board. */
class cVAXIF {
private:
  unsigned int baseAddr;
  virtual long int getOne(int offset);
  virtual long int getOneUntilStaysSame(int offset);
  virtual double getConverted(int offset);
  virtual void sendOne(int offset, long int value);
  virtual void sendConverted(int offset, double value);
public:
  virtual void init(int baseAddr);
  virtual void cleanup(void);
  virtual double getAz(void);
  virtual double getEl(void);
  virtual void sendAz(double az);
  virtual void sendEl(double el);
};  /* cVAXIF */

void cVAXIF::init(int baseAddr) {
  this->baseAddr = baseAddr;
  /* xxx: Set outputs to VAX to ? */
}

void cVAXIF::cleanup(void) {
}

long int cVAXIF::getOne(int offset) {
  long int value;

  value = inportb(this->baseAddr+offset+2);  /* high */
  value = (value<<8) | inportb(this->baseAddr+offset+1);  /* mid */
  value = (value<<8) | inportb(this->baseAddr+offset);  /* low */
    /* At the moment the board doesn't have any special latching */
    /* based on the order in which bytes are being read. */
  return value;
}

long int cVAXIF::getOneUntilStaysSame(int offset) {
  long int prevValue;
  long int value;
  int retries = 0;

  value = this->getOne(offset);
  do {
    prevValue = value;
    value = this->getOne(offset);
    retries++;
  } while ((prevValue != value) && (retries < 10));
  /* xxx: what to do if retries > 10? */
  return value;
}

double cVAXIF::getConverted(int offset) {
  long int value;

  value = this->getOneUntilStaysSame(offset);

  /* Extend MSB == 1  => negative in 2-complement. */
  if (value & 0x00200000) {
    /* MSB sign bit == 1, extend. */
    value |= 0xFFE00000;
  }
  return ((double)value * 360.0 / 2097152.0);
}  /* getConverted */

void cVAXIF::sendOne(int offset, long int value) {
  value <<= 2;  /* VAX aligns MSBs, so fill two LSBs with 0 */
  outportb(this->baseAddr+offset,  value & 0x000000FF);  /* low */
  outportb(this->baseAddr+offset+1, (value & 0x0000FF00)>>8);  /* mid */
  outportb(this->baseAddr+offset+2, (value & 0x00FF0000)>>16);  /* high */
    /* High byte latches all of 24(22/21) bits to VAX. */
}

void cVAXIF::sendConverted(int offset, double value) {
  long int signBit;

  if (value < 0.0) {
    value = -value;
    signBit = 0x00000000;
  } else {
    signBit = 0x00200000;  /* positive == sign bit is on; not 2-compl. */
  }
  this->sendOne(offset, (long int)(value * 2097152.0 / 360.0) | signBit);
}  /* sendConverted */

double cVAXIF::getAz(void) {
  return this->getConverted(/*offset=>*/0);
}  /* getAz */

double cVAXIF::getEl(void) {
  return this->getConverted(/*offset=>*/4);
}  /* getEl */

void cVAXIF::sendAz(double az) {
  this->sendConverted(/*offset=>*/0, az);
}  /* sendEl */

void cVAXIF::sendEl(double el) {
  this->sendConverted(/*offset=>*/4, el);
}  /* sendEl */


/* Definition of the struct of all sensor values on the serial line down. */
#include "downlink.h"

typedef union {
  tDownlink downlink;
  unsigned char buff[sizeof(tDownlink)];
} tDLBuff;


int getVal(char *prompt) {
  char buff[255];
  int temp;

  printf("%s", prompt);
  gets(buff);
  (void)sscanf(buff, "%d", &temp);
  return (temp);
}

double getDouble(char *prompt) {
  char buff[255];
  double temp;

  printf("%s", prompt);
  gets(buff);
  (void)sscanf(buff, "%lg", &temp);
  return (temp);
}


/* ------------------------------------------------------------------ */
int main(int argc, char **argv) {
#if LOCAL_IK121
  class cIK121 vIK121;
#endif
  class cET1728 vET1728;
  class cVAXIF vVAXIF;
  double vTargetAz = -36.0;
  double vTargetEl = 46.0;
  double vAzPanelOffset = 0.0;
  double vElPanelOffset = 0.0;
  double vAzErr;
  double vElErr;
  double vDiffVoltsAz;
  double vDiffVoltsEl;
  double vJumpDegrees = 0.0;
  int vRunMode = 1;  /* 0==stopped as manual position, 1==VAX-track */
  double vLogInterval = -1.0;  // seconds; when negative, logging disabled
  FILE *vLogFile;
  time_t vNextLogTime = 0;  // immediately, since 0 means year 1970
  time_t vLastTime;
  long int vSerialReadErrors = 0;

  /* "Use" argc, argv to prevent compiler warning messages. */
  (void)argc; (void)argv;

  /* New VAXIF MACH211-based VAX interface board. */
  /* Supports both getting VAX commanded az/el and */
  /* sending encoder az/el values back to VAX. */
  vVAXIF.init(
    /*baseAddr=>*/ 0x300
  );

#if LOCAL_IK121
  /* Incremental encoders. */
  vIK121.init(/*baseAddr=>*/ 0x320);
#endif

  /* Position error signal D/A converter. */
  vET1728.init(/*baseAddr=>*/ 0x2c0);

  /* Serial uplink. */
  OpenSerial(COM2, 115200L, D8S1N, NULL);

  /* Open log file always in append mode. */
  assert ( (vLogFile = fopen("ant.log", "at")) != NULL );  // append in text mode

  clrscr();
  gotoxy(1, 1);
  printf("                 Elevation   Azimuth  (degrees)\n");
  printf("Absolute\n");
  printf("Incremental\n");
  printf("Target\n");
  printf("Pos. error\n");
  printf("\n");
  printf("VAX Target\n");
  printf("Panel offsets\n");
  gotoxy(1, 15);  printf("Press 'q' to stop, 'r' to search for ref mark zero.");
  for(;;) {
    tDLBuff dl;

    do {
      int i;

      /* The other end is not supposed to send any responses
         before receiving a command, make sure by: */
      FlushInBuf(COM2);

      /* Ask for all sensor values by sending a command character. */
      WriteSerial(COM2, '*');

      /* Try to read exactly the expected number of reply bytes. */
      for (i=0; i<sizeof(tDownlink); i++) {
        dl.buff[i] = getOne();
      }
        /* Too few characters: trailing characters will be 0xff;
           extra characters will be flushed before the next read. */

    /* Message must have correct length at end. */
      if (dl.downlink.length != sizeof(dl)) {
        vSerialReadErrors++;
      }
    } while (dl.downlink.length != sizeof(dl));

	/* Test for keyboard actions. */
	if (kbhit()) {
	  switch (getch()) {
		case 'q':
		case 'Q':
		  goto lEnd;
        case 'r': {
#if LOCAL_IK121
          vIK121.startCountingFromZeroAtRefMark();
#else
          WriteSerial(COM2, 'r');
#endif
          break;
        }
        case 'l': {
#if LOCAL_IK121
          vIK121.latchAtNextRefMark();
#else
          WriteSerial(COM2, 'l');
#endif
          break;
        }
        case 'a': {
#if TEST_DA
          vET1728.set(0, getVal(" Azimuth D/A   ->"));
#endif
          if (vRunMode) {
            vAzPanelOffset = getDouble(" Azimuth offset (deg) ->");
          } else {
            vTargetAz = getDouble(" Azimuth deg   ->");
          }
          break;
        }
        case 'e': {
#if TEST_DA
          vET1728.set(1, getVal(" Elevation D/A ->"));
#endif
          if (vRunMode) {
            vElPanelOffset = getDouble(" Elevation offset (deg) ->");
          } else {
            vTargetEl = getDouble(" Elevation deg ->");
          }
          break;
        }
        case 'j': {
          vJumpDegrees = getDouble(" Jump amount deg ->");
          break;
        }
        case '8': {
          vTargetEl += vJumpDegrees;
          break;
        }
        case '2': {
          vTargetEl -= vJumpDegrees;
          break;
        }
        case '4': {
          vTargetAz -= vJumpDegrees;
          break;
        }
        case '6': {
          vTargetAz += vJumpDegrees;
          break;
        }
        case 't': {  /* toggle run mode / "track" VAX */
          vRunMode = ! vRunMode;
          break;
        }
        case 'i': {
          vLogInterval = getDouble(" Log interval (sec, <0 to stop) ->");
          vNextLogTime = 0;  // log next line immediately
          break;
        }
	  }  /* switch */
	}  /* if */

    /* If we are VAX-tracking, update target position. */
    if (vRunMode) {
      vTargetAz = vVAXIF.getAz();
      vTargetEl = vVAXIF.getEl();
    }

    /* Update VAX parallel encoder output. */
    {
      double vAzAsEssco = dl.downlink.incr.az + vAzPanelOffset;
      double vElAsEssco = dl.downlink.incr.el + vElPanelOffset;
      vVAXIF.sendAz(vAzAsEssco);
      vVAXIF.sendEl(vElAsEssco);
    }

    /* Update D/A output signals. */
    {
      static const double coeff = 10.0;  /* V per degree */
        /* We want -1.0--+1.0 degree error to match
           the maximum D/A output range which will be
           adjusted with resistors (at the moment 10V -> 1.2V). */
        /* +-1.0 degree comes from the original ESSCO 11-bit D/A
           which took input in BCD millidegrees, -0.999--+0.999. */

      vAzErr = vTargetAz - (dl.downlink.incr.az + vAzPanelOffset);
      vElErr = vTargetEl - (dl.downlink.incr.el + vElPanelOffset);
      vET1728.set(0 /*az*/, (vDiffVoltsAz = vAzErr * coeff));
      vET1728.set(1 /*el*/, (vDiffVoltsEl = vElErr * coeff));
    }

	/* Update display values. */
	{
      long int val;

#define COL1 16
#define COL2 30
	  gotoxy(COL1, 2);
	  printf("%6.1f", dl.downlink.abs.el);
	  gotoxy(COL2, 2);
	  printf("%6.1f", dl.downlink.abs.az);

#if LOCAL_IK121
      /* Trigger encoder value measurement. */
      vIK121.trigger();

      gotoxy(1,3);
      val = vIK121.get(/*axis=>*/ 0);
      printf("Local IK121 axis 0 (az) = %11ld = %11.6f  (ref search %1d)",
        val, ((double)val / (100.0*1024.0)),
          /* RON806 counts 36000 per 360degrees,
             and IK121 interpolates 1024-fold. */
          /* LSBit matches quite well tens of microdegrees,
             i.e. 5th digit after the decimal point. */
        vIK121.isSearchingForRefMark(/*axis=>*/ 0)
      );

      gotoxy(1,4);
      val = vIK121.get(/*axis=>*/ 1);
      printf("            axis 1 (el) = %11ld = %11.6f  (ref search %1d)",
        val, ((double)val / (100.0*1024.0)),
        vIK121.isSearchingForRefMark(/*axis=>*/ 1)
      );

      gotoxy(1,5);
      printf(" latched refmarks at axis 0 = %11ld, axis 1 = %11ld",
        vIK121.getLatchedRefMark(/*axis=>*/ 0),
        vIK121.getLatchedRefMark(/*axis=>*/ 1)
      );

#else
      /* Remote IK121 values. */
	  gotoxy(COL1, 3);
	   printf("%11.6f", dl.downlink.incr.el);
	   gotoxy(COL2, 3);
	   printf("%11.6f", dl.downlink.incr.az);
#endif

	  gotoxy(COL1, 4);
	  printf("%11.6f", vTargetEl + vElPanelOffset);
	  gotoxy(COL2, 4);
	  printf("%11.6f", vTargetAz + vAzPanelOffset);

	  gotoxy(COL1, 5);
	  printf("%11.6f", vElErr);
	  gotoxy(COL2, 5);
	  printf("%11.6f", vAzErr);

      /* xxx: vDiffVoltsEl, vDiffVoltsAz */

      gotoxy(COL1,7);
      printf("%11.6f", vVAXIF.getEl());
      gotoxy(COL2,7);
      printf("%11.6f\n", vVAXIF.getAz());

      gotoxy(COL1,8);
      printf("%11.6f", vElPanelOffset);
      gotoxy(COL2,8);
      printf("%11.6f\n", vAzPanelOffset);


	  gotoxy(1,10);
	  printf("A/D tilt-y-az: %9.4f", dl.downlink.tilt.az);
	  printf("  tilt-x-el: %9.4f (millidegrees)\n", dl.downlink.tilt.el);

	  printf("    tilt-Temp:  %8.4f", dl.downlink.tiltTemp);
	  printf("   air-Temp:  %8.4f C", dl.downlink.extTemp);

      gotoxy(60,12);
      printf("%10ld", vSerialReadErrors);

      gotoxy(1,13);
      if (vRunMode) {
        printf("VAX-TRACK, press 't' for manual position.");
      } else {
        printf("MANUAL POS, press 't' for VAX-tracking.");
      }
      clreol();
	}  /* of block update display values */

    // If log line writing interval has expired, write one line.
    if ((vLogInterval >= 0.0) && (time(&vLastTime) >= vNextLogTime)) {
      fprintf(vLogFile,
        "%ld\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%9.4f\t%9.4f\t%8.4f\t%8.4f\n",
        vLastTime,
        dl.downlink.incr.az,
        dl.downlink.incr.el,
        vAzErr,
        vElErr,
        dl.downlink.tilt.az,
        dl.downlink.tilt.el,
        dl.downlink.tiltTemp,
        dl.downlink.extTemp
      );
      vNextLogTime = vLastTime + (int)(vLogInterval + 0.5);
    }  // if needs to write a log line
  }  /* for ever */

lEnd:
  /* Cleanups. */
  printf("\n\n\n");
  (void)fclose(vLogFile);
  CloseSerial(COM2);
  vET1728.cleanup();
#if LOCAL_IK121
  vIK121.cleanup();
#endif
  vVAXIF.cleanup();
  return (0);
}  /* main */
