#include "registration.h"
RCSID(serial_c, "$Id: serial.c,v 1.5 1999/10/19 12:42:15 amn Exp $");
/* ------------------------------------------------------------------ */
/*  SERIAL.C -- Buffered IBM Serial Port handler  */

/* $Log: serial.c,v $
/* Revision 1.5  1999/10/19 12:42:15  amn
/* Added new-style command long words; debugging serial.c TIMEOUT (50000L?).
/*
/* Revision 1.4  1999/10/15 12:46:44  amn
/* Patched ROC412 to be 10 bits only until wiring is corrected.
/* Both downlink and uplink in 'downlink.h'.
/* 'serial.c' enhanced to support (and require!) receiver FIFOs.
/*
/* Revision 1.3  1999/10/12 07:55:18  amn
/* Added registration and log.
/* */

/* 08-Sep-1996  amn  Suspecting a shared variable problem. */
/* 02-Sep-1996  amn  Original adaptation to Borland C++ 3.1. */

#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include <alloc.h>
#include "serial.h"

/* ring buffer definitions */

#define BUFLEN	1024                   /* buffer lenght               */

/* i/o-port definitions */

#define PIC_MASK  0x21                 /* 8259 mask register          */
#define PIC_EOI   0x20                 /* 8259 EOI command            */
#define TIMEOUT  50000L

static /***was: const***/ struct
{
  int data,                            /* 8250 data register          */
      ier,                             /* interrupt enable register   */
      fcr,                             /* amn: FIFO control register  */
      lcr,                             /* line control register       */
      mcr,                             /* modem control register      */
      stat,                            /* status register             */
      intno,                           /* interrupt number            */
      mask,                            /* PIC mask                    */
      opened;                          /* 0/1 - port closed/opened    */
  volatile unsigned char *head, *tail;
  unsigned char *buffer;
  void interrupt (*handler)(void);     /* interrupt handler address   */
  void interrupt (*old_handler)(void); /* old interrupt vector        */
  long max_timeout;
} ports[2] =
{
  /* COM1 */
    {0x3f8, 0x3f9, 0x3fa, 0x3fb, 0x3fc, 0x3fd, 0x0c, 0x10, 0, 0,0,0,0,0,0},

  /* COM2 */
    {0x2f8, 0x2f9, 0x2fa, 0x2fb, 0x2fc, 0x2fd, 0x0b, 0x08, 0, 0,0,0,0,0,0}
};

static void (*interceptor[2])(char) = {NULL,NULL};

/***************************************************************************/

static void interrupt ReadChr1 (void)

/* 8250 UART Interrupt server routine for COM1 */

{
  do {  /* amn: added for emptying FIFOs */
    *(ports[0].head) = (unsigned char)inportb(ports[0].data);
    if (interceptor[0] != NULL)
      interceptor[0](*ports[0].head);

    if (++ports[0].head >= &ports[0].buffer[BUFLEN])
      ports[0].head = ports[0].buffer;
  } while ( inportb(ports[0].stat) & 0x01);

  outportb(PIC_EOI, 0x20);
}

/***************************************************************************/

static void interrupt ReadChr2 (void)

/* 8250 UART Interrupt server routine for COM2 */

{
  do {  /* amn: added for emptying FIFOs */
    *(ports[1].head) = (unsigned char)inportb(ports[1].data);
    if (interceptor[1] != NULL)
      interceptor[1](*ports[1].head);

    if (++ports[1].head >= &ports[1].buffer[BUFLEN])
      ports[1].head = ports[1].buffer;
  } while ( inportb(ports[1].stat) & 0x01);

  outportb(PIC_EOI, 0x20);
}

/***************************************************************************/

int BufLen (int pn)

/* Returns number of characters in buffer. */

{
  unsigned char /***was: huge***/ *head,*tail;

  head = (unsigned char *)ports[pn].head;
  tail = (unsigned char *)ports[pn].tail;

  if (head >= tail)
    return(head-tail);
  else {
    int diff = tail-head;

    return(BUFLEN-diff);
  }
}

/***************************************************************************/

void CloseSerial (int pn)

/* Terminates serial communication */

{
  if (ports[pn].opened)
  {
    /* set 8259 PIC chip (disable UART interrupts) */
    outportb(PIC_MASK, inportb(PIC_MASK) | ports[pn].mask);

    /* reset interrupt vector */
    setvect(ports[pn].intno, ports[pn].old_handler);

    free(ports[pn].buffer);
    ports[pn].opened = 0;
    interceptor[pn] = NULL;
#define DEBUG_TIMEOUT 1
#if DEBUG_TIMEOUT
    printf("serial.c: maximum timeout for COM%d: = %ld\n",
      pn+1,
      ports[pn].max_timeout
    );
#endif
    ports[pn].max_timeout = 0L;
  }
}

/***************************************************************************/

void FlushInBuf (int pn)
{
  ports[pn].head = ports[pn].tail = ports[pn].buffer;
}

/***************************************************************************/

void MaskInt2 (void)

/* This disables INTR2 */

{
  outportb(PIC_MASK, inportb(PIC_MASK) | 4);
}

/***************************************************************************/
/* Initializes serial communication. If you want to intercept characters
 * from a serial port, give pointer to the interceptor as parameter
 * intercept. If not, use NULL for intercept.
 *
 * returns:
 *  NULL if failed
 */

int OpenSerial(
  int pn,
  long baud,
  unsigned char cbyte,
  void (*intercept)(char)
) {
  unsigned int divisor;

  if (ports[pn].opened) return(-1);  /* Already opened! */

  divisor = (unsigned int)(115200L / baud);

  ports[0].handler = ReadChr1;
  ports[1].handler = ReadChr2;

  /* allocate receiving buffer */
  if ((ports[pn].buffer = (unsigned char *)calloc(1,BUFLEN)) == NULL)
    return(NULL);

  /* reset buffer */
  ports[pn].head = ports[pn].tail = ports[pn].buffer;

  /* set interrupt vector */
  ports[pn].old_handler = getvect(ports[pn].intno);
  setvect(ports[pn].intno, ports[pn].handler);

  /* set 8250 UART chip */
  outportb(ports[pn].lcr, 0x83);        /* set baudrate               */
  outportb(ports[pn].data, divisor & 0xff);
  outportb(ports[pn].ier, divisor >> 8);
  outportb(ports[pn].lcr, cbyte);       /* set data/stop bits, parity */
  outportb(ports[pn].mcr, 0x0b);        /* DTR, RTS, OUT1 & OUT2      */
  outportb(ports[pn].ier, 0x01);        /* enable rec interrupt       */

  /* amn: Clearing & disabling FIFOs. */
  outportb(ports[pn].fcr, 0x06);  /* 0x02 rcvr, 0x04 xmit FIFOs */

  /* Enable receive FIFO. */
  outportb(ports[pn].fcr, 0x01);  /* 0x01 enable, 0x40 mask for trigger at 4 */

  inportb(ports[pn].data);              /* make UART happy            */

  /* set 8259 PIC chip (enable data ready interrupts from UART) */
  outportb(PIC_MASK, inportb(PIC_MASK) & ~ports[pn].mask);

  ports[pn].opened = 1;
  interceptor[pn] = intercept;
  return(-1);
}

/***************************************************************************/
/* Set line control register */

void PortControl(
  int pn,
  unsigned char cbyte
) {
  long i=0;

  while((inportb(ports[pn].stat) & 0x60) != 0x60 && i++ < TIMEOUT);
  if (i > ports[pn].max_timeout) {
    ports[pn].max_timeout = i;
  }
  outportb(ports[pn].lcr, cbyte);
}

/***************************************************************************/

int ReadSerial (int pn)

/* Reads one character from serial port,
 * returns -1 if there are no characters waiting.
 */

{
  register int c;

  if (ports[pn].head != ports[pn].tail)
  {
    c = *ports[pn].tail;
    if (++ports[pn].tail >= &ports[pn].buffer[BUFLEN])
      ports[pn].tail = ports[pn].buffer;
    return(c);
  }
  else return(-1);
}

/***************************************************************************/

void UnMaskInt2 (void)

/* This enables INTR2 */

{
  outportb(PIC_MASK, inportb(PIC_MASK) & 0xFB);
}

/***************************************************************************/

void WriteSerial (int pn, unsigned char ch)

/* Writes one character to serial port. */

{
  long i=0;

  while(!(inportb(ports[pn].stat) & 0x20) && i++ < TIMEOUT);
  if (i > ports[pn].max_timeout) {
    ports[pn].max_timeout = i;
  }
  outportb(ports[pn].data, ch);
}

/***************************************************************************/

void WriteStrSerial (int pn, unsigned char *str, int len)

/* Writes len characters from str to serial port. */

{
  long i=0;

  for (; len > 0; str++, len--)
  {
    while(!(inportb(ports[pn].stat) & 0x20) && i++ < TIMEOUT);
    if (i > ports[pn].max_timeout) {
      ports[pn].max_timeout = i;
    }
    outportb(ports[pn].data, *str);
  }
}

/***************************************************************************/
/*
    void main(void) {
	int c;

        OpenSerial(COM1,9600,D8S1N,NULL);

	while (1) {
	    while ((c = ReadSerial(COM1)) == -1)
		if (kbhit())
		    WriteSerial(COM1,getch());

	    putch(c & 0x7f);
	}

	CloseSerial(COM1);
    }
*/
