/* up/ant.cpp  --  reading sensors and encoders in "up" PC.  */

/* When         Who  What */
/* 15-Feb-1999  amn, jen  Separated cSensor and cET28xx to files. */
/* 27-Sep-1996  amn  Length added to downlink struct. */
/* 09-Sep-1996  amn  Added IK121/RON806 handling. */
/* 08-Sep-1996  amn  Adding getting integer raw values to cSensor. */
/* 05-Sep-1996  amn  Putting in azimuth abs. encoder coefficients. */
/* 03-Sep-1996  amn  Gray code decoding was missing in ROC412 code. */
/* 02-Sep-1996  amn  Serial communications. */
/* 30-Aug-1996  amn  Sensor value conversions to 'double'. */
/* 27-Aug-1996  amn  To C++; decoding Gray code of ROC412. */
/* 26-Aug-1996  amn  Original version. */

#include <conio.h>
#include <stdio.h>
#include <dos.h>

#include <assert.h>
#include <math.h>  /* for M_PI */
#include <mem.h>  /* for memcpy() */

#include "serial.h"

#define ET2816_BASE 0x210
#define ET1722B_BASE 0x220

#define DEBUG_REPEATS 0
#define DEBUG_MEAS_LENGTH 1


#include "cSensor.cpp"

#include "cROC4xx.cpp"


/* ------------------------------------------------------------------ */
/* Heidenhain IK121/RON806 incremental encoders. */
#include "cIK121.cpp"


#include "cET28xx.cpp"


/* Definition of the struct of all sensor values on the serial line down. */
#include "downlink.h"


/* ------------------------------------------------------------------ */
int main(int argc, char **argv) {
  class cPortROC412 vAzAbs;
  class cPortROC412 vElAbs;
  class cIK121 vIK121;
  class cET28xx vET2816;
  class cSensorChannel vTiltYAz, vTiltXEl, vTiltTemp, vExtTemp;

  /* "Use" argc, argv to prevent compiler warning messages. */
  (void)argc; (void)argv;

  vAzAbs.init(
	/*baseAddr=>*/ ET1722B_BASE,
	/*conn=>*/ 0,
	/*numOfBitsMask=>*/ 0x0800
  );
  vAzAbs.setCalibration(/*offset=>*/ 540.2062, /*coeff=>*/ -3.000867);
    /* Measured on 04-Sep-1996 with ant. range -360..0. */

  vElAbs.init(
	/*baseAddr=>*/ ET1722B_BASE,
	/*conn=>*/ 1,
	/*numOfBitsMask=>*/ 0x0800
  );
  vElAbs.setCalibration(/*offset=>*/ +185.24, /*coeff=>*/ -1.0);
    /* xxx: not measured yet */

  /* Incremental encoders. */
  vIK121.init(/*baseAddr=>*/ 0x320);

  vET2816.init(/*baseAddr=>*/ ET2816_BASE);
  vET2816.setCalibration(/*offset=>*/ 0.0, /*coeff=>*/ 1.0);
    /* calibration of A/D board itself */

  /* Individual channel calibration coefficients. */
  vTiltYAz.init(&vET2816);
  vTiltYAz.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 0.4833 /* low gain; high=0.05002 uradians / mV */
                 * 180.0 / M_PI  /* millidegrees / V */
  );

  vTiltXEl.init(&vET2816);
  vTiltXEl.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 0.4740 /* low gain; high=0.04991 */
                 * 180.0 / M_PI  /* millidegrees / V */
  );

  vTiltTemp.init(&vET2816);
  vTiltTemp.setCalibration(/*offset=>*/ 0.0, /*coeff=>*/ 100.0);

  vExtTemp.init(&vET2816);
  vExtTemp.setCalibration(/*offset=>*/ 0.0, /*coeff=>*/ 100.0);

  /* Serial downlink. */
  OpenSerial(COM2, 115200L, D8S1N, NULL);

  clrscr();
  gotoxy(1, 10);  printf("Press 'q' to stop.");
  for(;;) {

    /* ----- */
    /* Test for incoming serial characters. */
    if (BufLen(COM2) > 0) {
      unsigned char c = (unsigned char)ReadSerial(COM2);
      union {
        char line[81];
        tDownlink downlink;
      };
      int len;
      double dval;

      switch (c) {
        case '*': {
          vIK121.trigger();
          downlink.incr.az = (double)vIK121.get(/*axis=>*/ 0) / 102400.0
            + 1.11 - 0.022 /* measured 29-Sep-1996 */;
          downlink.incr.el = -(double)vIK121.get(/*axis=>*/ 1) / 102400.0
           + +42.516475 + 0.035 /* measured 29-Sep-1996 */;
          downlink.abs.az = vAzAbs.get();
          downlink.abs.el = vElAbs.get();
          vET2816.setChannel(/*channel=>*/ 0, SE, /*gain=>*/ 1);
          downlink.tilt.az = vTiltYAz.get();
          vET2816.setChannel(/*channel=>*/ 1, SE, /*gain=>*/ 1);
          downlink.tilt.el = vTiltXEl.get();
          vET2816.setChannel(/*channel=>*/ 3, SE, /*gain=>*/ 1);
          downlink.extTemp = vExtTemp.get();
          vET2816.setChannel(/*channel=>*/ 2, SE, /*gain=>*/ 1);
          downlink.tiltTemp = vTiltTemp.get();
          /* 'downlink' is layered on top of 'line'. */
          len = downlink.length = sizeof(downlink);
          break;
        }
        case 'r': {
          vIK121.startCountingFromZeroAtRefMark();
          break;
        }
        case 'l': {
          vIK121.latchAtNextRefMark();
          break;
        }
        case 'a': {
          len = sprintf(line, "%c -> absolute azimuth = %7.2f (Bin = %04X, Gray = %04X)\r\n",
            c,
            vAzAbs.get(),
            vAzAbs.getStableBin(),
            vAzAbs.getStableGray()
          );
          break;
        }
        case '0': {
          vET2816.setChannel(/*channel=>*/ (c-'0'), SE, /*gain=>*/ 1);
          dval = vTiltYAz.getUncalibrated();
          len = sprintf(line, "%c -> A/D ch %c (y/\"az\" tilt) = %8.5fV = %8.3fmdeg\r\n",
            c, c,
            dval, vTiltYAz.calibrated(dval)
          );
          break;
        }
        case '1': {
          vET2816.setChannel(/*channel=>*/ (c-'0'), SE, /*gain=>*/ 1);
          dval = vTiltXEl.getUncalibrated();
          len = sprintf(line, "%c -> A/D ch %c (x/elev tilt) = %8.5fV = %8.3fmdeg\r\n",
            c, c,
            dval, vTiltXEl.calibrated(dval)
          );
          break;
        }
        case '2': {
          vET2816.setChannel(/*channel=>*/ (c-'0'), SE, /*gain=>*/ 8);
          dval = vTiltTemp.getUncalibrated();
          len = sprintf(line,
            "%c -> A/D ch %c (tiltm. temp) = %8.5fV = %5.1fC\r\n",
            c, c,
            dval, vTiltTemp.calibrated(dval)
          );
          break;
        }
        case '3': {
          vET2816.setChannel(/*channel=>*/ (c-'0'), SE, /*gain=>*/ 8);
          dval = vExtTemp.getUncalibrated();
          len = sprintf(line,
            "%c -> A/D ch %c (extern temp) = %8.5fV = %5.1fC\r\n",
            c, c,
            dval, vTiltTemp.calibrated(dval)
          );
          break;
        }
        default: {
          len = sprintf(line, "%c -> unrecognized command letter, try 'a/0/1/2/3'\r\n", c);
          break;
        }
      }  /* switch */

      /* Send a response down. */
      WriteStrSerial(COM2, (unsigned char *)line, len);
      /* The other end is not supposed to send new commands
         before receiving a previous response, make sure by: */
      FlushInBuf(COM2);
    }  /* if serial characters have arrived */
    /* ----- */

	/* Test for keyboard actions. */
	if (kbhit()) {
	  switch (getch()) {
		case 'q':
		case 'Q':
		  goto lEnd;
	  }  /* switch */
	}  /* if */

	/* Update display values. */
	{
#if 0
	  int val;
#endif
	  double dval;

#if 0
	  gotoxy(1,3);
	  val = vAzAbs.getRawGray();
	  printf("Gray %04X = %4d", val, val);
#endif
	  gotoxy(1,4);
#if 0
	  val = vAzAbs.getStableBin();
#endif
	  dval = vAzAbs.get();
	  printf("AzAbs %7.2f", dval);

#if 0
	  gotoxy(1,5);
	  vET2816.setChannel(/*channel=>*/ 0, SE, /*gain=>*/ 1);
	  val = vET2816.getBin();
	  dval = vET2816.get();
	  printf("A/D  tilt-y-az: %04X = %6d = %8.4f", val, val, dval);

	  vET2816.setChannel(/*channel=>*/ 1, SE, /*gain=>*/ 1);
	  val = vET2816.getBin();
	  dval = vET2816.get();
	  printf("  tilt-x-el: %04X = %6d = %8.4f\n", val, val, dval);

	  vET2816.setChannel(/*channel=>*/ 2, SE, /*gain=>*/ 8);
	  val = vET2816.getBin();
	  dval = vET2816.get();
	  printf("     tilt-Temp: %04X = %6d = %8.4f", val, val, dval);

	  vET2816.setChannel(/*channel=>*/ 3, SE, /*gain=>*/ 8);
	  val = vET2816.getBin();
	  dval = vET2816.get();
	  printf("  air-Temp:  %04X = %6d = %8.4f", val, val, dval);
#endif
	}

  }  /* for ever */

lEnd:
  /* Cleanups. */
  printf("\n\n\n");
  CloseSerial(COM2);
  vExtTemp.cleanup();
  vTiltTemp.cleanup();
  vTiltXEl.cleanup();
  vTiltYAz.cleanup();
  vET2816.cleanup();
  vIK121.cleanup();
  vElAbs.cleanup();
  vAzAbs.cleanup();
  return (0);
}  /* main */
