/* vaxd/ant.cpp  --  Conveying 'downf'-up PC sensor data to VAX.  */
/*                   and VAX commands back via VAXINTF board. */

/* When         Who  What */
/* 11-May-1999  amn  Panel offsets were ignored. */
/* 16-Mar-1999  amn  Original 'vaxd' version. */
/* 16-Apr-1997  amn  Adding logging. */
/* 02-Dec-1996  amn  Fixing 'getUntilStaysSame()'.
/* 29-Nov-1996  amn  Jouko's new VAXIF board. */
/* 01-Oct-1996  amn  c8255::setPortMode didn't use 'this->out()' correctly. */
/* 27-Sep-1996  amn  VAX, serial lost bytes check, VAX out. */
/* 09-Sep-1996  amn  Remote IK121. */
/* 08-Sep-1996  amn  Initializing the reference mark. */
/* 05-Sep-1996  amn  Original version. */

#include <conio.h>
#include <stdio.h>
#include <dos.h>

#include <assert.h>
#include <time.h>  /* for time functions */

#include "serial.h"

/* Reads one character; if timeout, returns 0xff. */
unsigned char getOne(void) {
  unsigned int i = 50000;
  int c;

  while (i && ((c = ReadSerial(COM1)) == -1))
    i--;
  return ((unsigned char)c);
}


/* -------------------------------------------------------------------- */
/* New VAXIF MACH211-based VAX interface board. */
class cVAXIF {
private:
  unsigned int baseAddr;
  virtual long int getOne(int offset);
  virtual long int getOneUntilStaysSame(int offset);
  virtual double getConverted(int offset);
  virtual void sendOne(int offset, long int value);
  virtual void sendConverted(int offset, double value);
public:
  virtual void init(int baseAddr);
  virtual void cleanup(void);
  virtual double getAz(void);
  virtual double getEl(void);
  virtual void sendAz(double az);
  virtual void sendEl(double el);
};  /* cVAXIF */

void cVAXIF::init(int baseAddr) {
  this->baseAddr = baseAddr;
  /* xxx: Set outputs to VAX to ? */
}

void cVAXIF::cleanup(void) {
}

long int cVAXIF::getOne(int offset) {
  long int value;

  value = inportb(this->baseAddr+offset+2);  /* high */
  value = (value<<8) | inportb(this->baseAddr+offset+1);  /* mid */
  value = (value<<8) | inportb(this->baseAddr+offset);  /* low */
    /* At the moment the board doesn't have any special latching */
    /* based on the order in which bytes are being read. */
  return value;
}

long int cVAXIF::getOneUntilStaysSame(int offset) {
  long int prevValue;
  long int value;
  int retries = 0;

  value = this->getOne(offset);
  do {
    prevValue = value;
    value = this->getOne(offset);
    retries++;
  } while ((prevValue != value) && (retries < 10));
  /* xxx: what to do if retries > 10? */
  return value;
}

double cVAXIF::getConverted(int offset) {
  long int value;

  value = this->getOneUntilStaysSame(offset);

  /* Extend MSB == 1  => negative in 2-complement. */
  if (value & 0x00200000) {
    /* MSB sign bit == 1, extend. */
    value |= 0xFFE00000;
  }
  return ((double)value * 360.0 / 2097152.0);
}  /* getConverted */

void cVAXIF::sendOne(int offset, long int value) {
  value <<= 2;  /* VAX aligns MSBs, so fill two LSBs with 0 */
  outportb(this->baseAddr+offset,  value & 0x000000FF);  /* low */
  outportb(this->baseAddr+offset+1, (value & 0x0000FF00)>>8);  /* mid */
  outportb(this->baseAddr+offset+2, (value & 0x00FF0000)>>16);  /* high */
    /* High byte latches all of 24(22/21) bits to VAX. */
}

void cVAXIF::sendConverted(int offset, double value) {
  long int signBit;

  if (value < 0.0) {
    value = -value;
    signBit = 0x00000000;
  } else {
    signBit = 0x00200000;  /* positive == sign bit is on; not 2-compl. */
  }
  this->sendOne(offset, (long int)(value * 2097152.0 / 360.0) | signBit);
}  /* sendConverted */

double cVAXIF::getAz(void) {
  return this->getConverted(/*offset=>*/0);
}  /* getAz */

double cVAXIF::getEl(void) {
  return this->getConverted(/*offset=>*/4);
}  /* getEl */

void cVAXIF::sendAz(double az) {
  this->sendConverted(/*offset=>*/0, az);
}  /* sendEl */

void cVAXIF::sendEl(double el) {
  this->sendConverted(/*offset=>*/4, el);
}  /* sendEl */


typedef struct sVaxlink {
  double az;
  double el;
  int length;  // Message length to detect lost bytes.
} tVaxlink;

typedef union uVLBuff {
  tVaxlink vaxlink;
  unsigned char buff[sizeof(tVaxlink)];
} tVLBuff;


int getVal(char *prompt) {
  char buff[255];
  int temp;

  printf("%s", prompt);
  gets(buff);
  (void)sscanf(buff, "%d", &temp);
  return (temp);
}

double getDouble(char *prompt) {
  char buff[255];
  double temp;

  printf("%s", prompt);
  gets(buff);
  (void)sscanf(buff, "%lg", &temp);
  return (temp);
}


/* ------------------------------------------------------------------ */
int main(int argc, char **argv) {
  class cVAXIF vVAXIF;
  double vTargetAz = -36.0;
  double vTargetEl = 46.0;
  double vAzAsEssco = 0.0;  // init to 0.0; before first serial message
  double vElAsEssco = 0.0;  // these are shown on-screen
  double vAzPanelOffset = 0.0;
  double vElPanelOffset = 0.0;
  double vAzErr;
  double vElErr;
  double vDiffVoltsAz;
  double vDiffVoltsEl;
  double vJumpDegrees = 0.0;
  int vRunMode = 1;  /* 0==stopped as manual position, 1==VAX-track */
  double vLogInterval = -1.0;  // seconds; when negative, logging disabled
  FILE *vLogFile;
  time_t vNextLogTime = 0;  // immediately, since 0 means year 1970
  time_t vLastTime;
  long int vSerialReadErrors = 0;

  /* "Use" argc, argv to prevent compiler warning messages. */
  (void)argc; (void)argv;

  /* New VAXIF MACH211-based VAX interface board. */
  /* Supports both getting VAX commanded az/el and */
  /* sending encoder az/el values back to VAX. */
  vVAXIF.init(
    /*baseAddr=>*/ 0x300
  );

  /* Serial "VAX-link". */
  OpenSerial(COM1, 115200L, D8S1N, NULL);

  /* Open log file always in append mode. */
  assert ( (vLogFile = fopen("ant.log", "at")) != NULL );  // append in text mode

  clrscr();
  gotoxy(1, 1);
  printf("                 Elevation   Azimuth  (degrees)\n");
  printf("\n");
  printf("Incremental\n");
  printf("Target\n");
  printf("Pos. error\n");
  printf("\n");
  printf("VAX Target\n");
  printf("Panel offsets\n");
  gotoxy(1, 15);  printf("Press 'q' to stop, 'r' to search for ref mark zero.");
  for(;;) {
    tVLBuff vl;

    // Serial line actions.
    if (BufLen(COM1) > 0) {
      unsigned char c;
      int i;

      /* Wait for initiating '*'. */
      c = getOne();
      if (c == '*') {

        /* Try to read exactly the expected number of incoming message bytes. */
        for (i=0; i<sizeof(tVaxlink); i++) {
          vl.buff[i] = getOne();
        }
          /* Too few characters: trailing characters will be 0xff;
             extra characters will be flushed while waiting for '*'. */

        /* Message must have correct length at end. */
        if (vl.vaxlink.length != sizeof(vl)) {
          vSerialReadErrors++;
        } else {
          // Finally serial incoming message can be assumed ok,
          // get the values and reply back.

          /* Update encoder output to VAX parallel interface. */
          vAzAsEssco = vl.vaxlink.az + vAzPanelOffset;
          vElAsEssco = vl.vaxlink.el + vElPanelOffset;
          vVAXIF.sendAz(vAzAsEssco);
          vVAXIF.sendEl(vElAsEssco);

          /* If we are VAX-tracking, update target position. */
          if (vRunMode) {
            vTargetAz = vVAXIF.getAz();
            vTargetEl = vVAXIF.getEl();
          }
          vl.vaxlink.az = vTargetAz - vAzPanelOffset;
          vl.vaxlink.el = vTargetEl - vElPanelOffset;
          vl.vaxlink.length = sizeof(vl);

          /* Send a response to VAX-antcon-PC. */
          WriteStrSerial(COM1, (unsigned char *)(vl.buff), vl.vaxlink.length);
          /* The other end is not supposed to send new commands
             before receiving a previous response, make sure by: */
          FlushInBuf(COM1);
        }  // else if incoming serial message was ok
      }  // if initiator marker '*' found
    }  // if something on serial line

	/* Test for keyboard actions. */
	if (kbhit()) {
	  switch (getch()) {
		case 'q':
		case 'Q':
		  goto lEnd;
        case 'a': {
          if (vRunMode) {
            vAzPanelOffset = getDouble(" Azimuth offset (deg) ->");
          } else {
            vTargetAz = getDouble(" Azimuth deg   ->");
          }
          break;
        }
        case 'e': {
          if (vRunMode) {
            vElPanelOffset = getDouble(" Elevation offset (deg) ->");
          } else {
            vTargetEl = getDouble(" Elevation deg ->");
          }
          break;
        }
        case 'j': {
          vJumpDegrees = getDouble(" Jump amount deg ->");
          break;
        }
        case '8': {
          vTargetEl += vJumpDegrees;
          break;
        }
        case '2': {
          vTargetEl -= vJumpDegrees;
          break;
        }
        case '4': {
          vTargetAz -= vJumpDegrees;
          break;
        }
        case '6': {
          vTargetAz += vJumpDegrees;
          break;
        }
        case 't': {  /* toggle run mode / "track" VAX */
          vRunMode = ! vRunMode;
          break;
        }
        case 'i': {
          vLogInterval = getDouble(" Log interval (sec, <0 to stop) ->");
          vNextLogTime = 0;  // log next line immediately
          break;
        }
	  }  /* switch */
	}  /* if */

    /* If we are VAX-tracking, update target position. */
    /* (Also when ant/downf-PC doesn't ask via serial line.) */
    if (vRunMode) {
      vTargetAz = vVAXIF.getAz();
      vTargetEl = vVAXIF.getEl();
    }

    /* Update position error values (on-screen). */
    vAzErr = vTargetAz - vAzAsEssco;
    vElErr = vTargetEl - vElAsEssco;

	/* Update display values. */
	{
      long int val;

#define COL1 16
#define COL2 30

      /* Remote IK121 values. */
	  gotoxy(COL1, 3);
	   printf("%11.6f", vElAsEssco - vElPanelOffset);
	   gotoxy(COL2, 3);
	   printf("%11.6f", vAzAsEssco - vAzPanelOffset);

	  gotoxy(COL1, 4);
	  printf("%11.6f", vTargetEl + vElPanelOffset);
	  gotoxy(COL2, 4);
	  printf("%11.6f", vTargetAz + vAzPanelOffset);

	  gotoxy(COL1, 5);
	  printf("%11.6f", vElErr);
	  gotoxy(COL2, 5);
	  printf("%11.6f", vAzErr);

      gotoxy(COL1,7);
      printf("%11.6f", vVAXIF.getEl());
      gotoxy(COL2,7);
      printf("%11.6f\n", vVAXIF.getAz());

      gotoxy(COL1,8);
      printf("%11.6f", vElPanelOffset);
      gotoxy(COL2,8);
      printf("%11.6f\n", vAzPanelOffset);

      gotoxy(60,12);
      printf("%10ld", vSerialReadErrors);

      gotoxy(1,13);
      if (vRunMode) {
        printf("VAX-TRACK, press 't' for manual position.");
      } else {
        printf("MANUAL POS, press 't' for VAX-tracking.");
      }
      clreol();
	}  /* of block update display values */

    // If log line writing interval has expired, write one line.
    if ((vLogInterval >= 0.0) && (time(&vLastTime) >= vNextLogTime)) {
      fprintf(vLogFile,
        "%ld\t%11.6f\t%11.6f\t%11.6f\t%11.6f\n",
        vLastTime,
        vAzAsEssco,
        vElAsEssco,
        vAzErr,
        vElErr
      );
      vNextLogTime = vLastTime + (int)(vLogInterval + 0.5);
    }  // if needs to write a log line
  }  /* for ever */

lEnd:
  /* Cleanups. */
  printf("\n\n\n");
  (void)fclose(vLogFile);
  CloseSerial(COM1);
  vVAXIF.cleanup();
  return (0);
}  /* main */
